/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.assemblies;

import ist.palcom.collections.LinearMap;
import ist.palcom.collections.LinearSet;
import ist.palcom.debugging.DebugDispatcher;
import ist.palcom.discovery.CommandEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.palcomthreads.Semaphore;
import ist.palcom.resource.descriptor.AssemblyID;
import ist.palcom.resource.descriptor.Command;
import ist.palcom.resource.descriptor.CommandInfo;
import ist.palcom.resource.descriptor.ConnectionDeclList;
import ist.palcom.resource.descriptor.ConnectionInfo;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.GroupInfo;
import ist.palcom.resource.descriptor.Info;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDAssemblyVer;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.Param;
import ist.palcom.resource.descriptor.ParamInfo;
import ist.palcom.resource.descriptor.ServiceDecl;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.resource.descriptor.ServiceUse;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.resource.descriptor.URNLeaf;
import ist.palcom.resource.descriptor.VersionPart;
import ist.palcom.resourceAddressing.AInfoEvent;
import ist.palcom.services.AbstractService;
import ist.palcom.services.SelfTestProvider;
import ist.palcom.services.SelfTester;
import ist.palcom.services.ServiceProxyEvent;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * An Assembly is a service that manages a set of connections, and that can have
 * a script coordinating services.
 * 
 * The assembly info is read from disk and stored on disk for later use. The
 * infos are stored after each update, for robustness.
 */
public class Assembly extends AbstractService {
	
	private AbstractService dp;
	private Command logCmd;
	private Command logMode;
	private String logStr;
	private LogCommandThread logCmdThread;
	private Command statusCmd;
	private String statusStr;
	protected boolean connectionEstablished[];
	
	private ResourceThread resourceThread;
	private PalcomThread infoThread;
		
	/* Commands for the debug service */
	private static final String CMD_CLEAR = "clear_log";
	private static final String CMD_RUN = "run";
	private static final String CMD_PAUSE = "pause";
	private static final String CMD_STEP = "step";
	private static final String CMD_TOGGLE_LOG = "toggle_logging";
	private static final String CMD_LOG_MODE = "logging_mode";
	private static final String CMD_LOG_TEXT = "log";
	
	
	/** The AssemblyInfo for this assembly. */
	public PRDAssemblyVer assemblyVer;
	
	protected CommandThread commandThread;
    
    private ServiceProxyThread spThread;
    
	private AssemblyDeviceContext context() {
		return (AssemblyDeviceContext) context;
	}

		
	public Assembly(AssemblyDeviceContext context,
			PRDAssemblyVer assemblyInfo, int period, LocalSID localSID) throws IOException {
		super(context, assemblyInfo.getName()+"Customer", false, true, PRDService.UNICAST, "P1",
                false, "1", "Assembly", localSID);
    	AssemblyID version = assemblyInfo.getVersion();
    	VersionPart vp1 = version.getCreatingVersionPart();
		VersionPart vp2 = version.getUpdatingVersionPart();
		VersionPart vp3 = version.getPreviousVersionPart();
		VersionPart vp4 = version.getMergedFromVersionPart();
		if(vp1 != null) {
			vp1 = (VersionPart) vp1.fullCopy();
			vp1.setServicePartID(vp1.getServicePartID()+"c");
		}
		if(vp2 != null) {
			vp2 = (VersionPart) vp2.fullCopy();
			vp2.setServicePartID(vp2.getServicePartID()+"c");
		}
		if(vp3 != null) vp3 = (VersionPart) vp3.fullCopy();
		if(vp4 != null) vp4 = (VersionPart) vp4.fullCopy();
		//XXX HERE BE DRAGONS!
		ServiceID serviceID = new ServiceID(vp1, vp2, vp3, vp4);
		
//		ServiceID serviceID = new ServiceID(
//        		new DeviceID("X:1"), "TF4", new DeviceID("X:1"), "TF4", null, null, null, null);
		setServiceInstanceID(new ServiceInstanceID(context.getDeviceInfo().getDeviceID(), serviceID, 
				context.getNewInstanceNumber(serviceID)));
		
//		proxies = new HashMap<String, ServiceProxy>();
		this.assemblyVer = assemblyInfo;
		
		commandThread = new CommandThread();
		logCmdThread = new LogCommandThread();
		
        spThread = new ServiceProxyThread();
        addServiceProxyListener(spThread);
        
        resourceThread = new ResourceThread();
		infoThread = new ConnectionInfoThread();
		
		int numC = assemblyInfo.getConnections().getNumConnectionDecl();
		connectionEstablished = new boolean[numC];
		for(int i=0; i<numC;i++) {
			connectionEstablished[i]=false;
		}
	}
	
	private String selfTestResult;
	
	protected void setSelfTestResult(String s) {
		selfTestResult = s;
		selfTestChanged=true;;
		doSelfTest();
	}
	
	protected SelfTester getSelfTester() {
		return new SelfTester() {
			public String doSelfTest() {
				if(selfTestResult == null || selfTestResult.equals(SelfTester.STATUS_OPERATIONAL)) {
					if(allConnectionsEstablished()) {
						return SelfTester.STATUS_OPERATIONAL;
					} else
						return SelfTester.STATUS_PARTIAL;
				}
				return selfTestResult;
			}
		};
	}

	/** for self-test
	 * 
	 * @return true if all declared connections are established
	 */
	protected boolean allConnectionsEstablished() {
//		System.err.println("*******Assembly.AllConnectionsEst: ");
//		int numC = assemblyInfo.getConnections().getNumConnectionDecl();
//		for(int i=0; i<numC;i++) {
//			if(!connectionEstablished[i]) return false;
//		}
//		for (Iterator iter = connectionStatus.keySet().iterator(); iter.hasNext();) {
//			Info o = (Info) iter.next();
//			Object val = connectionStatus.get(o);
//			if(val==null)return false;
//		}

		// why not iterate over the values directly?
		for(Iterator iter = connectionStatus.values().iterator(); iter.hasNext();) {
			Object next = iter.next();
//			System.err.println("*******Assembly.AllConnectionsEst: "+next);
			if(next == null) return false;
		}
		return true;
	}
	
	
	/** Start the assembly. */
	public void start() throws IOException {
		// This should be in the constructor, if Java had an "inner" construct...
		//XXX createDebugService();
		super.start();
		doSelfTest();
		context.getScheduler().scheduleThread(resourceThread);
        context.getScheduler().scheduleThread(commandThread);
        context.getScheduler().scheduleThread(spThread);
		context.getScheduler().scheduleThread(infoThread);
		context.getScheduler().scheduleThread(logCmdThread);
		if(dp != null) {
			dp.start();
		}
	}
	
	/** Stop the assembly. */
	public void stop() throws IOException {
		//Remove references to avoid leaking threads
		Iterator it = remoteServiceProxyIterator();
		while(it.hasNext()) {
			ServiceProxy sp = (ServiceProxy) it.next();
			sp.removeInCommandReceiver(commandThread);
			sp.removeOutCommandReceiver(commandThread);
		}
		if (dp != null) {
			context.getDiscoveryManager().removeLocalServiceDescription(
					(PRDServiceFMDescription) dp.getServiceProxy().getInfo());
			dp.stop();
		}
//		((AssemblyDeviceContext) context).getResourceManager().putEvent(
//                new ResourceRequestEvent(assemblyInfo.getConnections(),
//                        Assembly.this, resourceThread, false));
		doRequestDisconnections();
		
		commandThread.terminate();
        spThread.terminate();
		logCmdThread.terminate();
		infoThread.terminate();
		resourceThread.terminate();
		super.stop();
	}
			
	/** Updates the assembly info. */
	void updateAssemblyInfo(PRDAssemblyVer info) throws IOException {
		this.assemblyVer = info;
	}

	public void doRequestDisconnection(ArrayList l) {
		for (Iterator iter = l.iterator(); iter.hasNext();) {
			PRDData element = (PRDData) iter.next();
			doRequestDisconnection(element);
		}
	}

	/**
	 * send connection requests to resource manager
	 * 
	 * @param l a list of Decl nodes
	 */
	public void doRequestConnection(ArrayList l) {
		for (Iterator iter = l.iterator(); iter.hasNext();) {
			PRDData element = (PRDData) iter.next();
			doRequestConnection(element);
		}
	}
	/**
	 * send cancel connection requests to resource manager
	 * 
	 * @param d a Decl node
	 */
	
	public void doRequestDisconnection(PRDData d) {
		DebugDispatcher.debug("Assembly.doRequestDisconnection:"+d, DebugDispatcher.CAT_ASSEMBLY);
		IResourceManager resourceManager = ((AssemblyDeviceContext) context).getResourceManager();
		resourceManager.putEvent(new ResourceRequestEvent(d, this, resourceThread, false));
	}

	public void doRequestConnection(PRDData d) {
		DebugDispatcher.debug("Assembly.doRequestConnection:"+d, DebugDispatcher.CAT_ASSEMBLY);
		IResourceManager resourceManager = ((AssemblyDeviceContext) context).getResourceManager();
		resourceManager.putEvent(new ResourceRequestEvent(d, this, resourceThread, true));
	}
	
	public void doRequestConnections() {
		doRequestConnection(assemblyVer.getConnections());
	}
	
	public void doRequestDisconnections() {
		doRequestDisconnection(assemblyVer.getConnections());
	}
	
	public void replaceServiceDecl(final ServiceDecl oldDecl, final ServiceDecl newDecl){
		doRequestDisconnection(assemblyVer.lookupConnections(new ServiceUse(oldDecl.getLocalName())));
		oldDecl.remove();
		assemblyVer.getServices().addServiceDecl(newDecl);
		DebugDispatcher.debug("replaceServiceDecl: new="+newDecl,DebugDispatcher.CAT_ASSEMBLY);
		doRequestConnection(assemblyVer.lookupConnections(new ServiceUse(newDecl.getLocalName())));
	}
	
	/** Handle a CommandEvent received from a ServiceDescription. */
	protected void commandReceived(CommandEvent event) {
	}
	
	private void createDebugService() {
		throw new Error("createDebugService broken by changes");
////		System.out.println("createDebugService");
//		ServiceProxy sp;
//		PRDServiceFMDescription sd = new PRDServiceFMDescription("Debug");
//        URN tmpUrn = getServiceInfo().getURN();
//		URNLeaf tmpLeaf = new URNLeaf("debug");
//		URN tmpUrn2 = tmpUrn.createSubUrn(
//		                tmpLeaf);
//		sd.setLocalSID(new LocalSID(new DeviceID(tmpUrn.getDeviceURN().toString()), tmpUrn2.getString()));
//		GroupInfo ctrl = new GroupInfo("Control");
//		CommandInfo toggle_logging = new CommandInfo(CMD_TOGGLE_LOG, Command.DIRECTION_IN);
//		CommandInfo logging_mode = new CommandInfo(CMD_LOG_MODE, Command.DIRECTION_OUT);
//		ParamInfo mode = new ParamInfo("mode", "text/plain");
//		logging_mode.addParamInfo(mode);
//		CommandInfo clear = new CommandInfo(CMD_CLEAR, Command.DIRECTION_IN);
//		CommandInfo run = new CommandInfo(CMD_RUN, Command.DIRECTION_IN);
//		CommandInfo pause = new CommandInfo(CMD_PAUSE, Command.DIRECTION_IN);
//		CommandInfo step = new CommandInfo(CMD_STEP, Command.DIRECTION_IN);
//		ctrl.addControlInfo(toggle_logging);
//		ctrl.addControlInfo(logging_mode);
//		ctrl.addControlInfo(clear);
//		ctrl.addControlInfo(pause);
//		ctrl.addControlInfo(run);
//		ctrl.addControlInfo(step);
//		sd.addControlInfo(ctrl);
//		CommandInfo status = new CommandInfo("status", Command.DIRECTION_OUT);
//		status.addParamInfo(new ParamInfo("status", "text/plain"));
//		sd.addControlInfo(status);
//		CommandInfo log = new CommandInfo(CMD_LOG_TEXT, Command.DIRECTION_OUT);
//		log.addParamInfo(new ParamInfo("log", "text/plain"));
//		sd.addControlInfo(log);
//		
//		sp = (ServiceProxy) sd.createInstance();
//		
//		logCmd = (Command) sp.findItem(CMD_LOG_TEXT);
//		logMode = (Command) sp.findItem(CMD_LOG_MODE);
//		statusCmd = (Command) sp.findItem("status");
//		
//		try {
//			dp = new AbstractService(context, sp.getProviderURN(), sp.getID(), sp) {
//				protected SelfTestProvider getSelfTestService(URN urn) {
//					return null;
//				}
//			};
////			System.out.println("Creating "+dp.getServiceProxy().getInfo());
////			System.out.println("urn="+dp.getServiceProxy().getProviderURN());
////			System.out.println("ID="+dp.getServiceProxy().getID());
////			addServiceDescription(sd);
//			sp.addInCommandReceiver(logCmdThread);
//			sp.addOutCommandReceiver(logCmdThread);
//			addSubservice(dp);
//		} catch (IOException e) {
//			e.printStackTrace();
//		}
//		
	}
	
	protected void logMsg(String msg) {
		if(enableLogging) {
			logStr = (logStr != null ? logStr+"\n" : "") +msg; 
			Param p = logCmd.getParam(0);
			byte [] val = logStr.getBytes();
			p.setData(val);
			logCmd.invoke();
		}
	}
	
	protected void logClear() {
		logStr = ""; 
		Param p = logCmd.getParam(0);
		byte [] val = logStr.getBytes();
		p.setData(val);
		logStr = null;
		logCmd.invoke();
	}
	
	private void logMode() {
		String logStr = enableLogging?"on":"off";
		System.out.println("logMode("+logStr+")");
		Param p = logMode.getParam(0);
		byte [] val = logStr.getBytes();
		p.setData(val);
		logMode.invoke();
	}
	
	
	private LinearSet activeConnections = new LinearSet();
	private LinearSet activeDecls = new LinearSet();
	private LinearMap connectionStatus = new LinearMap();
	/** Listening for responses from the resource manager
	 * 
	 * @author sven
	 *
	 */
	private class ResourceThread extends PalcomPriorityThread {
		public void run() {
			doRequestConnections();
			ConnectionDeclList connections = assemblyVer.getConnections();
			for(int i=0; i < connections.getNumConnectionDecl();i++) {
				connectionStatus.put(connections.getConnectionDecl(i), null);
			}
			doSelfTest();
			while(true) {
				try {
					ResourceResponseEvent e = (ResourceResponseEvent) waitEvent();
					Info i = e.getInfo();
					if(e.isAdded()) {
						if(i instanceof ConnectionInfo) {
							ConnectionInfo ci = (ConnectionInfo) i;
							if(! activeConnections.contains(ci)) {
								DebugDispatcher.debug("trying to connect: "+ci, DebugDispatcher.CAT_ASSEMBLY);
								System.out.println("***ResourceThread. try:\n"+
										ci.getProviderAddress().getDeviceAddress().getPRDDevice().getDeviceID()+
										"-->\n"+
										ci.getCustomerAddress().getDeviceAddress().getPRDDevice().getDeviceID()
										);
								context().getResourceAddressingManager().requestConnection(ci);
								activeConnections.add(ci);
								activeDecls.add(e.getDecl());
								connectionStatus.put(e.getDecl(), ci);
								//XXX listen for connections for the active connection
//								System.err.println("********** Assembly..... 1: "+ci.getCustomerAddress().getBrowserLabel());
								context().getResourceAddressingManager().listenForConnections(infoThread, ci.getCustomerAddress().getDeviceAddress());
//								System.err.println("********** Assembly..... 2: "+ci.getProviderAddress().getBrowserLabel());
								context().getResourceAddressingManager().listenForConnections(infoThread, ci.getProviderAddress().getDeviceAddress());
							} else {
								System.err.println("Assembly.ResourceThread ignoring active connection:\n"+ci.getProviderAddress()+"\n"+ci.getCustomerAddress());

							}
						}
					} else if(e.isChanged()) {
						DebugDispatcher.debug("Assembly.ResourceThread: change "+i, DebugDispatcher.CAT_ASSEMBLY);
						Info oi = e.getOldInfo();
						if(i instanceof ConnectionInfo && oi instanceof ConnectionInfo) {
						    ConnectionInfo ci = (ConnectionInfo) i;
						    ConnectionInfo oci = (ConnectionInfo) oi;
						    DebugDispatcher.debug("trying to (replace) connect: "+ci, DebugDispatcher.CAT_ASSEMBLY);
						    context().getResourceAddressingManager().requestDisconnection(oci);
                            context().getResourceAddressingManager().requestConnection(ci);
						    activeConnections.remove(oci);
						    activeConnections.add(ci);
							connectionStatus.put(e.getDecl(), ci);
//						    dumpSet(activeConnections);
							//XXX listen for connections for the active connection
							context().getResourceAddressingManager().listenForConnections(infoThread, ci.getCustomerAddress().getDeviceAddress());
							context().getResourceAddressingManager().listenForConnections(infoThread, ci.getProviderAddress().getDeviceAddress());
						    //XXX when to remove?
							//context.getResourceManager2().updatePRDConnectionReceiver(ci.getDeviceID(), infoThread);
						}
					} else if(e.isRemoved()) {						
						if(i instanceof ConnectionInfo) {
						    ConnectionInfo ci = (ConnectionInfo) i;
						    DebugDispatcher.debug("trying to disconnect: "+ci, DebugDispatcher.CAT_ASSEMBLY);
						    context().getResourceAddressingManager().requestDisconnection(ci);
						    activeConnections.remove(ci);
							activeDecls.remove(e.getDecl());
							connectionStatus.put(e.getDecl(), null);
							//XXX remove connection listener?
						}
//						DebugDispatcher.debug("Assembly.ResourceThread: remove "+i, DebugDispatcher.CAT_ASSEMBLY);						
					}
				}catch (Exception e) {e.printStackTrace();}
			    doStatusCmd();
			}
		}
	}
	private boolean oldConnectionStatus;
	
	private void doSelfTest() {
		invokeSelfTestIfChanged();
	}

	private boolean selfTestChanged;
	
	private void invokeSelfTestIfChanged() {
		boolean cs = allConnectionsEstablished();
		if((selfTestChanged || cs != oldConnectionStatus)) {
//			selfTest.doSelfTestCmd();
			if(cs) {
//				System.err.println("********** GREEN");
				setStatus(PRDData.FULLY_OPERATIONAL, "");
			}else {
//				System.err.println("********** YELLOW");
				setStatus(PRDData.PARTIALLY_OPERATIONAL, "Missing connections");
			}
			oldConnectionStatus = cs;
			selfTestChanged = false;
		}
	}

	private void doStatusCmd() {
//		System.err.println("Warning! Assembly.doStatusCmd broken by changes");
		String status="Connection status";
//		for(int i=0; i<assemblyInfo.getConnections().getNumConnectionDecl();i++) {
//			status=status+"\n"+assemblyInfo.getConnections().getConnectionDecl(i).getAssemblyLabel();
//		}
//		status=status+"\nestablished:";
//		for (Iterator iter = activeDecls.iterator(); iter.hasNext();) {
////			ConnectionInfo ci = (ConnectionInfo) iter.next();
////			status=status+"\n"+(ci.getProviderURN());
//			Decl d = (Decl)iter.next();
//			status=status+"\n"+d.getAssemblyLabel();
//		}
//		for (Iterator iter = connectionStatus.keySet().iterator(); iter.hasNext();) {
//			PRDData o = (PRDData) iter.next();
//			Object val = connectionStatus.get(o);
//			status=status+"\n"+(val!=null?"[UP]":"[DOWN]")+"\t"+o.getAssemblyLabel();
//			
//		}
//		Param p = statusCmd.getParam(0);
//		byte [] val = status.getBytes();
//		p.setData(val);
//		statusCmd.invoke();
		invokeSelfTestIfChanged();
	}
	
//	private void dumpSet(LinearSet activeConnections) {
//		for (Iterator iter = activeConnections.iterator(); iter.hasNext();) {
//			ConnectionInfo ci = (ConnectionInfo) iter.next();
//			System.out.println(ci.getProviderURN());
//		}
//
//	}
	
	private class ConnectionInfoThread extends PalcomPriorityThread {
		public void run() {
			while(true){
				Event e = waitEvent();
				//XXX put here just to get periodic updates...
				doStatusCmd();

				if(e instanceof AInfoEvent) {
					AInfoEvent ie = (AInfoEvent)e;
					
					if(ie.getInfo() instanceof ConnectionInfo) {
						ConnectionInfo ci = (ConnectionInfo) ie.getInfo();
						
						if(ie.infoShouldBeKept()) {
//							System.err.println("ConnectionInfoThread:"+ci);
//							dumpSet(activeConnections);
//							System.err.println("Comparing to "+assemblyInfo.getURN());
							if(ci.getCustomerAddress().equals(assemblyVer.getCustServiceAddr())) {
								if(activeConnections.contains(ci)) {
									
								} else {
								//remove unwanted connections
//								System.err.println("********UNWANTED********");
//XXX                                    context().getResourceAddressingManager().requestDisconnection(ci);
								}
							}
						} 
						else {
							System.err.println("Assembly.ConnectionInfoThread: connection down:\n"+ci.getProviderAddress()+"\n"+ci.getCustomerAddress());
//							activeConnections.remove(ci);
							if(activeConnections.contains(ci)) {
//								//try to re-establish wanted connections
////							    context.getRemoteConnectThread().putEvent(new InfoEvent(ci, true));
								context().getResourceAddressingManager().requestConnection(ci);
							} else {
								System.err.println("*** unknown connection:\n"+ ci.getProviderAddress()+"\n-->\n"+ci.getCustomerAddress());
							}
						}
					}
				}
			}
		}
	}
	
	/**
	 * A Thread that received CommandEvents from ServiceDescriptions and calls
	 * the commandReceived method.
	 */
	
	private class CommandThread extends PalcomPriorityThread {
		
		public void run() {
			pauseSem.give();
			while (true) {
				pauseSem.take();
				if(!singleStep) pauseSem.give();
				CommandEvent event = (CommandEvent) waitEvent();
				commandReceived(event);
			}
		}
	}
    
    private class ServiceProxyThread extends PalcomPriorityThread {
        public void run() {
            while (true) {
                ServiceProxyEvent event = (ServiceProxyEvent) waitEvent();
                event.getServiceProxy().addOutCommandReceiver(commandThread);
            }
        }
    }
	
	private boolean enableLogging;
	
	/* QnD hack for pausing/single-stepping an assembly */
	private Semaphore pauseSem = new Semaphore(context.getScheduler());
	private boolean singleStep = false;
	
	/** Command thread for the Log service */
	private class LogCommandThread extends PalcomPriorityThread {
		public void run() {
			while (true) {
				CommandEvent event = (CommandEvent) waitEvent();
				String cmd = event.getCommand().getID();
				if(cmd.equals(CMD_CLEAR)) {
					logStr = null;
					logClear();
					logMsg("Event log:");
				} else if (cmd.equals(CMD_TOGGLE_LOG)) {
					enableLogging = !enableLogging;
					logMode();
				} else if (cmd.equals(CMD_RUN)) {
					singleStep = false;
					pauseSem.give();
				} else if (cmd.equals(CMD_PAUSE)) {
					singleStep = true;
					pauseSem.take();
				} else if (cmd.equals(CMD_STEP)) {
					pauseSem.give();
				}
			}
		}
	}
//	/**
//	* A thread that keeps the connections up, while the assembly is active.
//	*/
//	private class ConnectionKeeperThread extends PalcomPriorityThread {
//	private int period;
//	
//	ConnectionKeeperThread(int period) {
//	this.period = period;
//	}
//	
//	public void run() {
//	enablePeriodicEvents(period);
//	boolean[] tryingToConnect = new boolean[assemblyInfo.getConnections().getNumConnectionDecl()];
//	for (int j = 0; j < assemblyInfo.getConnections().getNumConnectionDecl(); j++) {
//	tryingToConnect[j] = false;
//	}
//	
//	while (true) {
//	waitEvent();
//	// Try to activate (establish) in-active
//	// connections if the assembly is active
//	statusStr = "Connection Status:";
//	for (int j = 0; j < assemblyInfo.getConnections().getNumConnectionDecl(); j++) {
//	ConnectionDecl cd = assemblyInfo.getConnections().getConnectionDecl(j);
//	ConnectionInfo ci = cd.asInfo();
//	
//	// Look up the current connectionInfo
//	ConnectionInfo current = (ConnectionInfo) remoteInfoHandler
//	.getInfoThatIsAboutTheSameEntity(ci);
//	
//	if (current == null) {
//	connectionEstablished[j]=false;
//	statusStr = statusStr+"\n"+"[**] "+cd.getProvider()+" --> "+cd.getCustomer();
//	if(remoteInfoHandler.connectionIsPossible(ci)) {
//	tryingToConnect[j] = true;
//	try {
//	System.out.println("trying to connect: "+ci);
//	context.getRemoteConnectCustomer().establishConnection(ci);
//	} catch (IOException e) {
//	// Don't bother, just keep on
//	}
//	} else {
//	tryLoadingSoftwareComponent(assemblyInfo, cd.getProvider().getURN());
//	}
//	} else {
//	tryingToConnect[j] = false;
//	connectionEstablished[j]=true;
//	statusStr = statusStr+"\n"+"[OK] "+cd.getProvider()+" --> "+cd.getCustomer();	
//	}
//	}
//	Param p = statusCmd.getParam(0);
//	byte [] val = statusStr.getBytes();
//	p.setData(val);
//	statusCmd.invoke();
//	}
//	}
//	
//	/** tryLoadingSoftwareComponent 
//	* 
//	* Connect to the SoftwareComponentManager on a device with a missing service
//	* and attempt to load the service with the specified URN.
//	*/
//	
//	protected void tryLoadingSoftwareComponent(AssemblyInfo ai, String urn) {
//	// Lookup device declaration in assembly info
//	DeviceDecl dev = ai.lookupDevice(Info.urnToDeviceURN(urn));
//	if(dev != null) {
//	String swcmanurn = dev.getURN()+SoftwareComponentManager.URN_SUFFIX;
//	ConnectionInfo ci = new ConnectionInfo(swcmanurn, getServiceInfo().getURN());
//	ConnectionInfo current = (ConnectionInfo) remoteInfoHandler
//	.getInfoThatIsAboutTheSameEntity(ci);
//	// check for or establish connection to the SoftwareComponentManager
//	if (current == null) {
//	try {
//	context.getRemoteConnectCustomer().establishConnection(ci);
//	} catch (IOException e) {
//	e.printStackTrace();
//	}        			
//	}        		
//	// Send the load command
//	ServiceProxy sp = getServiceProxyForUrn(swcmanurn);
//	if (sp != null) {
//	synchronized(Coroutine.class) {
//	ControlItem tmp = sp.findItem("load");
//	if(tmp instanceof Command) {
//	Command command = (Command) tmp;
//	command.getParam(0).setData(urn.getBytes());
//	logMsg("[OUT] trying to load SWC "+urn);
//	command.invoke();
//	try {
//	context.getRemoteConnectCustomer().closeConnection(ci);
//	} catch (IOException e) {
//	e.printStackTrace();
//	}
//	} else {
//	System.out.println("tryLoading...:Not a command "+tmp);
//	}
//	}
//	} 
//	}
//	}
//	}

	public void setThisReferencesToDeviceID() {
		assemblyVer.setThisReferencesToDeviceID(context);
//XXX		assemblyVer.setThisReferencesToDeviceID(context.getDeviceInfo().getDeviceID());
	}
	
//	private class InfoThread extends PalcomPriorityThread {
//	public void run() {
//	while (true) {
//	InfoEvent event = (InfoEvent) waitEvent();
//	remoteInfoHandler.infoReceived(event);
//	String deviceURN = event.getInfo().getDeviceURN();
//	context.getDiscoveryManager().updateServiceListReceiver(deviceURN, this);
//	context.getDiscoveryManager().updateConnectionInfoReceiver(deviceURN, this);
//	}
//	}
//	}
//	
	
}
