/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.assembly.commandline;

import ist.palcom.assemblies.Assembly;
import ist.palcom.assemblies.AssemblyDeviceContext;
import ist.palcom.assemblies.AssemblyParser;
import ist.palcom.assemblies.IResourceManager;
import ist.palcom.assembly.manager.AssemblyManager;
import ist.palcom.assembly.manager.NewResourceManager;
import ist.palcom.device.AbstractUDPDevice;
import ist.palcom.device.Service;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.PRDAssemblyD;
import ist.palcom.resource.descriptor.PRDAssemblyVer;
import ist.palcom.services.AbstractService;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Iterator;

/**
 * Simple commandline-based assembler, running .ass scripts.
 * Takes commands from stdin as well as auto-loading any .ass file(s)  
 * given as parameter.
 * 
 * @author Henrik Gammelmark, geemark@daimi.au.dk
 */
public class CommandLineAssembler extends AbstractUDPDevice {
	private AssemblyDeviceContext adc;
	private AssemblyManager mgr;
	
	private String[] toLoad;
	
	public static void main(String[] args) throws Exception {
		CommandLineAssembler asm = new CommandLineAssembler(args);
		asm.run();
	}
	
	
	public CommandLineAssembler(String[] assemblies, DeviceID deviceID) throws Exception {
		super(deviceID);
		setName("CommandLineAssembler");
		
		//This device should be active, i.e. look for other devices by sending heartbeats requesting answers.
		setHeartBeatPeriod(2000);
		
		toLoad = assemblies;
		
		// Command input
		System.out.println("PalCom commandline assembler");
		System.out.println("Local DeviceID: " + getDeviceInfo().getDeviceID());
	}

	public CommandLineAssembler(String[] assemblies) throws Exception {
		super(new DeviceID("X:cli-assembler" + Math.random()));
		setName("CommandLineAssembler");
		
		//This device should be active, i.e. look for other devices by sending heartbeats requesting answers.
		setHeartBeatPeriod(2000);
		
		toLoad = assemblies;
		
		// Command input
		System.out.println("PalCom commandline assembler");
		System.out.println("Local DeviceID: " + getDeviceInfo().getDeviceID());
	}
	
    protected void initDevice() throws IOException {
    	// Create context and assign ressource manager
    	adc = new AssemblyDeviceContext(getDeviceContext());
    	
    	IResourceManager resourceService = new NewResourceManager(adc, IResourceManager.SERVICE_NAME);
		adc.addServiceAtDeviceLevel((AbstractService)resourceService);
		adc.setResourceManager(resourceService);
    	
		// Create and start assembly manager
		mgr = new AssemblyManager(adc, AssemblyManager.SERVICE_NAME);
    	adc.addServiceAtDeviceLevel(mgr);
    	mgr.start();
    	
    	// Framework init
        super.initDevice();
        
		// Load assemblies if any
		for (int i = 0; i < toLoad.length; i++)
			startAssembly(toLoad[i]);
		
		new Thread(new CLIThread()).start();
    }
    
    /**
     * Given an assembly .ass file, parse the XML into an AssemblyDescriptor.
     */
    private PRDAssemblyD readAssemblyFromFile(File file) throws IOException {
    	// Read raw XML from file
    	StringBuilder xml = new StringBuilder();
    	BufferedReader reader = new BufferedReader(new FileReader(file));
    	
    	while (true) {
    		String tmp = reader.readLine();
    		if (tmp == null)
    			break;
    		xml.append(tmp);
    	}
    	reader.close();
    	
    	// Parse as assembly
    	return (PRDAssemblyD) AssemblyParser.parseAssemblyXML(xml.toString());
    }
    
    /*
     * Given a filename for a .ass assembly description, load it and run it.
     */
    private void startAssembly(String asm) {
    	try {
    		File f = new File(asm);
	    	PRDAssemblyD asmDesc = readAssemblyFromFile(f);
	        PRDAssemblyVer asmVer = asmDesc.getPRDAssemblyVer(0);
	        mgr.loadAssembly(asmVer);
	        System.out.println("Loaded " + f);
	        
		} catch (Exception e) {
			System.out.println("Unable to load and start assembly " + asm + ": " + e);
		}

    }
    
    
    /**
     * Handles the input from the command line interface, and processes user commands.
     */
    private class CLIThread implements Runnable {
    	// Command loop
    	public void run() {
    		BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
    		
    		try {
	    		while (true) {
	    			System.out.print("> ");
		    		String cmd = reader.readLine();
		    		if (cmd == null)
		    			break;
		    				    		
		    		// Exit
		    		if (cmd.equals("exit")) {
		    			System.exit(0);
		    		
		    		// Load and start assembly
		    		} else if (cmd.startsWith("load ")) {
		    			String ass = cmd.substring(5);
		    			startAssembly(ass);			    			
		    			
		    		// Stop assembly
		    		} else if (cmd.startsWith("stop ")) {
		    			String asm = cmd.substring(5);
		    			
		    			Iterator it = adc.serviceIterator();
		    			boolean found = false;
		    			while (it.hasNext()) {
		    				Service s = (Service) it.next();
		    				if (s instanceof Assembly && s.getServiceInfo().getName().equalsIgnoreCase(asm)) {
		    					found = true;
		    					mgr.stopAssembly((Assembly) s);
		    				}
		    			}
		    			
		    			if (found)
		    				System.out.println("Stopped");
		    			else
		    				System.out.println("No assemblies named " + asm + " running");
		    			
		    			
		    		// Stop all assemblies
		    		} else if (cmd.equals("stop")) {
		    			mgr.removeAssemblies();
		    			System.out.println("All stopped");
		    			
		    		// List running assemblies
		    		} else if (cmd.equals("list")) {
		    			Iterator it = adc.serviceIterator();
		    			while (it.hasNext()) {
		    				Service s = (Service) it.next();
		    				if (s instanceof Assembly) {
		    					System.out.println(s.getServiceInfo().getName());
		    				}
		    			}
		    			
		    			
		    		// Usage
		    		} else {
			    		System.out.println("Valid commands are:");
			    		System.out.println(" list\t\tList loaded assemblies");
			    		System.out.println(" load <file>\tLoad .ass file and start the assembly");
			    		System.out.println(" stop <name>\tStop the assembly with the given name");
			    		System.out.println(" stop\t\tStop all assemblies");
			    		System.out.println(" exit\t\tExit the assembler");
		    		}
	
	    		}
    		} catch (IOException e) {
    			System.err.println("IO error when reading from stdin.");
    			e.printStackTrace();
    		}
    	}
    }
}
