/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.base;

import java.io.File;
import java.io.FilenameFilter;
import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;

public class System {
    private final long startTime = java.lang.System.currentTimeMillis();

    private static System instance = new System();

    /** Paths used by the newInstance method. */
    private ArrayList<String> lookupPaths;

    /** Singleton, private constructor. */
    private System() {
        lookupPaths = new ArrayList<String>();
    }

    public static System instance() {
        return instance;
    }

    /** Suspend entire service for the indicated duration. */
    public void sleep(int milliseconds) {
        int wakeUpTime = time() + milliseconds;
        int diff = milliseconds;
        try {
            while (diff > 0) {
                Thread.sleep(diff);
                diff = wakeUpTime - time();
            }
        } catch (InterruptedException e) {
            // Do nothing
        }
    }

    public int time() {
        return (int) (java.lang.System.currentTimeMillis() - startTime);
    }

    /**
     * Block for at most timeout milliseconds, or until data is available on at
     * least one of the first numberOfChannels Channels in channels. A negative
     * timeout will block indefinitely.
     * 
     * @param channels
     *            The Channels.
     * @param numberOfChannels
     *            The first numberOfChannels channels in channels are used.
     * @param results
     *            The result for each channel is put into this array (in the
     *            same order). A positive value means that data is available.
     * @param timeout
     *            The timeout.
     * 
     * @return The number of channels, possibly zero, that are now ready to read
     *         from.
     */
    public int select(Channel[] channels, int numberOfChannels, int[] results,
            long timeout) {
        long wakeUpTime = time() + timeout;
        int n = selectNow(channels, numberOfChannels, results);
        while (n == 0 && (timeout < 0 || time() < wakeUpTime)) {
            sleep(10);
            n = selectNow(channels, numberOfChannels, results);
        }
        return n;
    }

    /**
     * Check if data is available on at least one of the first numberOfChannels
     * Channels in channels, and return without blocking.
     * 
     * @param channels
     *            The Channels.
     * @param numberOfChannels
     *            The first numberOfChannels channels in channels are used.
     * @param results
     *            The result for each channel is put into this array (in the
     *            same order). A positive value means that data is available.
     * 
     * @return The number of channels, possibly zero, that are now ready to read
     *         from.
     */
    public int selectNow(Channel[] channels, int numberOfChannels, int[] results) {
        int n = 0;
        for (int i = 0; i < numberOfChannels; i++) {
            if (channels[i].dataAvailable()) {
                n++;
                results[i] = 1;
            } else {
                results[i] = 0;
            }
        }
        return n;
    }

    /**
     * Yield - this is a no-op in java
     */
    public void yield() {
    }

    /**
     * JVM creates its stack traces in a different way
     */
    public Object createStackTrace() {
      return null;
    }
    
    public Object getHmapRoot() {
    	return null;
    }

    private static final Object placeholder = new Object();

    public Object getExceptionPlaceholder() {
        return placeholder;
    }

    public void addExceptionMapping(String s, Object o) {
    }

    /**
     * Construct an object of a class, given by a class name, and return it. It
     * is platform dependent how classes are looked up. The lookup process can
     * be configured through calls to {@link #addLookupPath(String)} and
     * {@link #removeLookupPath(String)}.
     * 
     * @param className
     *            The class name
     * @param params
     *            Parameters to the constructor
     * @throws Exception
     *             if the object cannot be created.
     * 
     * @return The created object.
     */
    public Object newInstance(String className, Object[] params)
            throws Exception {
        synchronized (lookupPaths) {
            // Get a parent class loader (class loader that our class loader
            // will delegate to)
            ClassLoader parent;
            if (params == null || params.length == 0) {
                parent = System.class.getClassLoader(); // Default
            } else {
                parent = params[0].getClass().getClassLoader();
                // A more general solution here would be to go through the
                // classloaders for all constructors and try to find the one
                // with the best chance of constructing the object.
            }

            // Get classpath URLs (paths as dirs and JAR files inside dirs)
            ArrayList<URL> urlList = new ArrayList<URL>();
            for (int i = 0; i < lookupPaths.size(); i++) {
                try {
                    File path = new File(lookupPaths.get(i));
                    File[] jarFiles = path.listFiles(new FilenameFilter() {
                        public boolean accept(File dir, String name) {
                            return name.toLowerCase().endsWith(".jar");
                        }
                    });
                    urlList.add(path.toURL()); // Search the dir itself
                    for (int j = 0; j < jarFiles.length; j++) {
                        urlList.add(jarFiles[i].toURL());
                    }
                } catch (MalformedURLException e) {
                    // This should not happen
                    throw new RuntimeException(e);
                }
            }
            URL[] urls = new URL[urlList.size()];
            urls = urlList.toArray(urls);

            // Create class loader and load class
            ClassLoader classLoader = new URLClassLoader(urls, parent);
            Class clazz = classLoader.loadClass(className);

            // Find suitable constructor, if any
            Constructor[] constructors = clazz.getConstructors();
            Constructor constructor = null;
            int constructorIndex = 0;
            while (constructor == null
                    && constructorIndex < constructors.length) {
                Class[] paramTypes = constructors[constructorIndex]
                        .getParameterTypes();
                boolean paramsMatch = paramTypes.length == params.length;
                int paramIndex = 0;
                while (paramsMatch && paramIndex < paramTypes.length) {
                    if (!paramTypes[paramIndex]
                            .isAssignableFrom(params[paramIndex].getClass())) {
                        paramsMatch = false;
                    }
                    paramIndex++;
                }
                if (paramsMatch) {
                    constructor = constructors[constructorIndex];
                }
                constructorIndex++;
            }

            if (constructor == null) {
                throw new NoSuchMethodException(className);
            } else {
                // Create and return object
                return constructor.newInstance(params);
            }
        }
    }

    /**
     * Add a (file system) path where classes are looked up on calls to
     * {@link #newInstance(String, Object[])}. It is platform dependent how the
     * search is performed at the path.
     * 
     * @param path
     *            The path
     */
    public void addLookupPath(String path) {
        synchronized (lookupPaths) {
            lookupPaths.add(path);
        }
    }

    /** Remove a path added by {@link #addLookupPath(String)}. */
    public void removeLookupPath(String path) {
        synchronized (lookupPaths) {
            lookupPaths.remove(path);
        }
    }
}
