/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.base.networking;

/**
 * This class implements core.BluetoothDiscovery
 * 
 * @author Jacob Frolund Pedersen, frolund@daimi.au.dk
 */

import ist.palcom.base.Channel;

import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;

public class BluetoothDiscovery extends Channel {
	private Vector services = new Vector();

	private Vector devices = new Vector();

	DiscoveryAgent discoveryAgent;

	BluetoothDiscoveryThread discoveryThread;
	
	private boolean discoveryFinished = false;
	
	UUID[] uuidSet;
	
	private int timeToWait = 0;

	public static UUID PALCOM_SERVICE_ID = new UUID(
			"88888888888888888888888888888810", false);
	
	// Speed up discovery, by limiting service inquires to predefined devices
	private static final boolean ONLY_PREDEFINED_DEVICES = false;
	private static List<String> predefinedDevices = new LinkedList<String>();
	
	static {
		// UAA Avetana licenses
		predefinedDevices.add("00092d14020f");
		predefinedDevices.add("000d93123f1f");
		predefinedDevices.add("00112465960d");
		predefinedDevices.add("000c76d32c0c");
		predefinedDevices.add("000c76d3770a");
		predefinedDevices.add("000e9b4c1fd8");
		predefinedDevices.add("000c764837af");
		predefinedDevices.add("000c7648a24d");
		predefinedDevices.add("0010c6dbcea3");
		
		// Whitestein Avetana licenses
		predefinedDevices.add("000ea1309b6b");
		predefinedDevices.add("000ea1309b79");
		predefinedDevices.add("0020e04ca247");
	}
	
	/*
	 * This method is called by the system class to identify if data is ready.
	 * Data is ready when a new palcom service is discovered (non-Javadoc)
	 * 
	 * @see ist.palcom.base.Channel#dataAvailable()
	 */
	public boolean dataAvailable() {
		if(discoveryFinished){
			discoveryFinished = false;
			return true;
		} else {
			return false;
		}
	}

	public int getNumberOfServices(){
		return services.size();
	}
	
	public int getNumberOfDevices(){
		return devices.size();
	}
	
	public String getServiceAddress(){
		return (String) services.remove(0);
	}
	
	public long timeToNextDiscovery(){
		timeToWait = timeToWait * 2;
		if(timeToWait == 0){
			timeToWait = timeToWait + (new Random()).nextInt(4)+2;
		}
		return timeToWait;
	}
	
	public BluetoothDiscovery() throws BluetoothStateException{
		if (ONLY_PREDEFINED_DEVICES)
			System.out.println("Warning: BT discovery is set to ignore devices that are not hardcoded");
		
		uuidSet = new UUID[2];
		uuidSet[0] = new UUID(0x1101); // btspp services
		uuidSet[1] = PALCOM_SERVICE_ID;

		discoveryThread = new BluetoothDiscoveryThread();
		
		discoveryAgent = LocalDevice.getLocalDevice().getDiscoveryAgent();
	}

	public synchronized void searchDevices() throws BluetoothStateException {
		//System.err.println("searchDevices()");
		discoveryAgent.startInquiry(DiscoveryAgent.GIAC, discoveryThread);
	}

	public synchronized void searchServices() throws BluetoothStateException {
		//System.err.println("searchServices()");
		RemoteDevice device = null;

		if (devices.size() > 0) {
			device = (RemoteDevice) devices.remove(0);
			//System.err.println("  -- searching device " + device.getBluetoothAddress());
			discoveryAgent.searchServices(null, uuidSet, device, discoveryThread);
		}
	}
	
	/*
	 * We need a new class here, othervise the package ist.palcom.process will
	 * have to know about javax.**** and we dont want that.
	 */
	private class BluetoothDiscoveryThread implements DiscoveryListener {
		public void deviceDiscovered(RemoteDevice btDevice,
				DeviceClass btDeviceClass) {
			
			//System.err.println("BT device discovered: " + btDevice.getBluetoothAddress().toLowerCase() + " - in list: " + predefinedDevices.contains(btDevice.getBluetoothAddress().toLowerCase()));
			if ((!ONLY_PREDEFINED_DEVICES || predefinedDevices.contains(btDevice.getBluetoothAddress().toLowerCase())) && !devices.contains(btDevice)) {
				devices.addElement(btDevice);				
			}
		}

		public void servicesDiscovered(int transID,
				ServiceRecord[] serviceRecord) {
			services.addElement(serviceRecord[0].getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT,false));
		}

		public void serviceSearchCompleted(int transID, int respCode) {
			discoveryFinished = true;
		}

		public void inquiryCompleted(int discoveryResult) {
			switch (discoveryResult) {
			case DiscoveryListener.INQUIRY_COMPLETED:
				// the discovery of devices ended, with no errors
				break;
			case DiscoveryListener.INQUIRY_ERROR:
				// There were an error during the discovery process
				break;
			case DiscoveryListener.INQUIRY_TERMINATED:
				// The discovery process were terminated used
				// discoveryAgent.cancelInquiry()
				break;
			default:
				// This should never happen
				break;
			}
			discoveryFinished = true;
		}
	}
}
