/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.base.networking;

import ist.palcom.base.Channel;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.StringTokenizer;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.StreamConnectionNotifier;

/**
 * This class implements a JVM and Avetana version of core.BluetoothSocket.
 * 
 * Notes: Since the implementation is based on rfcomm, the api only allows one
 * connection to each host.
 */
public class BluetoothSocket extends Channel {
	private final static boolean DEBUG = false;
	
	private boolean justConnected = false;

	protected LocalDevice localDevice;

	ReceivingThread thread = null;

	BluetoothPacket packetWithReceivedData = null;

	ServerThread sThread = null;
	
	private IOException lastError = null;

	/*
	 * Open a bluetooth connection to a specific host, this function call is non
	 * blocking
	 */
	public BluetoothSocket(String BTURL) throws IOException {
		thread = new ReceivingThread(BTURL);
		thread.setDaemon(true);
		thread.start();
	}

	public BluetoothSocket() throws BluetoothStateException, IOException {
		sThread = new ServerThread();
		sThread.setDaemon(true);
		sThread.start();
	}

	public void send(BluetoothPacket p) throws IOException {
		if (thread != null) {
			thread.send(p);
		}
	}

	/**
	 * @throws IOException if no data is available, but only an error
	 */
	public boolean receive(BluetoothPacket packet) throws IOException {
		if (lastError != null) {
			IOException e = lastError;
			lastError = null;
			System.out.println("BTSocket: Disconnect, propagating error");
			throw e;
		}
		
		if (packetWithReceivedData != null) {
			int length = Math.min(packetWithReceivedData.getLength(), packet
					.getLength());
			packet.setReceiverHost(packetWithReceivedData.getReceiverHost());
			packet.setLength(length);
			java.lang.System.arraycopy(packetWithReceivedData.getData(), 0,
					packet.getData(), 0, length);

			packetWithReceivedData = null;
			synchronized (this) {
				notifyAll();
			}
			return true;
		} else {
			return false;
		}
	}

	public void close() {
		// Closing the ServiceThread
		try {
			if (sThread != null)
				sThread.notifier.close();
		} catch (IOException e) {
			java.lang.System.err.println("Shut down of Bluetooth server socket failed");
		}
		// Closing the ConnectionThread
		thread.disconnectDetected();
		// TODO: close connections (connectionThread)
	}

	/**
	 * After a call to System's select or selectNow with this BluetoothSocket,
	 * returns true if data is available, false otherwise. If no call to select
	 * or selctNow has been made, result is undefined.
	 */
	public synchronized boolean dataAvailable() {
		if (lastError != null) return true;		// Provoke the client to observe the error by calling receive()
		
		// To create an event when a connection is created and othervise only
		// when data is avaliable
		if (justConnected) {
			justConnected = false;
			return true;
		} else {
			return packetWithReceivedData != null;
		}
	}

	public void startReceiving(){
		if((null != thread) && (!thread.isAlive())){
			thread.start();
		}
	}
	
	public boolean isConnected() {
		return (thread != null) && thread.connected;
	}

	public String getURL() {
		return thread.getURL();
	}
	
	public String getDeviceAddress() {
		if (thread.getURL().indexOf("//") > -1) {		// String.contains - java 1.3.1 compliant
			String[] splitUrl = stringSplitJava131(thread.getURL(), ":");
			String[] tmp = stringSplitJava131(splitUrl[1], "/");
			return tmp[0];
		} else {
			return thread.getURL();
		}
	}
	
	/**
	 * Used instead of String.split(s), to be java 1.3.1 compliant.
	 */
	private static String[] stringSplitJava131(String s, String delim) {
		StringTokenizer tk = new StringTokenizer(s, delim);
		String[] tmp = new String[tk.countTokens()];
		for (int i = 0; i < tmp.length; i++) {
			tmp[i] = tk.nextToken();
		}
		return tmp;
	}

	/**
	 * @return The local host address (without port).
	 * @throws Exception
	 */
	public static java.lang.String getLocalAddress() {
		try {
			return LocalDevice.getLocalDevice().getBluetoothAddress();
		} catch (IOException e) {
			System.err.println("BluetoothSocket: Error getting the local device address");
			return "00:00:00:00:00:00";
		}
	}
	
	private synchronized void dataReceived(BluetoothPacket packet)
			throws InterruptedException {
		packetWithReceivedData = packet;
		notifyAll();
		while (packetWithReceivedData != null) {
			wait();
		}
	}

	private class ServerThread extends Thread {

		private StreamConnectionNotifier notifier;

		ServerThread() throws BluetoothStateException, IOException {
			// Set name for debugging
			setName("BluetoothSocket "
					+ Integer.toHexString(BluetoothSocket.this.hashCode()));
			localDevice = LocalDevice.getLocalDevice();
			localDevice.setDiscoverable(DiscoveryAgent.GIAC);
			
			String url = "btspp://localhost:"
					+ BluetoothDiscovery.PALCOM_SERVICE_ID.toString()
					+ ";name=Palcom Server" + ";authorize=false";
			notifier = (StreamConnectionNotifier) Connector.open(url);
		}

		public void run() {
			try {
				thread = new ReceivingThread((StreamConnection) notifier
						.acceptAndOpen());
				justConnected = true;
			} catch (IOException e) {
				// TODO: Error propagation - how to handle this one?
				System.err.println("Failed to accept and open");
			}
			justConnected = true;
		}
	}

	private class ReceivingThread extends Thread {

		private DataInputStream inStream;

		private DataOutputStream outStream;

		private StreamConnection sConnection = null;

		private boolean connected = true;

		private String BTURL = "";

		ReceivingThread(String BTURL) throws IOException {
			this.BTURL = BTURL;
			startUP();
		}

		ReceivingThread(StreamConnection sConnection) throws IOException {
			if (sConnection != null) {
				this.sConnection = sConnection;
				this.BTURL = RemoteDevice.getRemoteDevice(sConnection).getBluetoothAddress();
				// Potential error thrown: Could not get the address of the remote device");
			} else {
				disconnectDetected();
				throw new IOException("ReceiverThread not connected");
			}
			startUP();
		}

		public String getURL() {
			return BTURL;
		}

		// Potential error thrown: Could not open in/out streams in ConnectionThread
		private void startUP() throws IOException {
			if (sConnection == null && !BTURL.equals("")) {
				this.sConnection = (StreamConnection) Connector.open(BTURL);
			}
			this.inStream = sConnection.openDataInputStream();
			this.outStream = sConnection.openDataOutputStream();
			System.out.println("Bluetooth connection established with " + BTURL);
		}
		
		// TODO: Check if the connection is still open.
		// Potential error thrown: Could not write packet to Bluetooth Stream
		public void send(BluetoothPacket packet) throws IOException {
			if (sConnection != null) {
				try {
					if(outStream != null){
						byte[] messageAsBytes = packet.toByteRepresentation();
						outStream.writeInt(messageAsBytes.length);
						outStream.write(messageAsBytes);
						if (DEBUG) 
							System.out.println(" << BT socket: Sent " + messageAsBytes.length + " bytes to " + BTURL);
					} else {
						disconnectDetected();
						throw new IOException("BluetoothSocket: outStream is null, but connection exists");
					}
					
				// Act locally and rethrow
				} catch (IOException e) {
					disconnectDetected();
					throw e;
				}
			} else {
				disconnectDetected();
				throw new IOException("BluetoothSocket: Connection to device closed");
			}
		}
		
		public void disconnectDetected() {
			if (!connected) return;
			lastError = new IOException("Disconnection detected");
			//notifyAll();		// "java.lang.IllegalMonitorStateException: current thread not owner"
			System.out.println("Bluetooth connection closed to " + BTURL);
			connected = false;
			close();
		}
		
		// TODO: Check if the connection is still open.
		public void run() {
			while (isConnected()) {
				if(!justConnected){
					try {
						if (inStream.available() > 0) {
							// reading the size of the message that is reseaved
							int size = inStream.readInt();
							byte[] messageAsBytes = new byte[size];
							inStream.readFully(messageAsBytes, 0, size);
							BluetoothPacket p = new BluetoothPacket();
							p.fromByteRepresentation(messageAsBytes);
							p.setSenderHost(RemoteDevice.getRemoteDevice(sConnection)
									.getBluetoothAddress());
							if (DEBUG)
								System.out.println(" >> BT socket: Received " + messageAsBytes.length + " bytes from " + BTURL);
							try {
								dataReceived(p);
							} catch (InterruptedException e) {
								throw new IOException("BluetoothSocket: Could not get the address of the remote device. This is needed to create a BluetoothPacket");
								// TODO: Propagate, don't catch locally
							}
						} else {
							try {
								sleep(10);
							} catch (InterruptedException e) {
								// Do nothing
							}
						}
					} catch (IOException e) {
						disconnectDetected();
						// Ignore as this should be handled in the while condition
					}
				}
			}
			try {
				inStream.close();
				outStream.close();
				sConnection.close();
			} catch (IOException e) {
				// just ignore as we are closing anyway
			}
		}
	}
}
