/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.base.networking;

import ist.palcom.base.Channel;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;
import java.util.LinkedList;

/**
 * This class implements a JVM version of ist.palcom.base.DatagramSocket.
 * Before, it was implemented using Java non-blocking I/0 (Java NIO), but that
 * is apparently not supported for multicast (see, e.g.,
 * http://directory.apache.org/subprojects/network/faq.html#multicast), so now
 * it uses "normal" Java I/O.
 */
public class DatagramSocket extends Channel {

    protected java.net.DatagramSocket socket;

    private ReceivingThread thread;

    private LinkedList<java.net.DatagramPacket> packetWithReceivedData
        = new LinkedList<java.net.DatagramPacket>();

    private IOException lastReceiverError;
    
    /** Receive buffer size. */
    private int receiveBufferSize = 0;
    
    /** Send buffer size. */
    private int sendBufferSize = 0;
    
    /**
     * This flag is set if only the loopback interface is available. In that
     * case, localhost broadcast is used instead of multicast.
     */
    protected static boolean onlyLoopback;

    static {
        // Compute the onlyLoopback flag by looking at available interfaces
        onlyLoopback = true;
        try {
            Enumeration<NetworkInterface> interfaces = NetworkInterface
                    .getNetworkInterfaces();
            while (onlyLoopback && interfaces.hasMoreElements()) {
                NetworkInterface iface = interfaces.nextElement();
                Enumeration<InetAddress> addresses = iface.getInetAddresses();
                while (onlyLoopback && addresses.hasMoreElements()) {
                    InetAddress address = addresses.nextElement();
                    if (address instanceof Inet4Address
                            && !address.isLoopbackAddress()) {
                        onlyLoopback = false;
                    }
                }
            }
        } catch (SocketException e) {
            e.printStackTrace();
        }
    }

    public DatagramSocket() throws IOException {
        socket = createSocket();
    }

    public DatagramSocket(int port) throws IOException {
        socket = createSocket(port);
    }

    protected java.net.DatagramSocket createSocket() throws IOException {
        return new java.net.DatagramSocket();
    }

    protected java.net.DatagramSocket createSocket(int port) throws IOException {
        return new java.net.DatagramSocket(port);
    }

    public int getPort() {
        return socket.getLocalPort();
    }

    public void send(DatagramPacket p) throws IOException {
        InetAddress address = InetAddress.getByName(p.getAddress());
        if (onlyLoopback && address.isMulticastAddress()) {
            // Send to localhost broadcast address instead
            address = InetAddress
                    .getByAddress(new byte[] { 127, -1, -1, -1 });
        }
        java.net.DatagramPacket javaPacket = new java.net.DatagramPacket(p
                .getData(), p.getLength(), address, p.getPort());
        socket.send(javaPacket);
    }
    
    /** @return A packet if a packet was received, nil otherwise. */
    public synchronized DatagramPacket receive(int maxSize) throws IOException {
    	// Check if receiverthread caused error - if so, propagate
    	if (lastReceiverError != null) {
    		IOException tmp = lastReceiverError;
    		lastReceiverError = null;
    		throw tmp;
    	}
    	
    	// Check if any message is waiting
        startThreadIfNotStarted();
        try {
            java.net.DatagramPacket p = packetWithReceivedData.removeFirst();
            return new DatagramPacket(p);
        } catch(java.util.NoSuchElementException e){
            return null;
        }
    }

    public void close() {
        synchronized (this) {
            if (thread != null) {
                thread.interrupt();
                thread = null;
            }
        }
        lastReceiverError = null;		// Reset error if any
        socket.close();
    }

    /**
     * This method should only be called by the System class.
     * 
     * @see ist.palcom.base.Channel#dataAvailable()
     */
    public synchronized boolean dataAvailable() {
		startThreadIfNotStarted();
        return !packetWithReceivedData.isEmpty();
    }

    /** Starts the ReceivingThread, if not done already. */
    private synchronized void startThreadIfNotStarted() {
        if (thread == null) {
            thread = new ReceivingThread();
            thread.setDaemon(true);
            thread.start();
        }
    }

    private synchronized void dataReceived(java.net.DatagramPacket packet)
            throws InterruptedException {
        packetWithReceivedData.add(packet);
    }

    /** @return The local host address (without port). */
    public static java.lang.String getLocalAddress() {
        // The following code will return the first IPv4 address found. (Java
        // 1.4.2 and above)
        try {
            Enumeration nis = NetworkInterface.getNetworkInterfaces();
            while (nis.hasMoreElements()) {
                NetworkInterface ni = (NetworkInterface) nis.nextElement();
                Enumeration ias = ni.getInetAddresses();
                while (ias.hasMoreElements()) {
                    InetAddress ia = (InetAddress) ias.nextElement();
                    if (ia instanceof Inet4Address && !ia.isLoopbackAddress()) {
                        return ia.getHostAddress();
                    }
                }
            }
        } catch (SocketException e) {
        }
        return "127.0.0.1";

        /*
         * Java 1.3.1 version of the above (for Pocket LOOX PDA's etc) May not
         * work well on all platforms (Linux etc)
         */
        /*
         * try { return InetAddress.getLocalHost().getHostAddress(); } catch
         * (UnknownHostException e) { return "127.0.0.1"; }
         */
    }

    public int getReceiveBufferSize() throws IOException {
    	if (receiveBufferSize == 0) {
    		receiveBufferSize = socket.getReceiveBufferSize();
    	}
    	return receiveBufferSize;
    }

    public int getSendBufferSize() throws IOException {
        if (sendBufferSize == 0) {
            sendBufferSize = socket.getSendBufferSize();
        }
        return sendBufferSize;
    }

    public void setReceiveBufferSize(int size) throws IOException {
        if (receiveBufferSize != size) {
        	socket.setReceiveBufferSize(size);
        	receiveBufferSize = socket.getReceiveBufferSize();
        }
    }

    public void setSendBufferSize(int size) throws IOException {
        if (sendBufferSize != size) {
        	socket.setSendBufferSize(size);
        	sendBufferSize = socket.getSendBufferSize();
        }
    }

    private class ReceivingThread extends Thread {
        ReceivingThread() {
            // Set name for debugging
            String socketClassName = socket.getClass().getName();
            socketClassName = socketClassName.substring(socketClassName
                    .lastIndexOf('.') + 1);
            setName(socketClassName + ", port " + socket.getLocalPort());
            setPriority(Thread.MAX_PRIORITY);
        }

        public void run() {
            try {
                socket.setSoTimeout(100);

                // The buffer is allocated here once and used
                // for every receive.
                byte[] buffer = new byte[65535];
                while (!isInterrupted()) {
                    java.net.DatagramPacket receivePacket = new java.net.DatagramPacket(
                            buffer, buffer.length);
                    try {
                        socket.receive(receivePacket);

                        // Allocate a short buffer of exactly the length needed
                        byte[] shortBuf = new byte[receivePacket.getLength()];
                        java.lang.System.arraycopy(buffer, 0, shortBuf, 0, shortBuf.length);
                        receivePacket.setData(shortBuf);
                        
                        dataReceived(receivePacket);
                    } catch (InterruptedIOException e) {
                        // Timeout, go on to check if we have been interrupted
                    }
                }
            } catch (InterruptedException e) {
                // Exit
            } catch (IOException e) {
                // Exit, the socket was probably closed
            	lastReceiverError = e;	// Rethrow when/if the client calls receive() on this socket
            }
        }
    }
}