/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.communication;

import ist.palcom.collections.LinearMap;
import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.palcomthreads.PeriodicEvent;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.routing.RoutingManager;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class is the base in the communication layer and handles communication 
 * with the media managers through a routing manager. 
 * It also handles reliable communication by using acknowledge and resend messages.
 * @author Thomas Forsstrm, thomasf@cs.lth.se
 */
public class CommunicationManager implements IHgraphNode{

	protected HgraphNode _node;

	private static final Selector BROADCAST_SELECTOR = new Selector("0");

	private static final long DEFAULT_PERIOD = 2000;

	/**
	 * The number of times a reliable message can time out before it is sent again.
	 */
	private static final int TIMEOUTS_BEFORE_RESENDING = 2;

	/**
	 * The time interval between sending a reliable message and the timeout for 
	 * receiving an ack message.
	 */
	private long timeoutDelay;

	/**
	 * The number of times a reliable message will be sent unanswered 
	 * before giving up.
	 */
	private static final int RESENDS_BEFORE_GIVING_UP = 10;

	/**
	 * The maximum allowed number of reliable sent messages waiting to 
	 * be acknowledged.
	 */
	private static final int MAX_UN_ACKED_SENT_MESSAGES = 100;

	private UnicastDistributor unicastDistributor;

	private BroadcastDistributor broadcastDistributor;

	private DistributorThread distributor;

	private ReliableMessageHandler reliableHandler;

	private RoutingManager routingManager;

	/** List of TimeoutNotices with timeouts for infos. */
	private LinkedList timeouts;

	public CommunicationManager(PalcomScheduler scheduler, RoutingManager routingManager) {
		this.routingManager = routingManager;
		distributor = new DistributorThread();
		scheduler.scheduleThread(distributor);
		routingManager.registerAsListener(distributor);
		unicastDistributor = new UnicastDistributor();
		broadcastDistributor = new BroadcastDistributor();

		reliableHandler = new ReliableMessageHandler();

		timeouts = new LinkedList();
		timeoutDelay = DEFAULT_PERIOD * 2;

		/******hgraph impl*******/
		_node = new HgraphNode("communication manager");
	}

	/** Used for test purposes only. */
	public void setDistributorThreadAsListener() {
		routingManager.registerAsListener(distributor);
	}

	/** Set the device cache information for the device. */
	public void setLocalDeviceCache(String localDeviceCache) {
		routingManager.setLocalDeviceCache(localDeviceCache);
	}

	/** Returns the device cache for this device. */
	public String getLocalDeviceCache() {
		return routingManager.getLocalDeviceCache();
	}

	/** Set the device ID for this device. */
	public void setLocalDeviceID(DeviceID localDeviceID) {
		routingManager.setLocalDeviceID(localDeviceID.toString());
	}

	/** Set the status flag for this device. */
	public void setLocalStatusFlag(byte statusFlag) {
		routingManager.setLocalStatusFlag(statusFlag);
	}

	/** Set the discovery version for this device. */
	public void setLocalDiscoveryVersion(String discoveryVersion) {
		routingManager.setLocalDiscoveryVersion(discoveryVersion);
	}

	/** Set the discovery selector for this device. */
	public void setDiscoverySelector(Selector selector) {
		routingManager.setDiscoverySelector(selector);
	}

	/** Send a message to a receiver at a given device and selector, with selector set.*/
	public boolean send(Message message, DeviceID receiverDeviceID, boolean reliable) {
		if(reliable) {
			reliableHandler.addReliableNodeAndSend(receiverDeviceID.toString(), message);
			return true;
		}
		return routingManager.send(message, receiverDeviceID.toString());
	}

	/**
	 * Send a single-shot message to a given device.
	 * Any additional header nodes except the one used for single-shot messages
	 * must be inserted before calling this method.
	 * @param message The message, possibly with extra header nodes already set.
	 * @param receiverDeviceID The ID of the device which to send the message.
	 * @param receiverSelector The selector of the listening service.
	 * @param senderSelector The selector of the sending service.
	 */
	public boolean sendSingleShot(Message message, DeviceID receiverDeviceID, 
			Selector receiverSelector, Selector senderSelector, boolean reliable) {
		MessageHeaderNode hNode = new MessageHeaderNode(
				MessageHeaderNode.NODE_FORMAT_SINGLE_SHOT, 
				new String[]{receiverSelector.toString(), 
						senderSelector.toString()});
		message.insertHeaderNodeFirst(hNode);
		if(reliable) {
			reliableHandler.addReliableNodeAndSend(receiverDeviceID.toString(), message);
			return true;
		}
		return routingManager.send(message, receiverDeviceID.toString());
	}

	/** Send out a broadcast message, 
	 * with required message header nodes inserted before. */
	public void sendBroadcast(Message message) {
		routingManager.sendBroadcast(message);
	}

	/** Send a message saying that the device will close down*/
	public void sendHeartAttack() {
		routingManager.sendHeartAttack();
	}

	/**
	 * Start receiving point-to-point messages.
	 * 
	 * @return Selector to be used for sending messages to this receiver.
	 */
	public Selector startReceiving2(PalcomThread receiver) {
		Selector selector = unicastDistributor.getNextSelector();
		unicastDistributor.setReceiver(selector, receiver);
		return selector;
	}

	/** Stop the routing manager and all MediaManagers 
	 * as well as the internal thread. */
	public void stop() {
		unicastDistributor.clear();
		broadcastDistributor.clear();
		reliableHandler.clear();

		distributor.terminate();
	}

	/** Remove the device with ID deviceID from the list of registered devices. 
	 * Also removes any waiting reliable messages which are to be sent to the lost device.
	 * */
	public void unregisterDevice(DeviceID deviceID) {
		reliableHandler.removeDeviceInformation(deviceID.toString());
		routingManager.unregisterDevice(deviceID.toString());
	}
	
	/**
	 * Send a heartbeat message to all devices saying that this device is still
	 * active.
	 */
	public void sendHeartBeat(boolean isHeartBeatAck) {
		routingManager.sendHeartBeat(isHeartBeatAck);
	}
	
	/**
	 * Register as a listener for heartBeatEvents. This method is used by test
	 * programs and the ResourceManager. Currently only one listener at a time
	 * is supported.
	 */
	public void registerForHeartBeats(PalcomThread resourceManager) {
		routingManager.registerForHeartBeats(resourceManager);
	}

	/**
	 * Subscribe to error events from all media managers.
	 * Unhandled errors within the MediaManagers will be send to the handler
	 * in the form of {@link MALErrorEvent} events.
	 * Notice: Currently, only one error handler at a time is supported
	 */
	public void addErrorHandler(PalcomThread handler) {
		routingManager.addErrorHandler(handler);
	}

	/** Start receiving broadcast messages on topic topic. */
	public void startReceivingBroadcast(String topic, PalcomThread receiver) {
		broadcastDistributor.addMessageReceiver(topic, receiver);
	}

	/** Start receiving broadcast messages from device with urn remoteURN and selector remoteSelector. */
	public void startReceivingBroadcast(DeviceID remoteDeviceID, Selector remoteSelector, PalcomThread receiver) {
		broadcastDistributor.addMessageReceiver(remoteDeviceID.toString() + "/" + remoteSelector.toString(), receiver);
	}

	/** Stop receiving messages at a selector. */
	public void stopReceiving(Selector selector) {
		unicastDistributor.removeReceiver(selector);
	}

	/** Stop receiving messages for a receiver. */
	public void stopReceiving(PalcomThread receiver) {
		unicastDistributor.removeReceiver(receiver);
	}

	/** Stop receiving messages on topic for a receiver. */
	public void stopReceivingBroadcast(String topic, PalcomThread receiver) {
		broadcastDistributor.removeMessageReceiver(receiver, topic);
	}

	/** Stop receiving broadcast messages on all topics. */
	public void stopReceivingBroadcast() {
		broadcastDistributor.clear();
	}

	/** @return The TimeoutNotice for reliable message, or null if not set. */
	private TimeoutNotice getTimeout(Message message) {
		TimeoutNotice notice = (TimeoutNotice) timeouts.getFirst();
		while (notice != null) {
			if(!notice.reliableMessage.equals(message)) {
				notice = (TimeoutNotice) notice.getNext();
			} else {
				break;
			}
		}
		return notice;
	}

	/** Set timeout and resend information for a reliable message. */
	private void setTimeout(Message message, long timeout, int timeoutsLeft, int resendsLeft) {
		TimeoutNotice notice = getTimeout(message);
		if (notice == null) {
			timeouts.insertLast(new TimeoutNotice(message, timeout, timeoutsLeft, resendsLeft));
		} else {
			notice.nextTimeout = timeout;
			notice.timeoutsLeft = timeoutsLeft;
			notice.resendsLeft = resendsLeft;
		}
	}

	/** Remove timeout and resend information for a reliable message. */
	private void removeTimeout(Message message) {
		TimeoutNotice notice = getTimeout(message);
		if (notice != null) {
			timeouts.remove(notice);
		}
	}

	/** Thread for receiving messages from the routing layer and 
	 * distributing them to the correct listener by the use of selectors. */
	public class DistributorThread extends PalcomPriorityThread {
		public void run() {
			long threadPeriod = DEFAULT_PERIOD / 20;
			enablePeriodicEvents((int) threadPeriod);
			while (true) {
				Event event = waitEvent();

				//Handle resending reliable messages with no ack message received
				if(event instanceof PeriodicEvent) {
					long now = ist.palcom.base.System.instance().time();

					TimeoutNotice timeoutNotice = (TimeoutNotice) timeouts.getFirst();
					while(timeoutNotice != null) {
						if (now > timeoutNotice.nextTimeout) {
							if (timeoutNotice.timeoutsLeft <= 1) {
								if(timeoutNotice.resendsLeft<1) {
									timeouts.remove(timeoutNotice);
								} else {
									Message message = timeoutNotice.reliableMessage;

									//Reliable node is already inserted, hence "false"
									
									send(message, new DeviceID(message.getReceiverDeviceID()), false);
									setTimeout(timeoutNotice.reliableMessage, timeoutNotice.nextTimeout
											+ DEFAULT_PERIOD,
											TIMEOUTS_BEFORE_RESENDING, timeoutNotice.resendsLeft-1);
								}
							} else {
								setTimeout(timeoutNotice.reliableMessage, timeoutNotice.nextTimeout
										+ timeoutDelay,
										timeoutNotice.timeoutsLeft - 1, timeoutNotice.resendsLeft);
							}
						}
						timeoutNotice = (TimeoutNotice) timeoutNotice.getNext();
					}
					//Handle messages
				} else if (event instanceof MessageEvent) {
					MessageEvent me = (MessageEvent) event;
					Message message = me.getMessage();

					int offset = message.getHeaderNodesLength();
					byte[] messageArray = message.getRawData();

					//Are there more info in the message besides the already parsed header nodes?
					if(offset < message.getTotalLength() && messageArray!=null 
							&& messageArray.length == message.getTotalLength()) { 
						Message parsedMessage = Message.parse(messageArray, 0, messageArray.length);
						parsedMessage.setSenderDeviceID(message.getSenderDeviceID());
						parsedMessage.setSenderSelector(message.getSenderSelector());
						message = parsedMessage;
					}
					LinkedList headerNodes = message.getHeaderNodes();
					MessageHeaderNode hNode = (MessageHeaderNode) headerNodes.getFirst();
					if(hNode != null) {
						byte formatIdentifier = (byte) hNode.getFormatIdentifier();
						//Connection message
						if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_CONNECTION) {
							//Broadcasted message, handled different than the other.
							if(hNode.getFunctionIdentifier() == MessageHeaderNode.FUNCTION_BROADCASTED) {
								String[] data = hNode.getData();
								if(data.length==1) {
									Selector selector = new Selector(data[0]);
									message.setSenderSelector(selector);
									broadcastDistributor.distribute(message, 
											message.getSenderDeviceID().toString() + "/" + 
											message.getSenderSelector().toString(), 
											CommunicationManager.this);
									message = null; //Message has been handled
								} else {
//									TODO Syntax error in broadcasted message
								}
							} else {
								String[] data = hNode.getData();
								extractData(message, data);
							}
							//Single-shot message
						} else if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_SINGLE_SHOT) {
							String[] data = hNode.getData();
							extractData(message, data);
							//Publish message
						} else if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_PUBLISH) {
							broadcastDistributor.distribute(message, hNode.getData()[0], CommunicationManager.this);
							message = null; //Message has been handled
						}
						//Check for reliable message and handle it
						hNode = (MessageHeaderNode)headerNodes.getFirst();
						while(hNode != null && message != null) {
							formatIdentifier = hNode.getFormatIdentifier();
							if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_RELIABLE) {
								reliableHandler.reliableMessageReceived(message);
								//Message gets distributed by reliableHandler.
								message = null;
							} else if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_ACKMESSAGE) {
								String[] data = hNode.getData();
								if(data.length==1) {
									int seq = Integer.parseInt(data[0]);
									reliableHandler.ackMessageReceived(message.getSenderDeviceID(), seq);
								} else {
									//TODO Syntax error in ack message
								}
							} else if(formatIdentifier == MessageHeaderNode.NODE_FORMAT_RESENDMESSAGE) {
								String[] data = hNode.getData();
								if(data.length==1) {
									int seq = Integer.parseInt(data[0]);
									reliableHandler.resendMessageReceived(message.getSenderDeviceID(), seq);
								} else {
//									TODO Syntax error in resend message
								}
							}
							hNode = (MessageHeaderNode)hNode.getNext();
						}
					}
					if(message != null) {
						byte[] buffer = new byte[message.getTotalLength()];
						message.write(buffer, 0);
//						System.out.println("CommunicationManager: "+new String(buffer));
						unicastDistributor.distribute(message,
								CommunicationManager.this);
					}
				}
			}
		}

		/** Get and set the receiver and sender selectors. */
		private void extractData(Message message, String[] data) {
			if(data.length==2) {
				Selector receiverSelector = new Selector(data[0]);
				Selector senderSelector = new Selector(data[1]);
				message.setReceiverSelector(receiverSelector);
				message.setSenderSelector(senderSelector);
			} else if(data.length==1) {
				Selector receiverSelector = new Selector(data[0]);
				message.setReceiverSelector(receiverSelector);
			} else {
				//TODO Syntax error in data part of the message header node
			}
		}
	}

	/** Class for distributing (unicast) messages by selectors. */
	protected class UnicastDistributor {
		/** Counter for assigning unique selectors. */
		private int nextSelector;

		private ArrayList receivers;

		private ArrayList receiverSelectors;

		UnicastDistributor() {
			nextSelector = Integer.parseInt(BROADCAST_SELECTOR.getString()) + 1;
			receivers = new ArrayList();
			receiverSelectors = new ArrayList();
		}

		void clear() {
			receiverSelectors.clear();
			receivers.clear();
		}

		Selector getNextSelector() {
			Selector selector = new Selector(Integer.toString(nextSelector));
			nextSelector++; // For next time
			return selector;
		}

		/**
		 * @return the index of the receiver with selector selector, or -1 if
		 *         not found.
		 */
		private int getIndexOfReceiver(Selector selector) {
			int index = 0;

			while (index < receivers.size()
					&& !receiverSelectors.get(index).equals(selector)) {
				index++;
			}
			if (index == receiverSelectors.size()) {
				return -1;
			} else {
				return index;
			}
		}

		/**
		 * @return the index of the receiver, or -1 if not found.
		 */
		private int getIndexOfReceiver(PalcomThread receiver) {
			int index = 0;

			while (index < receivers.size()
					&& receiverSelectors.get(index) != receiver) {
				index++;
			}
			if (index == receiverSelectors.size()) {
				return -1;
			} else {
				return index;
			}
		}

		PalcomThread getReceiver(Selector selector) {
			PalcomThread receiver = null;
			int index = getIndexOfReceiver(selector);
			if (index != -1) {
				receiver = (PalcomThread) receivers.get(index);
			}
			return receiver;
		}

		/**
		 * Sets the receiver for selector selector. If set before, the old
		 * listener is removed.
		 */
		void setReceiver(Selector selector, PalcomThread receiver) {
			int index = getIndexOfReceiver(selector);
			if (index == -1) {
				receiverSelectors.add(selector);
				receivers.add(receiver);
			} else {
				receivers.set(index, receiver);
			}
		}

		void removeReceiver(Selector selector) {
			int index = getIndexOfReceiver(selector);
			if (index != -1) {
				receivers.remove(index);
				receiverSelectors.remove(index);
			}
		}

		/** Remove all occurences of the receiver from the list. */
		void removeReceiver(PalcomThread receiver) {
			int index = getIndexOfReceiver(receiver);
			while (index != -1) {
				receivers.remove(index);
				receiverSelectors.remove(index);
				index = getIndexOfReceiver(receiver);
			}
		}

		void distribute(Message message, Object source) {
			Selector receiverSelector = message.getReceiverSelector();

			PalcomThread receiver = getReceiver(receiverSelector);
			if (receiver != null) {
				receiver.putEvent(new MessageEvent(message, source));
			}
		}
	}

	/** Class for distributing publish (broadcasted) messages. */
	protected class BroadcastDistributor {
		private LinearMap topics;

		BroadcastDistributor() {
			topics = new LinearMap();
		}

		void addMessageReceiver(String topic, PalcomThread receiver) {
			ArrayList receivers = getReceivers(topic);
			if(receivers != null) {
				receivers.add(receiver);
			} else {
				receivers = new ArrayList();
				receivers.add(receiver);
			}
			topics.put(topic, receivers);
		}

		void clear() {
			topics.clear();
		}

		ArrayList getReceivers(String topic) {
			ArrayList receivers = (ArrayList) topics.get(topic);
			return receivers;
		}

		void removeMessageReceiver(PalcomThread receiver, String topic) {
			ArrayList receivers = getReceivers(topic);
			if(receivers != null) {
				int index = 0;
				while (index < receivers.size()
						&& receivers.get(index) != receiver) {
					index++;
				}
				if (index < receivers.size()) {
					receivers.remove(index);
					if(receivers.isEmpty()) {
						topics.remove(topic);
					}
				}
			}
		}

		void distribute(Message message, String topic, Object source) {
			ArrayList receivers = getReceivers(topic);
			if(receivers != null) {
				for (int i = 0; i < receivers.size(); i++) {
					PalcomThread receiver = (PalcomThread) receivers.get(i);
					receiver.putEvent(new MessageEvent(message, source));
				}
			}
		}
	}

	/** Handles reliable messages */
	protected class ReliableMessageHandler {

		/** List with counters for sequence numbers, 
		 * last Ack:ed number, last sent and last received. */
		private ArrayList seqCounters;

		/** List with the DeviceIDs of the devices with current ongoing reliable communications. */
		private ArrayList deviceIDs;

		/** Buffer with sent messages waiting to be Ack:ed. */
		private ArrayList sentMessages;

		/** Buffer with messages waiting to be sent. */
		private ArrayList messagesToSend;

		/** Buffer for incoming messages that are out of order. */
		private ArrayList receivedMessages;

		ReliableMessageHandler() {
			sentMessages = new ArrayList(); 
			seqCounters = new ArrayList();
			deviceIDs = new ArrayList();
			messagesToSend = new ArrayList();
			receivedMessages = new ArrayList();
		}

		void clear() {
			sentMessages.clear();
			seqCounters.clear();
			deviceIDs.clear();
			messagesToSend.clear();
			receivedMessages.clear();
		}

		/** Called when a AckMessage is received. */
		void ackMessageReceived(String senderDevice, int seq) {
			int index = deviceIDs.indexOf(senderDevice);
			if(index>=0) {
				int lastAck = ((SequenceNumber)seqCounters.get(index)).lastAck;
				int counter = lastAck+1;
				if(counter == Integer.MAX_VALUE) {
					counter = 0;
				}
				if(lastAck >= seq && seq > lastAck - MAX_UN_ACKED_SENT_MESSAGES) {
//					System.out.println("Ignoring received Ack message with " +
//					"lower sequence number than the previously received.");
					return;
				}
				
				LinearMap messages = (LinearMap)sentMessages.get(index);
				
				while(counter <= seq || MAX_UN_ACKED_SENT_MESSAGES < (counter - seq)) {
					Message removedMessage = (Message) messages.remove(new Integer(counter));
					removeTimeout(removedMessage);
					counter++;
					if(counter == Integer.MAX_VALUE) {
						counter = 0;
					}
					sendQueuedMessage();
				}
				((SequenceNumber)seqCounters.get(index)).lastAck = seq;
				sentMessages.set(index, messages);
			} else {
				System.out.println("Received Ack message from device with no previously known reliable commmunication.");
			}
		}

		/**
		 * Sends a message from the queue with unsent messages.
		 */
		private void sendQueuedMessage() {
			//Send a queued message
			if(timeouts.getSize() < MAX_UN_ACKED_SENT_MESSAGES && 
					!messagesToSend.isEmpty()) {
				Message message = (Message)messagesToSend.remove(0);

				//Find the header node with function reliable
				MessageHeaderNode hNode = (MessageHeaderNode) 
				message.getHeaderNodes().getFirst();
				while (hNode != null
						&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_RELIABLE) {
					hNode = (MessageHeaderNode) hNode.getNext();
				}
				if (hNode != null) {
					int sequence = Integer.parseInt(hNode.getData()[0]);
					int i = deviceIDs.indexOf(message.getReceiverDeviceID());
					LinearMap messageMap = ((LinearMap)sentMessages.get(i));
					messageMap.put(new Integer(sequence), message);
					sentMessages.set(i, messageMap);
					setTimeout(message, ist.palcom.base.System.instance().time() + timeoutDelay, 
							TIMEOUTS_BEFORE_RESENDING, RESENDS_BEFORE_GIVING_UP);
					send(message, new DeviceID(message.getReceiverDeviceID()), false);
				}
			}
		}

		/** Called when a ResendMessage is received. */
		void resendMessageReceived(String senderDevice, int seq) {
//			System.out.println("Resend message received from: " + senderDevice.toString() + " seq:" + seq);
			int index = deviceIDs.indexOf(senderDevice);
			if(index>=0) {
				LinearMap messages = (LinearMap)sentMessages.get(index);
				if(messages.containsKey(new Integer(seq))) {
					Message message = (Message)messages.get(new Integer(seq));
					setTimeout(message, ist.palcom.base.System.instance().time() + timeoutDelay, 
							TIMEOUTS_BEFORE_RESENDING, RESENDS_BEFORE_GIVING_UP);
					send(message, new DeviceID(senderDevice), false);
				} else {
					System.out.println("Could not find message asked to resend: " + senderDevice.toString()+ " " + seq);
				}
			} else {
				System.out.println("Received resend message from device with no previously known reliable commmunication.");
			}
		}

		/** Called when sending a reliable message. */
		void addReliableNodeAndSend(String receiverDevice, Message message) {
			if(!deviceIDs.contains(receiverDevice)) {
				deviceIDs.add(receiverDevice);
				seqCounters.add(new SequenceNumber());
				sentMessages.add(new LinearMap());
				receivedMessages.add(new LinearMap());
			} 
			int index = deviceIDs.indexOf(receiverDevice);
			if(((SequenceNumber)seqCounters.get(index)).lastSent+1 == Integer.MAX_VALUE) {
				((SequenceNumber)seqCounters.get(index)).lastSent = 0;
			} else {
				((SequenceNumber)seqCounters.get(index)).lastSent++;
			}
			int seq = ((SequenceNumber)seqCounters.get(index)).lastSent;
			MessageHeaderNode reliableNode = new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_RELIABLE, new String[]{
							"" + seq});
			message.insertHeaderNodeLast(reliableNode);
			message.setReceiverDeviceID(receiverDevice);
			//Only send out the message if the current number of sent un-Ack:ed messages 
			//are lower than MAX_UN_ACKED_SENT_MESSAGES.
			if(timeouts.getSize() < MAX_UN_ACKED_SENT_MESSAGES) {
				((LinearMap)sentMessages.get(index)).put(new Integer(seq), message);
				setTimeout(message, ist.palcom.base.System.instance().time() + timeoutDelay, 
						TIMEOUTS_BEFORE_RESENDING, RESENDS_BEFORE_GIVING_UP);
				send(message, new DeviceID(message.getReceiverDeviceID()), false);
			} else {
				messagesToSend.add(message);
			}
		}

		/** Called when a reliable message is received. */
		void reliableMessageReceived(Message message) {
			LinkedList headerNodes;
			headerNodes = message.getHeaderNodes();
			MessageHeaderNode hNode = (MessageHeaderNode) headerNodes
			.getFirst();
			// Find header node with function reliable
			while (hNode != null
					&& hNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_RELIABLE) {
				hNode = (MessageHeaderNode) hNode.getNext();
			}
			if (hNode != null) {
				int seq = Integer.parseInt(hNode.getData()[0]);
				String senderDevice = message.getSenderDeviceID();
//				headerNodes.remove(hNode); //TODO Should the reliable node be removed here?
//				message = adjustMessageRawData(message);
//				message.setHeaderNodes(headerNodes);
				if(!deviceIDs.contains(senderDevice)) {
					deviceIDs.add(senderDevice);
					seqCounters.add(new SequenceNumber());
					sentMessages.add(new LinearMap());
					receivedMessages.add(new LinearMap());
				} 
				int index = deviceIDs.indexOf(senderDevice);
				int lastReceived = ((SequenceNumber)seqCounters.get(index)).lastReceived;
				if(seq == lastReceived+1 || (seq == 0 && lastReceived+1 == Integer.MAX_VALUE)) {
					Message response = new Message(
							new MessageHeaderNode(
									MessageHeaderNode.NODE_FORMAT_ACKMESSAGE, 
									new String[]{"" + seq}));
					((SequenceNumber)seqCounters.get(index)).lastReceived = seq;

					send(response, new DeviceID(senderDevice), false);

					LinearMap map = (LinearMap)receivedMessages.get(index);
					int counter = seq+1;
					if(counter == Integer.MAX_VALUE) {
						counter = 0;
					}
					map.remove(new Integer(seq));
					unicastDistributor.distribute(message, distributor);
					while(map.get(new Integer(counter)) != null) {
						unicastDistributor.distribute((Message)map.get(new Integer(counter)), distributor);
						map.remove(new Integer(counter));
						send( new Message(
								new MessageHeaderNode(
										MessageHeaderNode.NODE_FORMAT_ACKMESSAGE, 
										new String[]{"" + counter})), new DeviceID(senderDevice), false);
						((SequenceNumber)seqCounters.get(index)).lastReceived = counter;
						counter++;
						if(counter == Integer.MAX_VALUE) {
							counter = 0;
						}
					}
					receivedMessages.set(index, map);

//					TODO What to do in this case? Received a already ack:ed message.
				} else if(seq < (lastReceived+1) && (lastReceived+1)-seq < MAX_UN_ACKED_SENT_MESSAGES) {
					
					Message response = new Message(new MessageHeaderNode(
									MessageHeaderNode.NODE_FORMAT_ACKMESSAGE, 
									new String[]{"" + seq}));
					send(response, new DeviceID(senderDevice), false);
				} else {
					int counter = lastReceived + 1;
					if(counter == Integer.MAX_VALUE) { 
						counter = 0;
					}
					LinearMap map = (LinearMap) receivedMessages.get(index);
					map.put(new Integer(seq), message);
					receivedMessages.set(index, map);
					do {
						if(!map.containsKey(new Integer(counter))) {
							Message response = new Message(
									new MessageHeaderNode(
											MessageHeaderNode.NODE_FORMAT_RESENDMESSAGE, 
											new String[]{"" + counter}));
							send(response, new DeviceID(senderDevice), false);
						}
						counter++;
						if(counter == Integer.MAX_VALUE) { 
							counter = 0;
						}
					} while(counter < seq || (counter - seq) > MAX_UN_ACKED_SENT_MESSAGES);
				}
			}
		}

		/** Called if a device has shut down. */
		void removeDeviceInformation(String deviceID) {
			int index = deviceIDs.indexOf(deviceID);
			if(index >= 0) {
				sentMessages.remove(index);
				deviceIDs.remove(index);
				seqCounters.remove(index);
				receivedMessages.remove(index);
			}
			for(int i = 0; i < messagesToSend.size(); ++i) {
				Message message = (Message)messagesToSend.get(i);
				if(message.getReceiverDeviceID().equals(deviceID)) {
					messagesToSend.remove(i);
					--i;
				}
			}
		}

//		/**
//		* This method is required, in the current implementation, when a
//		* message header node as been removed. It creates a new message without
//		* the raw data array.
//		* 
//		* @param message
//		* @return
//		*/
//		private Message adjustMessageRawData(Message message) {
//		Message correctedMessage;
//		if (message.getType() != Message.TYPE_NONE) {
//		if (message.getType() == Message.TYPE_SINGLE) {
//		correctedMessage = new Message(message.getType(), message
//		.getData());
//		} else {
//		correctedMessage = new Message(message.getType());
//		for (int k = 0; k < message.getNumberOfParts(); ++k) {
//		correctedMessage.addPart(message.getPart(k));
//		}
//		}
//		} else {
//		correctedMessage = new Message();
//		}
//		return correctedMessage;
//		}

		/** Internal class for handling sequence numbers. */
		private class SequenceNumber {
			int lastAck;
			int lastSent;
			int lastReceived;

			SequenceNumber() {
				lastAck = 0;
				lastSent = 0;
				lastReceived = 0;
			}

		}
	}
	/**
	 * Keeps track of the next timeout for a sent reliable message, and the number of timeouts
	 * left for it as well as the number of times it should be resent.
	 */
	private class TimeoutNotice extends Link {
		Message reliableMessage;

		long nextTimeout;

		int timeoutsLeft;

		int resendsLeft;

		TimeoutNotice(Message reliableMessage, long nextTimeout, int timeoutsLeft, int resendsLeft) {
			this.reliableMessage = reliableMessage;
			this.nextTimeout = nextTimeout;
			this.timeoutsLeft = timeoutsLeft;
			this.resendsLeft = resendsLeft;
		}
	}

	/****************hgraph impl below********************/

	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}
}