/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.connections;

import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.communication.CommunicationManager;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.Selector;

import java.util.Iterator;

/**
 * A connection for the transport of messages. It reads messages, and passes
 * them to registered message receivers.
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class Connection implements IHgraphNode{

	private IHgraphNode _node;
	
	private CommunicationManager communicationManager;

    private Selector localSelector;
    
    private Selector remoteSelector;
    
    private DeviceID remoteDeviceID;

    /** True while the connection is open. */
    private boolean open;

    /** The list of registered message receivers. */
    private LinkedList messageReceivers;

    /** The list of registered disconnect receivers. */
    private LinkedList disconnectReceivers;

    protected Connection(CommunicationManager communicationManager, DeviceID remoteDeviceID, ConnectionsXMLFactory factory) {
		this.communicationManager = communicationManager;
		messageReceivers = new LinkedList();
		disconnectReceivers = new LinkedList();
		this.remoteDeviceID = remoteDeviceID;
		open = true;
	}

    /** @return The device ID of the remote party. */
    public DeviceID getRemoteDeviceID() {
        return remoteDeviceID;
    }

    protected void setLocalSelector(Selector localSelector) {
		this.localSelector = localSelector;
	}
	
	protected void setRemoteSelector(Selector remoteSelector) {
		this.remoteSelector = remoteSelector;
	}
	
	public Selector getLocalSelector() {
		return localSelector;
	}
	
	public Selector getRemoteSelector() {
		return remoteSelector;
	}

    /**
     * Sets the value of open.
     * 
     * @return The previous value of open.
     */
    private boolean setOpen(boolean value) {
        boolean previousValue = this.open;
        this.open = value;
        return previousValue;
    }

    public void close() {
        if (setOpen(false)) {
            sendDisconnectMessage();

//            if (localSelector != null) {
//            	communicationManager.stopReceiving(localSelector);
//            }

            fireDisconnected();
        }
    }
	
	public void sendMessage(Message message) {
		MessageHeaderNode hNode = new MessageHeaderNode(
				MessageHeaderNode.NODE_FORMAT_CONNECTION, 
				MessageHeaderNode.FUNCTION_MESSAGE_OVER_CONNECTION,
				new String[]{remoteSelector.toString()});
		message.insertHeaderNodeFirst(hNode);
		communicationManager.send(message, remoteDeviceID, false);
	}
	
	public void sendBroadcast(Message message) {
		MessageHeaderNode hNode = new MessageHeaderNode(
				MessageHeaderNode.NODE_FORMAT_CONNECTION, 
				MessageHeaderNode.FUNCTION_BROADCASTED,
				new String[]{localSelector.toString()});
		message.insertHeaderNodeFirst(hNode);
		communicationManager.sendBroadcast(message);
	}

    public void addMessageReceiver(PalcomThread receiver) {
        messageReceivers.insertLast(new MessageReceiverNotice(receiver));
    }

    public void removeMessageReceiver(PalcomThread receiver) {
        MessageReceiverNotice notice = (MessageReceiverNotice) messageReceivers
                .getFirst();
        while (notice != null && notice.receiver != receiver) {
            notice = (MessageReceiverNotice) notice.getNext();
        }
        if (notice != null) {
            messageReceivers.remove(notice);
        }
    }

    /**
     * Add a PalcomThread that will be notified when the connection is
     * disconnected. A ConnectionEvent with connected=false will be sent to the
     * receiver.
     */
    public void addDisconnectReceiver(PalcomThread listener) {
        disconnectReceivers.insertLast(new ConnectionListenerNotice(listener));
    }

    public void removeDisconnectReceiver(PalcomThread listener) {
        ConnectionListenerNotice notice = (ConnectionListenerNotice) disconnectReceivers
                .getFirst();
        while (notice != null && notice.listener != listener) {
            notice = (ConnectionListenerNotice) notice.getNext();
        }
        if (notice != null) {
            disconnectReceivers.remove(notice);
        }
    }

    /**
     * Notifies all listeners that have registered interest for notification on
     * messages.
     */
    protected void fireMessage(Message message) {
        MessageReceiverNotice receiverNotice = (MessageReceiverNotice) messageReceivers
                .getFirst();
        while (receiverNotice != null) {
            receiverNotice.receiver.putEvent(new MessageEvent(message, this));
            receiverNotice = (MessageReceiverNotice) receiverNotice.getNext();
        }
    }

    /**
     * Notifies all listeners that have registered interest for notification on
     * disconnects.
     */
    protected void fireDisconnected() {
        ConnectionListenerNotice notice = (ConnectionListenerNotice) disconnectReceivers
                .getFirst();
        while (notice != null) {
            notice.listener.putEvent(new ConnectionEvent(this, false));
            notice = (ConnectionListenerNotice) notice.getNext();
        }
    }

    void sendDisconnectMessage() {
    	Message disconnectMessage = new Message(new MessageHeaderNode(
        		MessageHeaderNode.NODE_FORMAT_CONNECTION, 
        		MessageHeaderNode.FUNCTION_CLOSE, new String[]{
        				remoteSelector.toString()}));
        communicationManager.send(disconnectMessage, remoteDeviceID, false);
    }

    /**
     * Called by the ConnectionsManager when a message for this connection has
     * been received.
     */
    void messageReceived(MessageEvent event) {
        fireMessage(event.getMessage());
    }

    private class MessageReceiverNotice extends Link {
        final PalcomThread receiver;

        MessageReceiverNotice(PalcomThread receiver) {
            this.receiver = receiver;
        }
    }

    private class ConnectionListenerNotice extends Link {
        final PalcomThread listener;

        ConnectionListenerNotice(PalcomThread listener) {
            this.listener = listener;
        }
    }
    
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}

}