/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.connections;

import ist.palcom.collections.LinearMap;
import ist.palcom.communication.CommunicationManager;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.Selector;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class ConnectionsManager implements IHgraphNode{

	protected HgraphNode _node;

    private PalcomScheduler scheduler;

    private CommunicationManager communicationManager;

    private ConnectionsXMLFactory factory;

    private ReceiverThread receiverThread;

    /** Map from local selectors to customer side connection receivers. */
    private LinearMap customerSideConnectionReceivers;

    /** Map from local selectors to provider side connection receivers. */
    private LinearMap providerSideConnectionReceivers;

    /** Map from remote PalCom URLs to Connection objects. */
    private LinearMap connections;
    private IHgraphNode connectionsNode = new HgraphNode("connections");
    
    public ConnectionsManager(PalcomScheduler scheduler,
            CommunicationManager communicationManager) {
        this.scheduler = scheduler;
        this.communicationManager = communicationManager;
        factory = new ConnectionsXMLFactory();
        customerSideConnectionReceivers = new LinearMap();
        providerSideConnectionReceivers = new LinearMap();
        connections = new LinearMap();
        
		/******hgraph impl*******/
		_node = new HgraphNode("connections manager");
    }

    private void startReceiverThreadIfNeeded() {
        if (receiverThread == null) {
            receiverThread = new ReceiverThread();
            scheduler.scheduleThread(receiverThread);
        }
    }

    /** Stop the manager. */
    public void stop() {
        if (receiverThread != null) {
            receiverThread.terminate();
        }
    }

    /**
     * Connect to a provider listening at a connection selector. When the
     * connection has been established, a ConnectionEvent with connected=true
     * will be sent to the connectionReceiver.
     * 
     * @param localDeviceID
     *            Device ID of the customer that connects.
     * @param remoteDeviceID
     *            Device ID of the listening provider.
     * @param connectionSelector
     *            The selector on which the provider listens.
     * @param connectionReceiver
     *            PalcomThread that will get notified with ConnectionEvents.
     * @return The local selector to be used for connecting.      
     */
    public Selector open(DeviceID remoteDeviceID, Selector connectionSelector,
            PalcomThread connectionReceiver) {
        startReceiverThreadIfNeeded();
        Selector localSelector = communicationManager.startReceiving2(receiverThread);
        Message connectMessage = new Message(new MessageHeaderNode(
        		MessageHeaderNode.NODE_FORMAT_CONNECTION, 
        		MessageHeaderNode.FUNCTION_OPEN, new String[]{
        				connectionSelector.toString(), localSelector.toString()}));
        communicationManager.send(connectMessage, remoteDeviceID, false);
        customerSideConnectionReceivers.put(localSelector, connectionReceiver);
        return localSelector;
    }

    /**
     * Start listening for connections from customers.
     * 
     * @param connectionReceiver
     *            The PalcomThread that will be notified when connections have
     *            been established, in a ConnectionEvent with connected=true.
     * 
     * @return The selector to be used for connecting.
     */
    public Selector startListening2(PalcomThread connectionReceiver) {
        startReceiverThreadIfNeeded();
        Selector selector = communicationManager.startReceiving2(receiverThread);
        providerSideConnectionReceivers.put(selector, connectionReceiver);
        return selector;
    }

    /** Stop listening at a specific selector. */
    public void stopListening(Selector selector) {
    	communicationManager.stopReceiving(selector);
        providerSideConnectionReceivers.remove(selector);
    }

    private class ReceiverThread extends PalcomPriorityThread {

        public void run() {
            while (true) {
                Event ev = waitEvent();
                MessageEvent event = (MessageEvent) ev;
                Message m = event.getMessage();
                MessageHeaderNode hNode = (MessageHeaderNode)m.getHeaderNodes().getFirst();
                byte functionType = (byte) hNode.getFunctionIdentifier();
                if (hNode.getFormatIdentifier() == MessageHeaderNode.NODE_FORMAT_CONNECTION) {
                	//Control message
                	if (functionType != MessageHeaderNode.FUNCTION_MESSAGE_OVER_CONNECTION) {
                		if (functionType == MessageHeaderNode.FUNCTION_OPEN) {
                			PalcomThread connectionReceiver = (PalcomThread)
                				providerSideConnectionReceivers.get(m
                                    .getReceiverSelector());
                            if (connectionReceiver != null) {
                                // We are on the provider side
                                DeviceID remoteDeviceID = new DeviceID(m.getSenderDeviceID());
                                Connection connection = new Connection(
                                		communicationManager, remoteDeviceID,
                                        factory);
                                Selector receiverLocalSelector = communicationManager
                                .startReceiving2(receiverThread);
                                connection.setLocalSelector(receiverLocalSelector);
                                connection.setRemoteSelector(m.getSenderSelector());
                                Message response = new Message(new MessageHeaderNode(
                                		MessageHeaderNode.NODE_FORMAT_CONNECTION, 
                                		MessageHeaderNode.FUNCTION_OPEN_REPLY, new String[]{
                                				m.getSenderSelector().toString(),
                                				receiverLocalSelector.toString()}));
                                communicationManager.send(response, remoteDeviceID, false);
                                connectionReceiver
                                        .putEvent(new ConnectionEvent(
                                                connection, true));
                                connections.put(remoteDeviceID + "/" + m.getSenderSelector().toString(), connection);
                            }
                		} else if (functionType == MessageHeaderNode.FUNCTION_OPEN_REPLY) {
                			PalcomThread connectionReceiver = (PalcomThread)
                            	customerSideConnectionReceivers.get(m
                                    .getReceiverSelector());
                            if (connectionReceiver != null) {
                                // This is the reply from the provider
                                DeviceID remoteDeviceID = new DeviceID(m.getSenderDeviceID());
                                Connection connection = new Connection(
                                		communicationManager, remoteDeviceID,
                                        factory);
                                connection.setLocalSelector(m.getReceiverSelector());
                                connection.setRemoteSelector(m.getSenderSelector());
                                connectionReceiver
                                        .putEvent(new ConnectionEvent(
                                                connection, true));
                                connections.put(remoteDeviceID + "/" + 
                                		m.getSenderSelector().toString(), connection);
                                customerSideConnectionReceivers.remove(m
                                        .getReceiverSelector());
                            }
                		} else if (functionType == MessageHeaderNode.FUNCTION_CLOSE) {
                			Selector localSelector = m.getReceiverSelector();
                			ArrayList connectionsList = connections.values();
                			int index = 0;
                			boolean found = false;
                			Connection connection = null;
                			while(index < connectionsList.size() && !found) {
                				connection = (Connection) connectionsList.get(index);
                				if(connection.getLocalSelector().equals(localSelector)) {
                					found = true;
                				}
                				++index;
                			}
                			if(found) {
                				if (connection != null) {
                					connection.close();
                					connections.remove(connection.getRemoteDeviceID().toString() + "/" + 
                							connection.getRemoteSelector().toString());
                				}
                			}
                			if (localSelector != null) {
                            	communicationManager.stopReceiving(localSelector);
                            }
                		} else if (functionType == MessageHeaderNode.FUNCTION_RE_OPEN) {
                			String remoteDeviceID = m.getSenderDeviceID();
                			Selector remoteSelector = m.getReceiverSelector();
                        	Connection connection = (Connection) connections
                             	.get(remoteDeviceID + "/" + remoteSelector.toString());
                            if (connection != null) {
                                connection.close();//TODO change to a re-open method?
                                connections.remove(remoteDeviceID + "/" + 
            							connection.getRemoteSelector().toString());
                            }
                		}
                	//Message over connection
                	} else {
                		Selector localSelector = m.getReceiverSelector();
            			ArrayList connectionsList = connections.values();
            			int index = 0;
            			boolean found = false;
            			Connection connection = null;
            			while(index < connectionsList.size() && !found) {
            				connection = (Connection) connectionsList.get(index);
            				if(connection.getLocalSelector().equals(localSelector)) {
            					found = true;
            				}
            				++index;
            			}
            			if(found) {
                            // Send the received message to the connection
                        	
//                        	Message parsedMessage = Message.parse(m.getRawData(),hNode.getLength(),m.getTotalLength());
//                          connection.messageReceived(new MessageEvent(parsedMessage, event.getSource()));
                        	connection.messageReceived(event);
                        } else { //The connection was not found, send re-open
                        	DeviceID remoteDeviceID = new DeviceID(m.getSenderDeviceID());
                        	Message response = new Message(new MessageHeaderNode(
                        			MessageHeaderNode.NODE_FORMAT_CONNECTION,
                        			MessageHeaderNode.FUNCTION_RE_OPEN, new String[]{
                        					localSelector.toString()}));
                        	communicationManager.send(response, remoteDeviceID, false);
                        }
                	}
                } else {
                	System.out.println("ConnectionManager received a non connection message.");
                	//TODO ConnectionManager received a non connection message?
                }
            }
        }
    }
    
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}

}
