/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.debugging;

import java.util.ArrayList;
import java.util.Iterator;

/** A class for dispatching debugging/tracing messages from anywhere to
 *  interested listeners, if any.
 *  
 *  The current motivation for this class is the desire to turn on/off
 *  debugging messages from within the Eclipse Browser.
 *
 *  The singleton object to be used is instantiated lazily, and accessed 
 *  through the static getInstance();
 *  
 *  For convenience, there is a static method, debug(String msg), which
 *  calls putMessage() on the singleton instance.
 *  
 *  Future functionality would include categories and priority levels for
 *  messages.
 *  
 * @author sven
 *
 */
public class DebugDispatcher {
		
	/* Categories for messages */
	public static final int CAT_COMMUNICATION = 1 <<0; // For the communication layer
	public static final int CAT_SYSTEM = 1 << 1;        // Basic system stuff
	public static final int CAT_ASSEMBLY = 1 << 2;		// Assembly manager 	& Assemblies
	public static final int CAT_SWCOMP = 1 << 4;		// Software Components
	public static final int CAT_RESOURCE = 1 << 5;		// Resource management
	public static final int CAT_SERVICE = 1 << 6;		// User services
	public static final int CAT_BROWSER = 1 << 7;		
	public static final int CAT_EDITOR = 1 << 8;
	public static final int CAT_UPDATE = 1<<9;			// Automatic update service
	
	private ArrayList receivers;
	private static DebugDispatcher instance;
	
	public DebugDispatcher() {
		receivers = new ArrayList();
	}
	
	public synchronized void registerDebugListener(DebugListener d){
		receivers.add(d);
	}
	
	public synchronized void removeDebugListener(DebugListener d) {
		receivers.remove(d);
	}
	
	public synchronized void putMessage(String msg) {
		for (Iterator iter = receivers.iterator(); iter.hasNext();) {
			DebugListener d = (DebugListener) iter.next();
			d.debug(msg);
		}
	}

	/**
	 * Returns the singleton instance.
	 */
	public synchronized static DebugDispatcher getDefault() {
		if(instance == null) {
			instance = new DebugDispatcher();
		}
		return instance;
	}
	
	public static void debug(String msg) {
		getDefault().putMessage(msg);
	}
	
	public static void debug(String msg, int category) {
		//TODO implement categories...
		getDefault().putMessage(msg);
	}
}
