/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.device;

import ist.palcom.mal.StopEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.PRDDevice;

import java.io.IOException;

/**
 * The base class of a PalCom device. This abstract class creates no specific
 * managers. Subclasses implement the
 * {@link #createDeviceContext(String, PalcomScheduler) createDeviceContext} method
 * for creating a DeviceContext with the right set of managers.
 * 
 * @author David Svensson Fors, david@cs.lth.se
 */
public abstract class AbstractDevice implements Runnable{

    protected PalcomScheduler scheduler;

    protected DeviceContext context;

    private DeviceLifeThread lifeThread;

    private EventChannel eventChannel;

    private EventChannelThread eventChannelThread;
    
    protected AbstractDevice(DeviceID deviceID) {
        scheduler = new PalcomScheduler();
        context = createDeviceContext(deviceID, scheduler);
        lifeThread = new DeviceLifeThread();
        eventChannel = new EventChannel();
        eventChannelThread = new EventChannelThread(eventChannel);
    }

    public DeviceContext getDeviceContext() {
        return context;
    }
    
    /**
     * Creates the {@link DeviceContext} of the device. This method has to be
     * implemented in a concrete subclass.
     * 
     * @param deviceID The device ID.
     * @param scheduler The scheduler.
     * @return The DeviceContext
     */
    protected abstract DeviceContext createDeviceContext(DeviceID deviceID,
            PalcomScheduler scheduler);

    public PRDDevice getDeviceInfo() {
        return context.getDeviceInfo();
    }
    
    /** Set the time interval between heartbeats. 
	 * This method has to be called in order for the device to send out heartbeats requesting answers,
	 * otherwise it will only send out heartbeats in reply to other devices.
	 */
    public void setHeartBeatPeriod(long period) {
    	context.setHeartBeatPeriod(period);
    }

    public void setName(String name) {
        context.setName(name);
    }

    public String getName() {
        return context.getName();
    }

    /**
     * Run the scheduler of the device until completion. This method is
     * blocking. May be called by a Java thread.
     */
    public final void run() {
        scheduler.scheduleThread(lifeThread);
        scheduler.scheduleThread(eventChannelThread);
        scheduler.run();
    }

    /**
     * Stop the device. May be called by Java threads. The actual cleanup is
     * done by a PalcomThread in the method stopDevice.
     */
    public void stop() {
        safePutEvent(new StopEvent(), lifeThread);
    }

    /**
     * Put an event in the mailbox of a PalcomThread in a way that does not
     * interfere with this device's PalcomScheduler.
     */
    public void safePutEvent(Event event, PalcomThread target) {
        eventChannel.safePutEvent(event, target);
    }

    /**
     * Initialize the device. Called from a PalcomThread. Here the actual setup
     * is done. Subclasses override this method, and call super.initDevice from
     * there.
     */
    protected void initDevice() throws IOException {
        context.getDiscoveryManager().start();
        context.getResourceManager2().start();
        context.getRemoteConnectManager().start();
    }

    /**
     * Stop the device. Called from a PalcomThread. Here the actual cleanup is
     * done. Subclasses override this method, and call super.stopDevice from
     * there.
     */
    protected void stopDevice() throws IOException {
        context.getDiscoveryManager().stop();
        context.getResourceManager2().stop();
        context.getPubSubManager().stop();
        context.getConnectionsManager().stop();
        context.getMediaManager().stop();
        context.getRemoteConnectManager().stop();
        context.getCommunicationManager().stop();
        context.getRoutingManager().stop();
        eventChannelThread.terminate();
    }

    /**
     * The PalcomThread that calls initDevice, waits for a StopEvent, and then
     * calls stopDevice.
     */
    private class DeviceLifeThread extends PalcomPriorityThread {
        public void run() {
            try {
                initDevice();
                waitEvent();
                stopDevice();
                eventChannel.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
}