/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.device;

import ist.palcom.collections.LinearMap;
import ist.palcom.communication.CommunicationManager;
import ist.palcom.connections.ConnectionsManager;
import ist.palcom.discovery.DiscoveryManager;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.manager.HgraphManager;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.MediaManager;
import ist.palcom.messaging.PointToPointManager;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.pubsub.PubSubManager;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.PRDSubList;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.manager.ResourceManager;
import ist.palcom.resourceAddressing.ResourceAddressingManager;
import ist.palcom.routing.BasicRoutingManager;
import ist.palcom.routing.RoutingManager;

import java.util.ArrayList;
import java.util.Iterator;

public abstract class AbstractDeviceContext implements DeviceContext {
	
	public static final boolean ENABLE_HGRAPH_DEBUG_OUTPUT = false;

    protected PRDDevice deviceInfo; //hg

    protected ArrayList services; //hg
    
    protected LinearMap serviceLists;

    protected PRDServiceList serviceList; //hg

    protected PalcomScheduler scheduler;
    
    protected CommunicationManager communicationManager; //hg

    protected PointToPointManager pointToPointManager; //hg

    protected ConnectionsManager connectionsManager; //hg

    protected PubSubManager pubSubManager; //hg
    
    protected RoutingManager routingManager; //hg
    
    protected DiscoveryManager discoveryManager; //hg

    protected ResourceManager resourceManager;
    
    protected ResourceAddressingManager addressingManager;

    protected RemoteConnectManager remoteConnectManager; //hg
    
    protected HgraphManager hgraphManager; //hg

    protected IDeviceIO deviceIO;
    
    private LinearMap instanceNumbers; //<ServiceID,Numbercounter>
    
    private long localSIDNbrs;

	/*****hgraph impl*****/
	protected HgraphNode _node;
	protected HgraphNode servicesNode;
	protected HgraphNode communicationNode;
	protected HgraphNode middlewareManagersNode;
	protected HgraphNode prdNode;
	/*********************/

    public AbstractDeviceContext(DeviceID deviceID, PalcomScheduler scheduler) {
        this.scheduler = scheduler;

		_node = new HgraphNode();
		prdNode = new HgraphNode("resourceDescriptor");
		_node.addNodeChild(prdNode, null);

		deviceInfo = new PRDDevice(deviceID);
		
		prdNode.addNodeChild(deviceInfo, null);
		
		instanceNumbers = new LinearMap();
		localSIDNbrs = 0;
		
		LocalSID localSID = new LocalSID(deviceID, null);
        serviceList = new PRDServiceList(localSID);
        
        serviceLists = new LinearMap();
        serviceLists.put(localSID, serviceList);
        
		prdNode.addNodeChild(serviceList, null);

        //TODO check if this is the way to go
        routingManager = new BasicRoutingManager(getMediaManager(), scheduler);
        communicationManager = new CommunicationManager(scheduler, routingManager);
        pointToPointManager = new PointToPointManager(communicationManager);
        connectionsManager = new ConnectionsManager(scheduler,communicationManager);
        pubSubManager = new PubSubManager(scheduler, communicationManager);
        discoveryManager = new DiscoveryManager(scheduler, pointToPointManager,communicationManager, deviceInfo);
        discoveryManager.updateLocalServiceList(serviceList);
        resourceManager = new ResourceManager(scheduler, pointToPointManager,communicationManager, deviceInfo);
        remoteConnectManager = new RemoteConnectManager(this);
        addressingManager = new ResourceAddressingManager(this);
        
        hgraphManager = new HgraphManager(scheduler);
        
        services = new ArrayList();
		servicesNode = new HgraphNode("services");
		communicationNode = new HgraphNode("communication");
		middlewareManagersNode = new HgraphNode("middleware-managers");
		
		_node.addNodeChild(servicesNode, null);
		_node.addNodeChild(communicationNode, null);
		_node.addNodeChild(middlewareManagersNode, null);
		
		_node.addNodeChild(hgraphManager, null);
		communicationNode.addNodeChild(routingManager, null);
		communicationNode.addNodeChild(communicationManager, null);
		communicationNode.addNodeChild(pointToPointManager, null);
		communicationNode.addNodeChild(connectionsManager, null);
		communicationNode.addNodeChild(pubSubManager, null);
		communicationNode.addNodeChild(discoveryManager, null);
		communicationNode.addNodeChild(remoteConnectManager, null);

//		middlewareManagersNode.addNodeChild(resourceManager);
        new HgraphPrintUtil(1000, scheduler, this);
		/***********************/
    }
    
    public PRDDevice getDeviceInfo() {
        return deviceInfo;
    }

    /** @return An iterator over the AbstractServices on the device. */
    public Iterator serviceIterator() {
        return services.iterator();
    }
    
    /** @return A, for this serviceID, unique instance number as a String object. */
    public String getNewInstanceNumber(ServiceID serviceID) {
    	if(instanceNumbers.get(serviceID) == null) {
    		instanceNumbers.put(serviceID, new Integer(1));
    	}
    	int counter = ((Integer)instanceNumbers.get(serviceID)).intValue();
    	instanceNumbers.put(serviceID, new Integer(counter+1));
    	return "" + counter;
    }
    
    /** @return A new unique string. */
    public String getNewLocalSIDString() {
    	localSIDNbrs++;
    	return "" + localSIDNbrs;
    }
    
    /** @return A new unique localSID */
    public LocalSID getNewLocalSID() {
    	return new LocalSID(getDeviceInfo().getDeviceID(), getNewLocalSIDString());
    }
    
    /** Set the time interval between heartbeats. 
	 * This method has to be called in order for the device to send out heartbeats requesting answers,
	 * otherwise it will only send out heartbeats in reply to other devices.
	 */
    public void setHeartBeatPeriod(long period) {
    	getMediaManager().setPeriod(period);
    }

    public void setName(String name) {
		if(setNodeName(name, getPath()) == HgraphNode.STATUS_OK){
			deviceInfo.setName(name);			
		}
    }

    public String getName() {
        return deviceInfo.getName();
    }

    public void setDeviceIO(IDeviceIO io) {
        deviceIO = io;
    }

    public IDeviceIO getDeviceIO() {
        return deviceIO;
    }

    /**
     * Adds a service to the device. Will add info about the service to the info
     * of the device. The DiscoveryManager will be notified if the service is
     * active.
     */
    public void addServiceAtDeviceLevel(Service service) {
        addServiceToList(service, new LocalSID(deviceInfo.getDeviceID(), null));
    }
    
    /**
     * Adds a service to a service list identified by listLocalSID. 
     * Will add info about the service to the info of the device. 
     * The DiscoveryManager will be notified if the service is active.
     * If there are no list with listLocalSID nothing happens.
     */
    public void addServiceToList(Service service, LocalSID listLocalSID) {
    	PRDServiceList list = (PRDServiceList) serviceLists.get(listLocalSID);
    	if(list != null) {
    		services.add(service);
    		list.addPRDServiceListItem(service.getServiceInfo());
    		discoveryManager.updateLocalServiceList(list);
    		
    		/******hgraph impl*******/
    		servicesNode.addNodeChild(service, getPath());
    		/************************/
    	}
    }
    
    /**
     * Adds subList to the service list idenfitied by listLocalSID. ArrayList services is a list
     * containing objects of Service. A PRDServiceList will be created pointing at subList and the 
     * services will be added to it.
     * If there are no list with listLocalSID nothing happens.
     */
    public void addListOfServicesToList(LocalSID listLocalSID, ArrayList newServices, PRDSubList subList) {
    	PRDServiceList list = (PRDServiceList) serviceLists.get(listLocalSID);
    	if(list != null) {
    		list.addPRDServiceListItem(subList);
    		PRDServiceList subServiceList = new PRDServiceList(subList.getListLocalSID());
    		for(int i = 0; i < newServices.size(); ++i) {
    			if(newServices.get(i) instanceof Service) {
    				Service service = (Service)newServices.get(i);
    				services.add(service);
    				subServiceList.addPRDServiceListItem(service.getServiceInfo());
    			}
    		}
    		serviceLists.put(subServiceList.getParentLocalSID(), subServiceList);
    		discoveryManager.updateLocalServiceList(list);
    		discoveryManager.updateLocalServiceList(subServiceList);
    	}
    }
    
    /**
     * Adds subList to the service list located at the highest level, direct under the device.
     * ArrayList services is a list containing objects of Service. A PRDServiceList will be created 
     * pointing at subList and the services will be added to it.
     */
    public void addListOfServicesAtDeviceLevel(ArrayList newServices, PRDSubList subList) {
    	addListOfServicesToList(new LocalSID(getDeviceInfo().getDeviceID(),null), newServices, subList);
    }

    /**
     * EXPERIMENTAL! Removes a service from the device, and notify the
     * DiscoveryManager.
     */
    public void removeService(Service service) {
        // XXX beware, duplicate code (of sorts)
        services.remove(service);
        
		/******hgraph impl*******/
		servicesNode.removeNodeChild(service.getNodeName(), getPath());
		/************************/
		if (discoveryManager.removeLocalServiceOrSubList(service.getServiceInfo())) {
            discoveryManager.updateLocalServiceList(serviceList);
            if (service.getServiceProxy() != null) {
                if (service.getServiceProxy() != null) {
                    discoveryManager
                            .removeLocalServiceDescription((PRDServiceFMDescription) service
                                    .getServiceProxy().getInfo());
                }
            }
        } else {
            System.out.println("***removeService: failed to remove" + service.getServiceInfo());
        }
    }
    
    public void removeSubList(PRDSubList subList) {
    	discoveryManager.removeLocalServiceOrSubList(subList);
    }

    public PalcomScheduler getScheduler() {
        return scheduler;
    }

    public abstract MediaManager getMediaManager();

    public PointToPointManager getPointToPointManager() {
        return pointToPointManager;
    }

    public ConnectionsManager getConnectionsManager() {
        return connectionsManager;
    }

    public PubSubManager getPubSubManager() {
        return pubSubManager;
    }

    public DiscoveryManager getDiscoveryManager() {
        return discoveryManager;
    }

    public ResourceManager getResourceManager2() {
        return resourceManager;
    }
    
    public ResourceAddressingManager getResourceAddressingManager() {
    	return addressingManager;
    }

    public RemoteConnectManager getRemoteConnectManager() {
        return remoteConnectManager;
    }
    
    public HgraphManager getHgraphManager(){
    	return hgraphManager;
    }
    
    public CommunicationManager getCommunicationManager() {
    	return communicationManager;
    }
    
    public RoutingManager getRoutingManager() {
		return routingManager;
	}
    
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}
	
	class HgraphPrintUtil {
		public HgraphPrintUtil(long timeMilis, PalcomScheduler scheduler, IHgraphNode hgraphNode){
			PrintThread thr = new PrintThread(hgraphNode);
			scheduler.scheduleThread(thr);
			thr.enablePeriodicEvents(timeMilis);
		}
		public class PrintThread extends PalcomPriorityThread{	
			private IHgraphNode printNode;
			
			public PrintThread(IHgraphNode node){
				this.printNode = node;
			}
			
			public void run(){
				while(true){
					waitEvent();
					if(ENABLE_HGRAPH_DEBUG_OUTPUT)
						System.out.println(printNode.toXMLRecursively());
				}
			}
		}
	}
}