/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.device;

import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class implements the sending of I/O events between 
 * the "hardware" and "Palcom" domains.
 * 
 * If polling is required, this class should be extended and
 * get{In|Out}Event and/or wait{In|Out}Event overridden. 
 * The implementations in this class throws an Error.
 * 
 * @author sven
 *
 */

public class DeviceIO implements IDeviceIO {
	private PalcomThread inEventReceiver;
	private ArrayList outEventListeners;

    private EventChannel eventChannel;
    
	public DeviceIO(PalcomScheduler scheduler) {
		outEventListeners = new ArrayList();
        eventChannel = new EventChannel();
        EventChannelThread thread = new EventChannelThread(eventChannel);
        scheduler.scheduleThread(thread);
    }

    public void close() {
        eventChannel.close();
        // This will make the EventChannelThread terminate
    }

    public void putInEvent(Event e) {
        if (inEventReceiver != null) {
            // Send the event to the receiver in a PalcomThread-safe way
            eventChannel.safePutEvent(e, inEventReceiver);
        }
    }
	
	public void putOutEvent(Event e) {
		for (Iterator iter = outEventListeners.iterator(); iter.hasNext();) {
			OutEventListener l = (OutEventListener) iter.next();
			l.outEventReceived(e);
		}
	}

	public void setInEventReceiver(PalcomThread t) {
		inEventReceiver = t;
	}

	public void removeInEventReceiver() {
		inEventReceiver = null;
	}
	
	public void addOutEventListener(OutEventListener l) {
		outEventListeners.add(l);
	}

	public void removeOutEventListener(OutEventListener l) {
		outEventListeners.remove(l);	
	}

	public Event getInEvent() {
		throw new Error("DeviceIO.getInEvent: polling not implemented");
	}

	public Event waitInEvent() {
		throw new Error("DeviceIO.waitInEvent: polling not implemented");
	}
	
	public Event getOutEvent() {
		throw new Error("DeviceIO.getOutEvent: polling not implemented");
	}

	public Event waitOutEvent() {
		throw new Error("DeviceIO.waitOutEvent: polling not implemented");
	}
}
