/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.device;

import ist.palcom.base.Channel;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomThread;

import java.util.ArrayList;

/**
 * EventChannel is a {@link ist.palcom.base.Channel} that can be used from Java
 * threads for sending PalcomEvents to PalcomThreads in a safe way (doing
 * operations such as putEvent directly on a PalcomThread from a Java thread is
 * not safe, because the PalComScheduler is not aware of the Java threads).
 * <p>
 * A {@link EventChannelThread} has to be scheduled for the EventChannel for the
 * event passing to work. Java threads may call the thread safe method
 * {@link #safePutEvent(Event, PalcomThread) safePutEvent} on the channel, and
 * events will be delivered to PalcomThreads, given that an EventChannelThread
 * runs.
 * <p>
 * When finished using the EventChannel, the method {@link #close() close} must
 * be called. Otherwise, running EventChannelThreads will not stop.
 */
public class EventChannel extends Channel {

    private ArrayList eventsAndTargets;

    private boolean closed;

    public EventChannel() {
        eventsAndTargets = new ArrayList();
        closed = false;
    }

    /**
     * Put an event in the mailbox of a PalcomThread in a way that does not
     * interfere with the PalcomScheduler.
     */
    public synchronized void safePutEvent(Event event, PalcomThread target) {
        eventsAndTargets.add(new EventAndTarget(event, target));
    }

    /**
     * Returns and removes the next (oldest) event and target that was put.
     * 
     * @return The EventAndTarget, or null if there are none (the channel was
     *         perhaps closed).
     */
    synchronized EventAndTarget getNextEventAndTarget() {
        if (closed || eventsAndTargets.isEmpty()) {
            return null;
        } else {
            return (EventAndTarget) eventsAndTargets.remove(0);
        }
    }

    /**
     * Close the channel. The EventChannelThread that handles the channel will
     * stop.
     */
    public synchronized void close() {
        closed = true;
        eventsAndTargets.clear();
    }

    synchronized boolean isClosed() {
        return closed;
    }

    /**
     * This method should only be called by the System class.
     * 
     * @see ist.palcom.base.Channel#dataAvailable()
     */
    public synchronized boolean dataAvailable() {
        return !eventsAndTargets.isEmpty() || closed;
    }

    /** A small class used for grouping an event and a target PalcomThread. */
    static class EventAndTarget {
        final Event event;

        final PalcomThread target;

        EventAndTarget(Event event, PalcomThread target) {
            this.event = event;
            this.target = target;
        }
    }
}