/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.device;

import ist.palcom.discovery.InfoEvent;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.xml.XMLException;
import ist.palcom.xml.XMLFactory;
import ist.palcom.xml.XMLRepresentable;

import java.io.IOException;
import java.util.Iterator;

public class RemoteConnectManager implements IHgraphNode{

    private DeviceContext context;
    protected HgraphNode _node;

    /** XMLFactory used for creating XMLRepresentables from received data. */
    private XMLFactory factory;

    private RemoteConnectThread thread;
    
    public RemoteConnectManager(DeviceContext context) {
        this.context = context;
        factory = new FullXMLFactory();
        thread = new RemoteConnectThread();
        _node = new HgraphNode("remoteconnect manager");
    }

    public void start() {
        context.getScheduler().scheduleThread(thread);
        Selector selector = context.getPointToPointManager().startReceiving2(
                thread);
        context.getDeviceInfo().setRemoteConnectSelector(selector);
    }

    public void stop() {
        context.getPointToPointManager().stopReceiving(thread);
        thread.terminate();
    }

    public void requestConnect(PRDConnection info) throws IOException {
        makeRequest(new RemoteConnect(info));
    }

    public void requestDisconnect(PRDConnection info) throws IOException {
        makeRequest(new RemoteDisconnect(info));
    }

    /**
     * This method exposes the PalcomThread that handles the event loop of the
     * manager, making it possible to communicate with the manager through
     * events.
     */
    public PalcomThread getThread() {
        return thread;
    }

    /**
     * Sends a request to the RemoteConnect of the customer in info. If the
     * customer cannot be found, nothing happens.
     * 
     * @throws IOException
     *             if there was an error in the RemoteConnect communication.
     */
    private void makeRequest(final Request request) throws IOException {
        PRDConnection connectionInfo = request.getConnectionInfo();
        final DeviceID deviceID = connectionInfo.getLocalSID2().getDeviceID();
        PalcomThread infoThread = new PalcomPriorityThread() {
            public void run() {
                while (true) {
                    InfoEvent event = (InfoEvent) waitEvent();
                    PRDDevice deviceInfo = (PRDDevice) event.getInfo();
                    if (deviceInfo.getDeviceID().equals(deviceID)) {
                        Selector remoteConnectSelector = deviceInfo
                                .getRemoteConnectSelector();
                        boolean sent = context.getPointToPointManager().send(
                                new Message(request.getXMLData()),
                                context.getDeviceInfo().getRemoteConnectSelector(),
                                deviceID, remoteConnectSelector, false);
                        terminate();
                    }
                }
            }
        };
        context.getScheduler().scheduleThread(infoThread);
        context.getResourceManager2().updatePRDDeviceReceiver(infoThread);
    }

    /**
     * Perform a RemoteConnect/RemoteDisconnect request on service. Searches the
     * subservice tree below service until a matching service is found.
     * 
     * @return True if the request was performed, false otherwise
     */
    private boolean serveRequest(Request request, Service service) {
        PRDConnection connectionInfo = request.getConnectionInfo();
        LocalSID customerLocalSID = connectionInfo.getLocalSID2();
        if (customerLocalSID.equals(service.getServiceInfo().getLocalSID())) {
            LocalSID providerLocalSID = connectionInfo.getLocalSID1();
            Service customer = service;
            try {
                if (request instanceof RemoteConnect) {
                    customer.connectTo(providerLocalSID);
                } else if (request instanceof RemoteDisconnect) {
                    customer.disconnectFrom(providerLocalSID);
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
            return true;
        }
        return false;
    }

    private class RemoteConnectThread extends PalcomPriorityThread {
        public void run() {
            while (true) {
                Event event = waitEvent();
                if (event instanceof MessageEvent) {
                    MessageEvent messageEvent = (MessageEvent) event;
                    if (messageEvent.getMessage().getType() == Message.TYPE_SINGLE) {
                        Message message =  messageEvent
                                .getMessage();
						try {
                            XMLRepresentable representable = factory
                                    .createFromXMLData(
                                            message.getData(), 0,
                                            message.getDataLength());
                            if (representable instanceof Request) {
                                Request request = (Request) representable;
                                request.getConnectionInfo().getLocalSID2().setDeviceID(
                                		context.getDeviceInfo().getDeviceID());

                                Iterator serviceIterator = context
                                        .serviceIterator();
                                boolean performed = false;
                                while (!performed && serviceIterator.hasNext()) {
                                    Service service = (Service) serviceIterator
                                            .next();
                                    performed = serveRequest(request, service);
                                }
                            }
                        } catch (XMLException e) {
                            e.printStackTrace();
                        }
                    }
                } else if (event instanceof InfoEvent) {
                    InfoEvent infoEvent = (InfoEvent) event;
                    PRDConnection ci = (PRDConnection) infoEvent.getInfo();
                    try {
                        if (infoEvent.infoShouldBeKept()) {
                            requestConnect(ci);
                        } else {
                            requestDisconnect(ci);
                        }
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
    }
    
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}
}
