/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import java.util.Iterator;

/**
 * Functionality for H-Graph nodes.
 * This class provides functionality for holding and maintaining a set of handlers, invokators and listeners.
 * It further facilitates functionality for passing Hgraph Events through the H-Graph
 */
public class FuncFunc {
	
	private IHgraphNode node;
	
	private Handler[] handlers;
	private Invokator[] invokators;
	private Listener[] listeners;
	private GenericInvokator genericInvokator;
	
	public FuncFunc(IHgraphNode node){
		this.node = node;
	}
	
	public void removeListener(String name, Path callerPath){
		if(listeners != null){
			for(int i = 0; i<listeners.length; i++){
				if(listeners[i].getName().equals(name)){
					removeListener(listeners[i], callerPath);
				}
			}
		}
	}
	public void removeHandler(String name, Path callerPath){
		if(handlers != null){
			for(int i = 0; i<handlers.length; i++){
				if(handlers[i].getName().equals(name)){
					removeHandler(handlers[i], callerPath);
				}
			}
		}		
	}
	public void removeInvokator(String name, Path callerPath){
		if(invokators != null){
			for(int i = 0; i<invokators.length; i++){
				if(invokators[i].getName().equals(name)){
					removeInvokator(invokators[i], callerPath);
				}
			}
		}
	}
	
	/**
	 * Adds an invokator to the node
	 * @param invokator
	 */
	public void addInvokator(Invokator invokator, Path callerPath){
		if(invokators == null){
			invokators = new Invokator[1];
			invokators[0] = invokator;
		}else{
			Invokator[] tmp = new Invokator[invokators.length + 1];
			System.arraycopy(invokators, 0, tmp, 0, invokators.length);
			tmp[invokators.length] = invokator;
			invokators = tmp;			
		}
		if(invokator.getName().equals("Generic Invokator")){
			genericInvokator = (GenericInvokator)invokator;
		}
		trigger(FuncHandler.CMD_ADDINVOKATOR, new Value("", invokator), callerPath);
	}
	
	/**
	 * Adds a listener to the node
	 * @see Listener
	 */
	public void addListener(Listener listener, Path callerPath){
		if(listeners == null){
			listeners = new Listener[1];
			listeners[0] = listener;
		}else{
			Listener[] tmp = new Listener[listeners.length + 1];
			System.arraycopy(listeners, 0, tmp, 0, listeners.length);
			tmp[listeners.length] = listener;
			listeners = tmp;
		}
		listener.setNode(node);
		trigger(FuncHandler.CMD_ADDLISTENER, new Value("listener", listener), callerPath);
	}
	
	/**
	 * Removes a listener from the node
	 * @see Listener
	 * @param listener
	 */
	public void removeListener(Listener listener, Path callerPath){
		if(listeners == null){
			return;
		}
		for(int i = 0; i<listeners.length; i++){
			if(listeners[i].equals(listener)){
				Listener[] tmp = new Listener[listeners.length - 1];
				System.arraycopy(listeners, 0, tmp, 0, i);
				System.arraycopy(listeners, i + 1, tmp, i, listeners.length - i - 1);
				listeners = tmp;
			}
		}
		trigger(FuncHandler.CMD_REMOVELISTENER, new Value("", listener), callerPath);
	}

	/**
	 * Removes an invokator from the node
	 * @see Invokator
	 * @param Invokator
	 */
	public void removeInvokator(Invokator invokator, Path callerPath){
		if(invokators == null){
			return;
		}
		for(int i = 0; i<invokators.length; i++){
			if(invokators[i].equals(invokator)){
				Invokator[] tmp = new Invokator[invokators.length - 1];
				System.arraycopy(invokators, 0, tmp, 0, i);
				System.arraycopy(invokators, i + 1, tmp, i, invokators.length - i - 1);
				invokators = tmp;
			}
		}
		if(invokator.getName().equals("Generic Invokator")){
			genericInvokator = null;
		}
		trigger(FuncHandler.CMD_REMOVEINVOKATOR, new Value("", invokator), callerPath);
	}

	public GenericInvokator getGenericInvokator(){
		return genericInvokator;
	}
	
	
	/**
	 * Invokes (a) handler(s) across the h-graph. It is the intension, that only locally predefined path dependencies are allowed to
	 * invoke. This is difficult to uphold, but is approximated by reqiring the Path parameter to be installed in the local node.
	 * The recipient path in the event parameter may be absolute or it may be relative.
	 * Relatve paths are designated using ".." and "." following standard convensions
	 * @param path A predefined path dependency
	 * @param event The event to invoke the graph with. It holds the recipient path, the command and the parameters
	 * @return an int indicating the status of the method
	 * @see HgraphEvent
	 */
	public int invoke(Invokator invokator, HgraphEvent event){
		if(invokators != null){
			for(int i = 0; i<invokators.length; i++){
				if(invokators[i].equals(invokator)){
					event.setFromPath(node.getPath());
					event.setToPath(relative2absolute(event.getToPath()));
					int common = 0;
					for(int j = 0; j<node.getPath().getLength(); j++){
						if((event.getToPath().getLength() > j) && node.getPath().getPathArray()[j].equals(event.getToPath().getPathArray()[j])){
							common++;
						}else{
							break;
						}
					}
					invokeHandler(event, common);
					return IHgraphNode.STATUS_OK;
				}
			}
		}
		return IHgraphNode.INVOKATOR_MISSING;
	}
	
	/**
	 * Invokes (a) handler(s) across the h-graph. It is the intension, that only locally predefined path dependencies are allowed to
	 * invoke. This is difficult to uphold, but is approximated by reqiring the Path parameter to be installed in the local node.
	 * The recipient path in the event parameter may be absolute or it may be relative.
	 * Relatve paths are designated using ".." and "." following standard convensions
	 * @param path A predefined path dependency
	 * @param event The event to invoke the graph with. It holds the recipient path, the command and the parameters
	 * @return an int indicating the status of the method
	 * @see HgraphEvent
	 */

	/**
	 * Recomputes the path from relative to absolute. If it is already absolute, it is simply returned itself
	 * @param relPath
	 * @return
	 */
	private Path relative2absolute(Path relPath){
		Path returnPath = relPath;
		
		int relative = 0;
		for(int j = 0; j<returnPath.getLength(); j++){
			if(returnPath.getPathArray()[j].equals("..")){
				relative++;
			}else{
				break;
			}
		}
		if(relative != 0){
			String[] tmpPath = new String[node.getPath().getLength() + returnPath.getLength() - 2 * relative];
			System.arraycopy(node.getPath().getPathArray(), 0, tmpPath, 0, node.getPath().getLength() - relative);
			System.arraycopy(returnPath.getPathArray(), relative, tmpPath, node.getPath().getLength() - relative, returnPath.getLength() - relative);
			returnPath = new Path(tmpPath);
		}else if(returnPath.getPathArray()[0].equals(".")){
			String[] tmpPath = new String[node.getPath().getLength() + returnPath.getLength() - 1];
			System.arraycopy(node.getPath().getPathArray(), 0, tmpPath, 0, node.getPath().getLength());
			System.arraycopy(returnPath.getPathArray(), 1, tmpPath, 0, returnPath.getLength() - 1);					
			returnPath = new Path(tmpPath);
		}
		
		return returnPath;
	}
	
	/**
	 * Triggers the listeners installed in the local node.
	 * They are able to call back through the hgraph 
	 */
	public void trigger(String command, Value params, Path callerPath){
		if(listeners != null){
			for(int i = 0; i<listeners.length; i++){
				listeners[i].trigger(command, params, callerPath);
			}
		}
	}
	
	/**
	 * Traverses h-graphs and invokes handlers in a recursive fashion.
	 * @param event
	 * @param common
	 */
	private void invokeHandler(HgraphEvent event, int common){
		if(event.getToPath().equals(node.getPath()) || (event.getToPath().getLength() == 1 && node.getNodeParent() == null)){
			if(handlers == null){
				return;
			}
			for(int i = 0; i<handlers.length; i++){
				handlers[i].handle(event);
			}
		}else{
			if((common < node.getPath().getLength()) && (node.getNodeParent() != null)){
				node.getNodeParent().getFuncFunc().invokeHandler(event, common);
			}else{
				Iterator iter = node.getChildIterator();
				while(iter.hasNext()){
					IHgraphNode tmp = (IHgraphNode)iter.next();
					if(event.getToPath().getPathArray()[node.getPath().getPathArray().length].equals(tmp.getNodeName())){
						tmp.getFuncFunc().invokeHandler(event, common + 1);							
					}
				}					
			}
		}
	}
	
	/**
	 * Adds a handler to the node.
	 * @see Handler
	 * @param handler
	 */
	public void addHandler(Handler handler, Path callerPath){
		if(handlers == null){
			handlers = new Handler[1];
			handlers[0] = handler;
		}else{
			Handler[] tmp = new Handler[handlers.length + 1];
			System.arraycopy(handlers, 0, tmp, 0, handlers.length);
			tmp[handlers.length] = handler;
			handlers = tmp;			
		}
		trigger(FuncHandler.CMD_ADDHANDLER, new Value("", handler), callerPath);
	}
	
	/**
	 * Removes a handler from the node
	 * @param handler the Handler to remove. 
	 */
	public void removeHandler(Handler handler, Path callerPath){
		if(handlers == null){
			return;
		}
		for(int i = 0; i<handlers.length; i++){
			if(handlers[i].equals(handler)){
				Handler[] tmp = new Handler[handlers.length - 1];
				System.arraycopy(handlers, 0, tmp, 0, i);
				System.arraycopy(handlers, i + 1, tmp, i, handlers.length - i - 1);
				handlers = tmp;
			}
		}
		trigger(FuncHandler.CMD_REMOVEHANDLER, new Value("", handler), callerPath);
	}
	
	public Handler[] getHandlers(){
		return handlers;
	}
	public Invokator[] getInvokators(){
		return invokators;
	}
	public Listener[] getListeners(){
		return listeners;
	}
}
