/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hgraph;

import java.util.Iterator;

import ist.palcom.base.HashMap;
import ist.palcom.base.StringBuilder;

/**
 * Value functionality for H-Graph nodes.
 * This class provides functionality for holding and maintaining a set of Values
 *
 */
public class ValueFunc {

	private IHgraphNode node;
	private HashMap values;
	
	public ValueFunc(IHgraphNode node){
		this.node = node;
		values = new HashMap();
	}
	
	/**
	 * Adds a value to the local list of inspectable values.
	 * Values must have unique names and the value will thus only be added, if its name is unique
	 * @return a status message
	 */
	public int addValue(Value value, Path callerPath){
		if(values.containsKey(value.getName())){
			return IHgraphNode.VALUE_ALREADY_EXISTS;
		}
		values.put(value.getName(), value);
		node.getFuncFunc().trigger(ValueHandler.CMD_ADDVALUE, value, callerPath);
		return IHgraphNode.STATUS_OK;
	}
	
	/**
	 * Removes a value from the local list of inspectable values based on its name
	 * @param name
	 * @return a status message
	 */
	public int removeValue(String name, Path callerPath){
		if(values == null){
			return IHgraphNode.STATUS_OK;
		}
		values.remove(name);
		node.getFuncFunc().trigger(ValueHandler.CMD_REMOVEVALUE, new Value("name", name), callerPath);
		return IHgraphNode.STATUS_OK;
	}
	
	/**
	 * Sets the value of a named Value
	 * @param name name of the Value to set the value for
	 * @param value the new value
	 * @return a status message
	 */
	public int setValue(String name, Object value, Path callerPath){
		if(values == null){
			addValue(new Value(name, value), callerPath);
		}
		values.put(name, new Value(name, value));
		node.getFuncFunc().trigger(ValueHandler.CMD_SETVALUE, new Value(name, value), callerPath);
		return IHgraphNode.STATUS_OK;
	}
	
	/**
	 * Gets the value Object of a named Value
	 * @param name the name of the Value to get
	 * @return the Value corresponding to the name. Null if none is found
	 */
	public Value getValue(String name){
		if(values == null){
			return null;
		}
		return (Value)values.get(name);
	}
	
	public Iterator valueIterator(){
		return values.iterator();
	}	
}
