/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hmap;

import java.util.Iterator;

import ist.palcom.base.StringBuilder;

public class HmapLeaf extends HmapNode {

	public static final String HMAPLEAF_COMMAND_GETVALUE = "getValue";
	public static final String HMAPLEAF_COMMAND_SETVALUE = "setValue";
	
	/**
	 * Entry value
	 */
	private Object val;

	private ReturnValueInvokator returnValueInvokator;
	
	public HmapLeaf() {
		super(HmapNode.HET_LEAF, null, "HmapLeaf", "1.0", "ist.palcom.hmap.HmapLeaf");
		addHandler(new ValueHandler());
		returnValueInvokator = new ReturnValueInvokator(this);
		addInvokator(nameInvokator);
	}

	/**
	 * 
	 * @param name Name of the entry
	 * @param val  Value contained by this entry
	 */
	public HmapLeaf(String name, Object val) {
		super(HmapNode.HET_LEAF, name, "HmapLeaf", "1.0", "ist.palcom.hmap.HmapLeaf");
		this.val = val;
		addHandler(new ValueHandler());
		returnValueInvokator = new ReturnValueInvokator(this);
		addInvokator(nameInvokator);
	}

	/**
	 * Sets the value of the leaf
	 * @param val The new value
	 */
	public void setValue(Object val) {
		this.val = val;
		invokeListeners(new HmapEvent(null, getPath(), HMAPLEAF_COMMAND_SETVALUE, new HmapLeaf("", val)));
	}
	
	public void setValue(HmapEvent event){
		HmapNode pars = event.getParameters(); 
		if((pars != null) && pars.isLeaf()){
			val = pars.asLeaf().getValue();
		}
		invokeListeners(event);
	}
	
	/**
	 * Gets the value of the leaf
	 * @return The value
	 */
	public Object getValue() {
		return val;
	}
	
	protected boolean isAncestorOf(HmapNode node){
		return false; //A leaf isn't ancestor of anything
	}

	protected void recalculatePath(HmapPath parentPath) {
		this.path = new HmapPath(parentPath, getKey());
	}

	public HmapNode resolve(HmapPath path, int type) {
		if(path.equals(this.path) && (type == this.getHmapEntryType())){
			return this;
		}
		return null;
	}
	
	public HmapNode resolveLocally(HmapKey key, int type){
		if(key.equals(getKey()) && (type == this.getHmapEntryType())){
			return this;
		}
		return null;
	}
	
	protected void invokeHandlers(HmapEvent event){
		if(path == null){
			recalculatePath();
		}
		if(event.getToPath().equals(path)){
			Iterator iter = handlerIterator();
			while(iter.hasNext()){
				Handler handler = (Handler)iter.next();
				handler.handle(event);
			}
		}else{
			int ctr = 0;
			int max = 0;
			if(path.getLength() > event.getToPath().getLength()){
				max = event.getToPath().getLength();
			}else{
				max = path.getLength();
			}
			while((ctr < max)){
				if(path.getKeyAt(ctr).equals(event.getToPath().getKeyAt(ctr))){
					ctr ++;
				}else{
					break;
				}
			}
			getParent().invokeHandlersUp(event, ctr);
		}
	}

	/**
	 * @return A stringified version of the leaf
	 */
	public String toString() {
		return toString(0);
	}

	/**
	 * @param indent Start the returned string with a number of whitespaces
	 * @return A stringified version of the leaf
	 */
	public String toString(int indent) {
		StringBuilder tmp = StringBuilder.instantiate();
		for (int i = 0; i < indent; i++){
			tmp.append(" ");
		}
		if(getKey().hasName()){
			tmp.append(getKey().getName().toString());			
		}else{
			tmp.append("Anonymous");
		}
		tmp.append(" : ");
		if(val != null){
			tmp.append(val.toString());
		}else{
			tmp.append("null");
		}
		tmp.append("\r\n");
		
		return tmp.toString();	
	}

	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node with all handlers, invokators and listernes also
	 */
	public String toStringAll(int indent) {
		StringBuilder tmp = StringBuilder.instantiate();
		for (int i = 0; i < indent; i++){
			tmp.append(" ");
		}
		
		tmp.append(getKey().toString());
		tmp.append(" : ");
		tmp.append(val.toString());
		tmp.append("\r\n");
		
		tmp.append(printHandlers(indent + 3));
		tmp.append(printInvokators(indent + 3));
		tmp.append(printListeners(indent + 3));
		
		return tmp.toString();	
	}
	
	public class ValueHandler extends AbstractHandler{
		
		private static final String name = "Leaf Value Handler";
		private static final String description = "Handles the getting and setting of the value of this leaf";
		
		public void handle(HmapEvent event){
			if(event.getCommand().equals(HMAPLEAF_COMMAND_GETVALUE)){
				returnValueInvokator.returnValue(event);
			}else if(event.getCommand().equals(HMAPLEAF_COMMAND_SETVALUE)){
				setValue(event);
			}
		}
		
		public String getName(){
			return name;
		}
		
		public String getDescription(){
			return description;
		}
	}
	
	private class ReturnValueInvokator extends AbstractInvokator{

		private static final String name = "Leaf Value Return Value Invokator";
		private static final String description = "Responsible for returning values through in the hmap.";
		
		public ReturnValueInvokator(HmapNode associatedNode){
			super(associatedNode);
		}
		
		public void returnValue(HmapEvent event) {
			HmapNode tmp = event.getParameters();
			if((tmp != null) && tmp.isLeaf()){
				if((tmp.asLeaf().getValue() != null) && (tmp.asLeaf().getValue() instanceof String)){
					invokation(this, event.getFromPath(), (String)tmp.asLeaf().getValue(), new HmapLeaf("", val));		
				}
			}
		}

		public String getDescription() {
			return description;
		}

		public String getName() {
			return name;
		}
	}
}
