/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.hmap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import ist.palcom.base.StringBuilder;

/**
 * HmapNode is the abstract superclass for all hmap directories and leaves
 * @author tonz
 *
 */
public abstract class HmapNode{
	
	public static final String HMAPNODE_COMMAND_GETNAME = "getName";
	public static final String HMAPNODE_COMMAND_SETNAME = "setName";

	/**
	 * The collection of handlers associated with this node (reactive functionality)
	 */
	private Collection handlers;
	
	/**
	 * The collection of invocators associated with this node (proactive functionality)
	 */
	private Collection invocators;
	
	/**
	 * The collection of listeners listening on events from this node
	 */
	private Collection listeners;
		
	/**
	 * The key for this node (name and type)
	 */
	private HmapKey key;
	
	/**
	 * Path of the node. Recalculated on moves
	 */
	protected HmapPath path;
	
	/**
	 * parent of the node
	 */
	protected HmapDir parent;
	
	public static final int HET_UNKNOWN = 0;
	public static final int HET_DIR = 1;
	public static final int HET_LEAF = 2;

	/**
	 * The specific type of this hmap entry
	 */
	private int hmapEntryType = HET_UNKNOWN;

	public NameInvokator nameInvokator;
	public ValueInvokator valueInvokator;
	public NodeInvokator nodeInvokator;
	
	private ReturnNameInvokator returnNameInvokator;
	
	/** 
	 * @param treeNodeType the type of the treenode
	 * @param name the name of the treenode
	 */
	protected HmapNode(int treeNodeType, String name, String typeName, String typeVersion, String typeFullyQualifiedName){
		this.key = new HmapKey(name, typeName, typeVersion, typeFullyQualifiedName);
		this.hmapEntryType = treeNodeType;

		addHandler(new NameHandler());
		nameInvokator = new NameInvokator(this);
		addInvokator(nameInvokator);
		valueInvokator = new ValueInvokator(this);
		addInvokator(valueInvokator);
		nodeInvokator = new NodeInvokator(this);
		addInvokator(nodeInvokator);
		returnNameInvokator = new ReturnNameInvokator(this);
		addInvokator(returnNameInvokator);		
	}
	
	public String getType() {
		return key.getTypeName();
	}

	public String getFullyQualifiedType() {
		return key.getTypeFullyQualifiedName();
	}

	public String getTypeVersion() {
		return key.getTypeVersion();
	}

	/**
	 * @param path the path to resolve
	 * @return The first node at the given path. null if no such node exists
	 */
	public HmapNode resolve(HmapPath path){
		return resolve(path, HET_UNKNOWN);
	}
	    		
	/**
	 * @return true if this node is a dir. False if not
	 */
	public boolean isDir(){
		return hmapEntryType == HET_DIR;
	}

	/**
	 * @return true if this node is a leaf. False if not
	 */
	public boolean isLeaf(){
		return hmapEntryType == HET_LEAF;
	}
	
	/**
	 * @return True if the node has a name, false if not
	 */
	public boolean hasName(){
		return key.hasName();
	}
	
	/**
	 * @return True if the node has a type, false if not
	 */
	public boolean hasType(){
		return key.hasTypeName();
	}
	
	/**
	 * @return this node as a dir - nul if it is not a dir
	 */
	public HmapDir asDir(){
		if(hmapEntryType == HET_DIR){
			return (HmapDir)this;
		}
		return null;
	}
	
	/**
	 * @return this node as a leaf - null if it is not a leaf
	 */
	public HmapLeaf asLeaf(){
		if(hmapEntryType == HET_LEAF){
			return (HmapLeaf)this;
		}
		return null;		
	}
	
	/**
	 * @return The path of this node (relative to its hmap root)
	 */
	public HmapPath getPath(){
		if(path == null){
			recalculatePath();
		}
		return path;
	}

	/**
	 * Sets the name of the node
	 * @param name new name
	 */
	public void setName(String name) {
		this.key.setName(name);
		recalculatePath();
		invokeListeners(new HmapEvent(null, getPath(), HMAPNODE_COMMAND_SETNAME, new HmapLeaf("", name)));
	}
	
	public void setName(HmapEvent event){
		HmapNode pars = event.getParameters(); 
		if((pars != null) && pars.isLeaf()){
			if(pars.asLeaf().getValue() instanceof String){
				this.key.setName((String)pars.asLeaf().getValue());
				recalculatePath();
				invokeListeners(event);
			}
		}
	}
	

	/**
	 * @return the name of the node
	 */
	public String getName() {
		return key.getName();
	}
		
	/**
	 * @return The key for this hmap node
	 */
	public HmapKey getKey(){
		return key;
	}

	/**
	 * Sets the parent of this node
	 * @param parent new parent
	 */
	public void setParent(HmapDir parent){
		if(this.parent != null){
			this.parent.removeNode(this);
		}
		this.parent = parent;
		recalculatePath(parent.getPath());
	}
	
	/**
	 * @return the parent dir of this node (which is null if it has none)
	 */
	public HmapDir getParent(){
		return parent;
	}
		
	/**
	 * Sets up a new path with this node as root
	 */
	protected void recalculatePath() {
		if(this.path == null){
			path = new HmapPath(this.key);
		}
	}
	
	/**
	 * Invokes a handler somewhere in the hmap
	 * --:: CAN ONLY BE CALLED FROM INVOCATORS ASSOCIATED WITH THIS NODE!!! ::--
	 * @param invocator The invocator caller
	 * @param path The path on which to invoka the handler
	 * @param command The command to invoke the handler with
	 * @param parameters The parameters with which the handler is invoked
	 */
	public void invokation(Invokator invocator, HmapPath path, String command, HmapNode parameters){
		if(invocators != null){
			Iterator iter = invocators.iterator();
			while(iter.hasNext()){
				if(((Invokator)iter.next()).equals(invocator)){
					invokeHandlers(new HmapEvent(getPath(), path, command, parameters));					
					break;
				}
			}
		}		
	}
	
	/**
	 * Invokes a handler somewhere in the hmap
	 * --:: CAN ONLY BE CALLED FROM LISTENERS ASSOCIATED WITH THIS NODE!!! ::--
	 * @param listener The listener caller
	 * @param path The path on which to invoka the handler
	 * @param command The command to invoke the handler with
	 * @param parameters The parameters with which the handler is invoked
	 */
	public void callback(Listener listener, HmapPath path, String command, HmapNode parameters){
		if(listeners != null){
			Iterator iter = listeners.iterator();
			while(iter.hasNext()){
				if(((Listener)iter.next()).equals(listener)){
					invokeHandlers(new HmapEvent(getPath(), path, command, parameters));
				}
			}
		}
	}

	/**
	 * The actual handler invocation
	 * @param event The event to invoke the handlers with
	 */
	protected void handle(HmapEvent event){
		if(handlers != null){
			Iterator iter = handlers.iterator();
			while(iter.hasNext()){
				((Handler)iter.next()).handle(event);
			}
		}
	}
	
	/**
	 * @param path the path to resolve
	 * @return The first dir at the given path. null if no such dir exists
	 */
	public HmapDir resolveDir(String path) {
		HmapNode node = resolve(HmapPath.fromString(path), HET_DIR);
		if ((node != null) && node.isDir()) {
			return node.asDir();
		}
		return null;
	}

	/**
	 * @param path the path to resolve
	 * @return The first dir at the given path. null if no such dir exists
	 */
	public HmapDir resolveDir(HmapPath path) {
		HmapNode node = resolve(path, HET_DIR);
		if ((node != null) && node.isDir()) {
			return node.asDir();
		}
		return null;
	}

	/**
	 * @param key the key of the dir to resolve
	 * @return The local dir with the given key. null if no such dir exists
	 */
	public HmapDir resolveDirLocally(HmapKey key) {
		HmapNode node = resolveLocally(key, HET_DIR);
		if ((node != null) && node.isDir()) {
			return node.asDir();
		}
		return null;
	}
	/**
	 * @param name the name of the dir to resolve
	 * @return The local dir with the given name. null if no such dir exists
	 */
	public HmapDir resolveDirLocally(String name) {
		HmapNode node = resolveLocally(new HmapKey(name), HET_DIR);
		if ((node != null) && node.isDir()) {
			return node.asDir();
		}
		return null;
	}

	/**
	 * @param path the leaf to resolve
	 * @return The first leaf at the given path. null if no such leaf exists
	 */
	public HmapLeaf resolveLeaf(String path) {
		HmapNode node = resolve(HmapPath.fromString(path), HET_LEAF);
		if ((node != null) && node.isLeaf()) {
			return node.asLeaf();
		}
		return null;
	}

	/**
	 * @param path the leaf to resolve
	 * @return The first leaf at the given path. null if no such leaf exists
	 */
	public HmapLeaf resolveLeaf(HmapPath path) {
		HmapNode node = resolve(path, HET_LEAF);
		if ((node != null) && node.isLeaf()) {
			return node.asLeaf();
		}
		return null;
	}

	/**
	 * @param key the key of the leaf to resolve
	 * @return The first leaf at the given key. null if no such leaf exists
	 */
	public HmapLeaf resolveLeafLocally(HmapKey key){
		HmapNode node = resolveLocally(key, HET_LEAF);
		if ((node != null) && node.isLeaf()) {
			return node.asLeaf();
		}
		return null;
	}
	
	/**
	 * @param name the name of the leaf to resolve
	 * @return The leaf at the given key. null if no such leaf exists
	 */
	public HmapLeaf resolveLeafLocally(String name){
		HmapNode node = resolveLocally(new HmapKey(name), HET_LEAF);
		if ((node != null) && node.isLeaf()) {
			return node.asLeaf();
		}
		return null;
	}
	
	/**
	 * @param path the path to resolve
	 * @return The node at the given path. null if no such node exists
	 */
	public HmapNode resolve(String path){
		return resolve(HmapPath.fromString(path), HET_UNKNOWN);
	}
	
	/**
	 * @param key The key of the node to resolve
	 * @return The node coresponding to the key. Null if none is found
	 */
	public HmapNode resolveLocally(HmapKey key){
		return resolveLocally(key, HET_UNKNOWN);
	}
	
	/**
	 * @param name The name of the node to resolve
	 * @return The node coresponding to the key. Null if none is found
	 */
	public HmapNode resolveLocally(String name){
		return resolveLocally(new HmapKey(name), HET_UNKNOWN);
	}
	
	/**
	 * Add a handler to this node
	 * @param handler The handler to be added
	 */
	public void addHandler(Handler handler){
		if(handlers == null){
			handlers = new ArrayList();
		}
		handlers.add(handler);
	}
	
	/**
	 * Add an invocator to this node
	 * @param invocator The invocator to be added
	 */
	public void addInvokator(Invokator invocator){
		if(invocators == null){
			invocators = new ArrayList();
		}
		invocators.add(invocator);
	}
	
	/**
	 * Add a listener to this node
	 * @param listener The listener to be added
	 */
	public void addListener(Listener listener){
		if(listeners == null){
			listeners = new ArrayList(); 
		}
		listeners.add(listener);
	}
	
	/**
	 * Remove a handler from this node
	 * @param handler The handler to remove
	 */
	public void removeHandler(Handler handler){
		if(handlers != null){
			handlers.remove(handler);			
		}
	}
	
	/**
	 * Remove an invocator from this node
	 * @param invocator The invocator to remove
	 */
	public void removeInvocator(Invokator invocator){
		if(invocators != null){
			invocators.remove(invocator);
		}
	}
	
	/**
	 * Remove a listener from this node
	 * @param listener The listener to remove
	 */
	public void removeListener(Listener listener){
		if(listeners != null){
			listeners.remove(listener);
		}
	}
	
	/**
	 * @return An iterator over the collection of handlers associated with this node
	 */
	public Iterator handlerIterator(){
		if(handlers != null){
			return handlers.iterator();
		}
		return null;
	}
	
	/**
	 * @return An iterator over the collection of invocators associated with this node
	 */
	public Iterator invokatorIterator(){
		if(invocators != null){
			return invocators.iterator();
		}
		return null;
	}
	
	/**
	 * @return An iterator over the collection of listeners associated with this node
	 */
	public Iterator listenerIterator(){
		if(listeners != null){
			return listeners.iterator();
		}
		return null;
	}
	
	/**
	 * @return true if the node has any handlers. False if not
	 */
	public boolean hasHandlers(){
		return handlers != null;
	}
	
	/**
	 * @return true if the node has any invokators. False if not
	 */
	public boolean hasInvokators(){
		return invocators != null;
	}
	
	/**
	 * @return true if the node has any listeners. False if not
	 */
	public boolean hasListeners(){
		return listeners != null;
	}
	
	/**
	 * @return The Entry type of this node; Can be Dir, Leaf or Unknown
	 */
	public int getHmapEntryType(){
		return hmapEntryType;
	}
	
	/**
	 * Triggers all listeners listening on this node with an event.
	 * @param event
	 */
	protected void invokeListeners(HmapEvent event){
		if(hasListeners()){
			Iterator iter = listenerIterator();
			while(iter.hasNext()){
				Listener listener = (Listener)iter.next();
				listener.trigger(event);
			}
		}
	}
	
	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node's handlers
	 */
	public String printHandlers(int indent){
		if(hasHandlers()){
			StringBuilder tmp = StringBuilder.instantiate();
			StringBuilder whitespace = StringBuilder.instantiate();
			for(int i = 0; i< indent; i++){
				whitespace.append(" ");
			}
			tmp.append(whitespace.toString());
			tmp.append("Handlers:\r\n");
			Iterator iter = handlerIterator();
			while(iter.hasNext()){
				Handler handler = (Handler)iter.next();
				tmp.append(whitespace.toString());
				tmp.append("-----------------\r\n");
				tmp.append(handler.toString(indent + 1));
				tmp.append(whitespace.toString());
				tmp.append("-----------------\r\n");
			}
			return tmp.toString();
		}
		return "";
	}
	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node's invokators
	 */
	public String printInvokators(int indent){
		if(hasInvokators()){
			StringBuilder tmp = StringBuilder.instantiate();
			StringBuilder whitespace = StringBuilder.instantiate();
			for(int i = 0; i< indent; i++){
				whitespace.append(" ");
			}
			tmp.append(whitespace.toString());
			tmp.append("Invokators:\r\n");
			Iterator iter = invokatorIterator();
			while(iter.hasNext()){
				Invokator invokator = (Invokator)iter.next();
				tmp.append(whitespace.toString());
				tmp.append("-----------------\r\n");
				tmp.append(invokator.toString(indent + 1));
				tmp.append(whitespace.toString());
				tmp.append("-----------------\r\n");
			}
			return tmp.toString();
		}
		return "";
	}
	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node's listeners
	 */
	public String printListeners(int indent){
		if(hasListeners()){
			StringBuilder tmp = StringBuilder.instantiate();
			StringBuilder whitespace = StringBuilder.instantiate();
			for(int i = 0; i< indent; i++){
				whitespace.append(" ");
			}
			tmp.append(whitespace.toString());
			tmp.append("Listeners:\r\n");
			Iterator iter = listenerIterator();
			while(iter.hasNext()){
				Listener listener = (Listener)iter.next();
				tmp.append(whitespace.toString());
				tmp.append("-----------------\r\n");
				tmp.append(listener.toString(indent + 1));
				tmp.append(whitespace.toString());
				tmp.append("-----------------\r\n");
			}
			return tmp.toString();
		}
		return "";
	}
	
	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node
	 */
    public abstract String toString(int indent);

	/**
	 * @param indent prepend a number of whitespaces to the retured string
	 * @return a stringified version of this node with all handlers, invokators and listernes also
	 */
    public abstract String toStringAll(int indent);

    /**
	 * Resovles and invokes handlers specified in the event
	 * @param event The event to invoke handlers with
	 */
	protected abstract void invokeHandlers(HmapEvent event);
	
    /**
     * Returns true if this node is a direct ancestor of <node>
     * @param node The potential (grand)child node
     */
    protected abstract boolean isAncestorOf(HmapNode node);

    /**
     * Resolves a node in the current h-map
     * @param path the path of the node to resolve
     * @param type the type of the given node. Can be leaf, dir and unknown
     * @return The resolved node; null if none was resolved
     */
	public abstract HmapNode resolve(HmapPath path, int type);
	
	/**
	 * Resolves a child of the current node
	 * @param key The key of the child
	 * @param type The type of the given node. Can be leaf, dir and unknown
	 * @return The resolved node; null if none was resolved
	 */
	public abstract HmapNode resolveLocally(HmapKey key, int type);
	
	/**
	 * Recalculates the path on the basis of the path of ones ancestors
	 * @param parentPath The path of ones parent
	 */
	protected abstract void recalculatePath(HmapPath parentPath);

	/**
	 * The handler class for HmapNode names.
	 * Responsible for handling: getName and setName events
	 * @author tonz
	 *
	 */
	public class NameHandler extends AbstractHandler{
		
		private static final String name = "Node Name Handler";
		private static final String description = "Handles the getting and setting of the name of this node. This is done with the getName and setName commands";
		
		public void handle(HmapEvent event){
			if(event.getCommand().equals(HMAPNODE_COMMAND_GETNAME)){
				returnNameInvokator.returnName(event);
			}else if(event.getCommand().equals(HMAPNODE_COMMAND_SETNAME)){
				setName(event);
			}
		}
		
		public String getName(){
			return name;
		}
		
		public String getDescription(){
			return description;
		}
	}
	
	private class ReturnNameInvokator extends AbstractInvokator{

		private static final String name = "Node Return Name Invokator";
		private static final String description = "TODO...";
		
		public ReturnNameInvokator(HmapNode associatedNode){
			super(associatedNode);
		}
		
		public void returnName(HmapEvent event) {
			HmapNode tmp = event.getParameters();
			if((tmp != null) && tmp.isLeaf()){
				if((tmp.asLeaf().getValue() != null) && (tmp.asLeaf().getValue() instanceof String)){
					invokation(this, event.getFromPath(), (String)tmp.asLeaf().getValue(), new HmapLeaf("", getKey().getName()));		
				}
			}
		}

		public String getDescription() {
			return description;
		}

		public String getName() {
			return name;
		}		
	}	
}
