/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.info;

import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDAssemblyVer;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.PRDServiceListItem;

/**
 * This interface defines an handler for a list of infos. The methods in this
 * interface are for a read-only handler (
 * {@link ist.palcom.info.ModifiableInfoHandler}extends the behavior with
 * updating methods).
 */
public interface InfoHandler {

    int getNumberOfInfos();

    /**
     * @return The Info at index index, or null if there is no such item.
     */
    PRDData getInfo(int index);

    /**
     * @return The Info in the list that equals info, or null if none exists.
     */
    PRDData getInfoThatEquals(PRDData info);

    /**
     * @return The Info in the list that is about the same entity, or null if
     *         none exists.
     */
    PRDData getInfoThatIsAboutTheSameEntity(PRDData info);

    /** Returns true if the InfoHandler contains an info that equals info. */
    boolean containsInfoThatEquals(PRDData info);

    /**
     * Returns true if the InfoHandler contains an info that is about the same
     * entity as info.
     */
    boolean containsInfoThatIsAboutTheSameEntity(PRDData info);

    /**
     * Adds a listener for InfoEvents posted after the list changes.
     */
    void addInfoListener(InfoListener l);

    /**
     * Removes a listener previously added with addInfoListener().
     */
    void removeInfoListener(InfoListener l);

    /**
     * @return The service list item in this handler's list which has LocalSID serviceLocalSID. The
     *         whole trees will be searched, and a service or a sublist will be
     *         returned if it matches. Null is returned if no matching service list item
     *         is found.
     */
    PRDServiceListItem getServiceListItem(LocalSID serviceLocalSID);

    /**
     * @return The DeviceInfo in this handler's list which has String deviceID.
     *         Null is returned if no matching device info is found.
     */
    PRDDevice getDeviceInfo(DeviceID deviceID);

    /** @return The list of subservices for the info with LocalSID parentLocalSID. */
    PRDServiceList getSubserviceList(LocalSID parentLocalSID);

    /**
     * @return The ServiceDescription in this handler's list for the
     *         provider which has LocalSID providerLocalSID. Null is returned if no
     *         matching device info is found.
     */
    PRDServiceFMDescription getServiceDescription(LocalSID providerLocalSID);

    /**
     * Returns the long name of the service with a given LocalSID, or null if it does
     * not exist in this handler.
     */
    public String getLongName(LocalSID serviceLocalSID);

    /**
     * @return True if the connection in connectionInfo is possible, i.e. if
     *         ServiceInfos for both its parties are in this InfoHandler.
     */
    boolean connectionIsPossible(PRDConnection connectionInfo);

    /**
     * @return True if the assembly in assemblyInfo is fully established, i.e.
     *         if ConnectionInfos for all its connections are in this
     *         InfoHandler, and they are all active.
     */
    boolean assemblyIsEstablished(PRDAssemblyVer assemblyInfo);

    /**
     * @return True if the assembly in assemblyInfo is possible, i.e. if all its
     *         connections are possible, in the same sense as for
     *         {@link #connectionIsPossible(PRDConnection)}.
     */
    boolean assemblyIsPossible(PRDAssemblyVer assemblyInfo);

    /**
     * @return true if there is at least one matching customer-service pair in
     *         info1 and info2, looking in this InfoHandler. The whole trees of
     *         subservices are examined.
     */
    boolean servicesMatch(PRDData info1, PRDData info2);

    /**
     * @return an array of ConnectionInfos for matching customer-service pairs
     *         in info1 and info2, looking in this InfoHandler.
     */
    PRDConnection[] getMatchingConnectionInfos(PRDData info1, PRDData info2);
}