/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.info;

import ist.palcom.collections.LinearMap;
import ist.palcom.discovery.InfoEvent;
import ist.palcom.discovery.ServiceStatusEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDAssemblyVer;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.PRDServiceListItem;

import java.util.ArrayList;

//TODO Add tests of the matching operations

//TODO Make this class a Monitor?

/** An InfoHandler implementation, with updating methods added. */
public class ModifiableInfoHandler implements InfoHandler, InfoListener {

	/** The infos. */
	private ArrayList infos;

	private ArrayList infoListeners;

	/** A HashMap for optimizing the getServiceInfo method. */
//	private HashMap<String, ServiceInfo> serviceInfos;
	private LinearMap serviceListItems;

	/** A map from provider LocalSIDs to ServiceDescriptions. */
//	private HashMap<String, ServiceDescription> serviceDescriptions;
	private LinearMap serviceDescriptions;

	/** Map from parent LocalSIDs to subservice lists. */
//	private HashMap<String, ServiceList> subserviceLists;
	private LinearMap subserviceLists;

	private LinearMap serviceStatusEvents;

	/**
	 * HashMap with cached looked-up parents, with mappings from LocalSIDs to parent
	 * ServiceInfos.
	 */
//	private HashMap cachedParents;
	private LinearMap cachedParents;

	public ModifiableInfoHandler() {
		infos = new ArrayList();
		infoListeners = new ArrayList();
//		serviceInfos = new HashMap<String, ServiceInfo>();
//		serviceDescriptions = new HashMap<String, ServiceDescription>();
//		subserviceLists = new HashMap<String, ServiceList>();
//		cachedParents = new HashMap();
		serviceListItems = new LinearMap();
		serviceDescriptions = new LinearMap();
		subserviceLists = new LinearMap();
		cachedParents = new LinearMap();

		serviceStatusEvents = new LinearMap();
	}

	public int getNumberOfInfos() {
		return infos.size();
	}

	public PRDData getInfo(int index) {
		try {
			return (PRDData) infos.get(index);
		} catch (IndexOutOfBoundsException e) {
			return null;
		}
	}

	public PRDData getInfoThatEquals(PRDData info) {
		for (int i = 0; i < getNumberOfInfos(); i++) {
			PRDData inf = getInfo(i).getPRDDataThatEquals(info);
			if (inf != null) {
				return inf;
			}
		}
		return null;
	}

	public PRDData getInfoThatIsAboutTheSameEntity(PRDData info) {
		for (int i = 0; i < getNumberOfInfos(); i++) {
			PRDData inf = getInfo(i).getPRDDataThatIsAboutTheSameEntityAs(info);
			if (inf != null) {
				return inf;
			}
		}
		return null;
	}

	private int getIndexOfInfoThatIsAboutTheSameEntity(PRDData info) {
		for (int i = 0; i < getNumberOfInfos(); i++) {
			PRDData inf = getInfo(i);
			if (inf.isAboutTheSameEntityAs(info)) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Adds an Info to the list, or, if an info about the same entity is already
	 * in the list, updates it.
	 * 
	 * InfoEvents will be sent to registered listeners, if the info is changed
	 * in the list.
	 */
	public void updateInfo(PRDData info) {
		int index = -1;
		boolean updated = false;
		index = getIndexOfInfoThatIsAboutTheSameEntity(info);
		if (index == -1) {
			infos.add(info);
			index = infos.size() - 1;
			updated = true;
		} else {
			PRDData current = getInfo(index);
			if (!info.equals(current)) {
				infos.set(index, info);
				updated = true;
			}
		}

		if (updated) {
			if (info instanceof PRDServiceFMDescription) {
				updateHashMapWithServiceDescription((PRDServiceFMDescription) info);
			} else if (info instanceof PRDServiceList) {
				updateHashMapsWithSubserviceList((PRDServiceList) info);
			}

			InfoEvent event = new InfoEvent(info, true);
			fireInfo(event);
		}
	}

	/**
	 * Removes the info from the list, that is about the same entity as info. If
	 * there is no such Info in the list, nothing happens (comparison with
	 * equals).
	 * 
	 * If an Info is removed, InfoEvents will be sent to registered listeners.
	 */
	public void removeInfo(PRDData info) {
		int index = index = getIndexOfInfoThatIsAboutTheSameEntity(info);
		if (index != -1) {
			infos.remove(index);

			if (info instanceof PRDServiceFMDescription) {
				PRDServiceFMDescription sd = (PRDServiceFMDescription) info;
				removeServiceDescriptionFromHashMap(
						sd.getLocalSID());
			} else if (info instanceof PRDServiceList) {
				removeSubserviceListFromHashMaps(((PRDServiceList) info)
						.getParentLocalSID());
			}
		}

		if (index != -1) {
			InfoEvent event = new InfoEvent(info, false);
			fireInfo(event);
		}
	}

	public void infoReceived(InfoEvent event) {
		if (event.infoShouldBeKept()) {
			updateInfo(event.getInfo());
		} else {
			removeInfo(event.getInfo());
		}
	}

	public void eventReceived(Event event) {
		if(event instanceof ServiceStatusEvent) {
			ServiceStatusEvent serviceStatus = (ServiceStatusEvent) event;
			serviceStatusEvents.put(serviceStatus.getLocalSID(), serviceStatus);
			PRDServiceListItem listItem = getServiceListItem(
					serviceStatus.getLocalSID());
			if(listItem != null) {
				fireInfo(new InfoEvent(listItem, true));
			}
		}
	}

	/** Send out InfoEvents for all infos to listeners. */
	public void fireAllInfos() {
		Object[] infosInArray = infos.toArray();
		for (int i = 0; i < infosInArray.length; i++) {
			PRDData info = (PRDData) infosInArray[i];
			InfoEvent event = new InfoEvent(info, true);
			fireInfo(event);
		}
	}

	private void fireInfo(InfoEvent event) {
		for (int i = 0; i < infoListeners.size(); i++) {
			InfoListener listener = (InfoListener) infoListeners.get(i);
			listener.infoReceived(event);
		}
	}

	// private void updateHashMapWithServiceInfo(ServiceInfo info) {
	// serviceInfos.put(info.getURN(), info);
	// }
	//
	// private void removeServiceInfoFromHashMap(String urn) {
	// serviceInfos.remove(urn);
	// }

	private void updateHashMapWithServiceDescription(
			PRDServiceFMDescription sd) {
		serviceDescriptions.put(sd.getLocalSID(), sd);
	}

	private void removeServiceDescriptionFromHashMap(LocalSID providerLocalSID) {
		serviceDescriptions.remove(providerLocalSID);
	}

	private void updateHashMapsWithSubserviceList(PRDServiceList sl) {
		subserviceLists.put(sl.getParentLocalSID(), sl);
		for (int i = 0; i < sl.getNumPRDServiceListItem(); i++) {
			PRDServiceListItem si = (PRDServiceListItem)sl.getPRDServiceListItem(i);
			serviceListItems.put(si.getLocalSID(), si);
		}
	}

	private void removeSubserviceListFromHashMaps(LocalSID parentLocalSID) {
		PRDServiceList sl = (PRDServiceList) subserviceLists.get(parentLocalSID);
		for (int i = 0; i < sl.getNumPRDServiceListItem(); i++) {
			serviceListItems.remove(sl.getPRDServiceListItem(i).getLocalSID());
			if(sl.getPRDServiceListItem(i) instanceof PRDService) {
				serviceStatusEvents.remove(sl.getPRDServiceListItem(i).getLocalSID());
			}
		}
		subserviceLists.remove(parentLocalSID);
	}

	/**
	 * Clears the list of Infos. InfoEvents may be generated and passed to
	 * listeners that have registered in this object.
	 */
	public void clear() {
		while (infos.size() > 0) {
			removeInfo(getInfo(0));
		}
	}

	public boolean containsInfoThatEquals(PRDData info) {
		return getInfoThatEquals(info) != null;
	}

	public boolean containsInfoThatIsAboutTheSameEntity(PRDData info) {
		return getInfoThatIsAboutTheSameEntity(info) != null;
	}

	public void addInfoListener(InfoListener l) {
		infoListeners.add(l);
	}

	public void removeInfoListener(InfoListener l) {
		infoListeners.remove(l);
	}

	public PRDServiceListItem getServiceListItem(LocalSID serviceLocalSID) {
		return (PRDServiceListItem) serviceListItems.get(serviceLocalSID);
	}

	public ServiceStatusEvent getServiceStatusEvent(LocalSID serviceLocalSID) {
		return (ServiceStatusEvent) serviceStatusEvents.get(serviceLocalSID);
	}

	public PRDDevice getDeviceInfo(DeviceID deviceID) {
		return (PRDDevice) getInfoThatIsAboutTheSameEntity(new PRDDevice(
				deviceID));
	}

	public PRDServiceFMDescription getServiceDescription(LocalSID providerLocalSID) {
		return (PRDServiceFMDescription) serviceDescriptions.get(providerLocalSID);
	}

	public PRDServiceList getSubserviceList(LocalSID parentLocalSID) {
		return (PRDServiceList) subserviceLists.get(parentLocalSID);
	}

	public String getLongName(LocalSID serviceLocalSID) {
		LocalSID localSID = serviceLocalSID;
		PRDData info = getServiceListItem(localSID);
		if (info == null) {
			return null;
		} else {
			StringBuffer result = new StringBuffer(info.getName());
			info = getParentInfo(localSID);
			while (info != null) {
				result.append(" of " + info.getName());
				info = getParentInfo(info.getLocalSID());
			}
			return result.toString();
		}
	}

	private PRDData getParentInfo(LocalSID serviceLocalSID) {
		PRDService parentInfo = (PRDService) cachedParents.get(serviceLocalSID);
		if (parentInfo != null) {
			PRDServiceList subservices = getSubserviceList(parentInfo.getLocalSID());
			if (subservices != null) {
				for (int i = 0; i < subservices.getNumPRDServiceListItem(); i++) {
					if(subservices.getPRDServiceListItem(i) instanceof PRDService) {
						PRDService si = (PRDService)subservices.getPRDServiceListItem(i);
						if (si.getLocalSID().equals(serviceLocalSID)) {
							//The mapping is still valid - return
							return parentInfo;
						}
					}
				}
				//Remove the mapping and go on to find a new one
				cachedParents.remove(serviceLocalSID);
			}

			for (int i = 0; i < getNumberOfInfos(); i++) {
				PRDData info = getInfo(i);
				subservices = getSubserviceList(parentInfo.getLocalSID());
				if (subservices != null) {
					for (int j = 0; j < subservices.getNumPRDServiceListItem(); j++) {
						if(subservices.getPRDServiceListItem(i) instanceof PRDService) {
							PRDService si = (PRDService)subservices.getPRDServiceListItem(j);
							if (si.getLocalSID().equals(serviceLocalSID)) {
								return info;
							}
						}
					}
				}
			}
		}

		return null;
	}

	public boolean connectionIsPossible(PRDConnection connectionInfo) {
		PRDServiceListItem providerInfo = getServiceListItem(connectionInfo
				.getLocalSID1());
		PRDServiceListItem customerInfo = getServiceListItem(connectionInfo
				.getLocalSID2());
		return providerInfo != null && customerInfo != null
		&& providerInfo instanceof PRDService && customerInfo instanceof PRDService ;
	}

	public boolean assemblyIsEstablished(PRDAssemblyVer ai) {
		for (int i = 0; i < ai.getConnections().getNumConnectionDecl(); i++) {
			// Do a lookup, do not use the ConnectionInfo directly
			PRDConnection ci = (PRDConnection) getInfoThatIsAboutTheSameEntity(ai
					.getConnections().getConnectionDecl(i).asInfo().getPRDConnection());
			if (ci == null) {
				return false;
			}
		}
		return true;
	}

	public boolean assemblyIsPossible(PRDAssemblyVer assemblyInfo) {
		for (int i = 0; i < assemblyInfo.getConnections()
		.getNumConnectionDecl(); i++) {
			PRDConnection connection = assemblyInfo.getConnections()
			.getConnectionDecl(i).asInfo().getPRDConnection();
			if (connection != null && !connectionIsPossible(connection)) {
				return false;
			}
		}
		return true;
	}

	public boolean servicesMatch(PRDData info1, PRDData info2) {
		// Get the current infos
		info1 = getInfoThatEquals(info1);
		info2 = getInfoThatEquals(info2);

		if (rootMatches(info1, info2)) {
			return true;
		} else {
			if (info1 != null) {
				LocalSID localSID = info1.getLocalSID();
				if(info1 instanceof PRDDevice) {
					localSID = new LocalSID(((PRDDevice)info1).getDeviceID(), null);
				}
				PRDServiceList subservices = getSubserviceList(localSID);
				if (subservices != null) {
					for (int i = 0; i < subservices.getNumPRDServiceListItem(); i++) {
						if (servicesMatch(subservices
								.getPRDServiceListItem(i), info2)) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * @return true if the root of info1 matches at least one customer or
	 *         service in the service identified by info2 (looking at
	 *         subservices, too).
	 */
	private boolean rootMatches(PRDData info1, PRDData info2) {
		if (rootsMatch(info1, info2)) {
			return true;
		} else {
			if (info2 != null) {
				LocalSID localSID = info2.getLocalSID();
				if(info2 instanceof PRDDevice) {
					localSID = new LocalSID(((PRDDevice)info2).getDeviceID(), null);
				}
				PRDServiceList subservices = getSubserviceList(localSID);
				if (subservices != null) {
					for (int i = 0; i < subservices.getNumPRDServiceListItem(); i++) {
						if (rootMatches(info1, subservices
								.getPRDServiceListItem(i))) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/** @return true if the root of info1 matches the root of info2. */
	private boolean rootsMatch(PRDData info1, PRDData info2) {
		if (info1 == null || info2 == null) {
			return false;
		} else if (info1 instanceof PRDService && info2 instanceof PRDService) {
			PRDService serviceInfo1 = (PRDService) info1;
			PRDService serviceInfo2 = (PRDService) info2;
			if ((serviceInfo1.getHasDesc() && !serviceInfo2.getHasDesc())
					|| (!serviceInfo1.getHasDesc() && serviceInfo2.getHasDesc())) {
				return true;
			} else {
				return false;
			}
		} else {
			return false;
		}
	}

	public PRDConnection[] getMatchingConnectionInfos(PRDData info1, PRDData info2) {
		// Get the current infos
		info1 = getInfoThatEquals(info1);
		info2 = getInfoThatEquals(info2);

		// Create result list
		ArrayList resultList = new ArrayList();

		// Add matches for the whole trees to the list
		addMatchingConnectionInfos(info1, info2, resultList);

		// Convert result list to array and return it
		PRDConnection[] result = new PRDConnection[resultList.size()];
		for (int i = 0; i < result.length; i++) {
			result[i] = (PRDConnection) resultList.get(i);
		}
		return result;
	}

	/**
	 * Adds ConnectionInfos for matching customer-service pairs in info1 and
	 * info2 to resultList.
	 */
	private void addMatchingConnectionInfos(PRDData info1, PRDData info2,
			ArrayList resultList) {
		addRootMatchingConnectionInfos(info1, info2, resultList);
		if (info1 != null) {
			LocalSID localSID = info1.getLocalSID();
			if(info1 instanceof PRDDevice) {
				localSID = new LocalSID(((PRDDevice)info1).getDeviceID(), null);
			}
			PRDServiceList subservices = getSubserviceList(localSID);
			if (subservices != null) {
				for (int i = 0; i < subservices.getNumPRDServiceListItem(); i++) {
					addMatchingConnectionInfos(subservices.getPRDServiceListItem(i), info2,
							resultList);
				}
			}
		}
	}

	/**
	 * Adds matching ConnectionInfos between the root of info1 and services of
	 * info2 to resultList.
	 */
	private void addRootMatchingConnectionInfos(PRDData info1, PRDData info2,
			ArrayList resultList) {

		if (info1 != null && info2 != null && info1 instanceof PRDService) {
			if (rootsMatch(info1, info2)) {
				PRDService serviceInfo1 = (PRDService) info1;
				PRDService serviceInfo2 = (PRDService) info2;
				PRDService providerInfo;
				PRDService customerInfo;
				if (serviceInfo1.isCustomer()) {
					providerInfo = serviceInfo2;
					customerInfo = serviceInfo1;
				} else {
					providerInfo = serviceInfo1;
					customerInfo = serviceInfo2;
				}

				resultList.add(new PRDConnection(providerInfo.getLocalSID(),
						customerInfo.getLocalSID()));
			} else {
				LocalSID localSID = info2.getLocalSID();
				if(info2 instanceof PRDDevice) {
					localSID = new LocalSID(((PRDDevice)info2).getDeviceID(), null);
				}
				PRDServiceList subservices = getSubserviceList(localSID);
				if (subservices != null) {
					for (int i = 0; i < subservices.getNumPRDServiceListItem(); i++) {
						if(subservices.getPRDServiceListItem(i) instanceof PRDService) {
							PRDService subserviceInfo = (PRDService)subservices
							.getPRDServiceListItem(i);
							addRootMatchingConnectionInfos(info1, subserviceInfo,
									resultList);
						}
					}
				}
			}
		}
	}
}