/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal.udp;

import ist.palcom.base.networking.DatagramPacket;
import ist.palcom.base.networking.DatagramSocket;
import ist.palcom.mal.AbstractMediaManager;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.MessageArrayWithURLEvent;
import ist.palcom.mal.StopEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.IOAvailableEvent;
import ist.palcom.palcomthreads.IOTimeoutEvent;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.URL;

import java.io.IOException;

/**
 * UDPCommunicationThread is a CommunicationThread for UDP communication. It
 * sends an receives on one socket (the concrete subclasses handle different
 * kinds of UDP sockets).
 * 
 * @see ist.palcom.mal.CommunicationThread CommunicationThread
 * 
 * @author David Svensson, david@cs.lth.se
 */
public abstract class UDPCommunicationThread extends CommunicationThread {

    protected static final String URL_PREFIX = "datagram://";

    private PalcomScheduler scheduler;

    protected DatagramSocket socket;

    private PalcomThread distributor;

    protected UDPCommunicationThread(PalcomScheduler scheduler,
            PalcomThread distributor, DatagramSocket socket) {
        this.socket = socket;
        this.scheduler = scheduler;
        this.distributor = distributor;
    }

    /** Close socket if needed. */
    protected void tearDown() {
        socket.close();
    }

    public int getReceiveBufferSize() throws IOException {
        return socket.getReceiveBufferSize();
    }

    public int getSendBufferSize() throws IOException {
        return socket.getSendBufferSize();
    }

    public void setReceiveBufferSize(int size) throws IOException {
        socket.setReceiveBufferSize(size);
    }

    public void setSendBufferSize(int size) throws IOException {
        socket.setSendBufferSize(size);
    }

    public void run() {
        final int timeout = 1000;
        enableIOEvent(socket, timeout);
        while (true) {
            Event event = waitEvent();
            if (event instanceof IOAvailableEvent) {
            	try {
                    // Get up to 64k packet (max UDP packet size).
                    DatagramPacket packet = socket.receive(getReceiveBufferSize());
//                    Message message = Message.parse(packet.getData(), 0, packet
//                            .getLength());
//                    if (message != null) {
                    	URL senderURL = new URL("datagram://"
                                + packet.getAddress() + ":"
                                + packet.getPort());
//                        message.setSenderURL(new URL("datagram://"
//                                        + packet.getAddress() + ":"
//                                        + packet.getPort()));
//                        message.setReceiverURL(getURL());

                        if (distributor != null) {
                            distributor.putEvent(new MessageArrayWithURLEvent(
                                    packet.getData(), senderURL));
                        }
//                    }
                    enableIOEvent(socket, timeout);
                    
                } catch (IOException e) {
                    distributor.putEvent(new MALErrorEvent(e));
                }
            } else if (event instanceof IOTimeoutEvent) {
                // Try again
                enableIOEvent(socket, timeout);
            } else if (event instanceof MessageArrayWithURLEvent) {
            	try {
                    MessageArrayWithURLEvent mawue = (MessageArrayWithURLEvent) event;
                    URL receiverURL = mawue.getDeviceURL();
                    byte[] messageData = mawue.getMessageArray();
    
                    String receiverHost = AbstractMediaManager.getHost(receiverURL);
                    int receiverPort = AbstractMediaManager.getPort(receiverURL);
                    DatagramPacket datagram = new DatagramPacket(messageData,
                            receiverHost, receiverPort);
                    socket.send(datagram);
                } catch (IOException e) {
                    distributor.putEvent(new MALErrorEvent(e));
                }
            } else if (event instanceof StopEvent) {
                tearDown();
                terminate();
            } else {
                //System.out.println("Unknown event");
            }
        }
    }
}
