/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.mal.udp;

import ist.palcom.mal.AbstractMediaManager;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.StopEvent;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.URL;

import java.io.IOException;
import java.util.ArrayList;

/**
 * This class implement is a media manager for UDP.
 * 
 * @see ist.palcom.mal.MediaManager MediaManager
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class UDPMediaManager extends AbstractMediaManager {

    private ArrayList threadURLs;

    private ArrayList threads;

    private URL defaultUnicastURL;

    private URL[] defaultGroupAddresses;

    /** The receive buffer size to set for new UDP sockets, not used when zero. */
    private int defaultReceiveBufferSize;

    /** The send buffer size to set for new UDP sockets, not used when zero. */
    private int defaultSendBufferSize;
    
    private int suggestedSendBufferSize;
    
    private int suggestedReceiveBufferSize;

    /**
     * @deprecated See superclass
     */
    public UDPMediaManager(PalcomScheduler scheduler) {
    	this(scheduler, null);
    	System.err.println("Warning: Using deprecated UDPMediaManager constructor");
    }
    
    public UDPMediaManager(PalcomScheduler scheduler, PalcomThread errorHandler) {
        super(scheduler, errorHandler);
        threadURLs = new ArrayList();
        threads = new ArrayList();

        defaultGroupAddresses = new URL[] { new URL(
                "palcom:datagram://255.255.255.255:8031") };
        // This is limited broadcast, port 8031
        //defaultGroupAddresses = new URL[] { new URL(
        //        "palcom:datagram://224.0.1.20:8031") };
        // This is "experiment.mcast.net", port 8031

        defaultReceiveBufferSize = 8192;
        defaultSendBufferSize = 8192;
        suggestedSendBufferSize = 8192;
        suggestedReceiveBufferSize = 8192;
    }

    public void stop() {
        for (int i = 0; i < threads.size(); i++) {
            CommunicationThread thread = (CommunicationThread) threads.get(i);
            thread.putEvent(new StopEvent());
        }
        threadURLs.clear();
        threads.clear();

        super.stop();
    }

    public URL getDefaultUnicastURL() {
        if (defaultUnicastURL == null) {
        	try {
	            UDPCommunicationThread defaultUnicastThread = new UDPUnicastCommunicationThread(
	                    scheduler, distributorThread);
	            setReceiveBufferSize(defaultUnicastThread);
	            setSendBufferSize(defaultUnicastThread);
	            scheduler.scheduleThread(defaultUnicastThread);
	            defaultUnicastURL = defaultUnicastThread.getURL();
	            setThread(defaultUnicastURL, defaultUnicastThread);
	            
        	} catch (IOException e) {
        		raiseErrorEvent(new MALErrorEvent(e));
        	}
        }
        return defaultUnicastURL;
    }

    protected CommunicationThread getReceivingThread(URL url) {
        if (url == null) {
            throw new NullPointerException();
        }
        url = getProtocolSpecificUrl(url);
        UDPCommunicationThread thread = null;
        int index = getIndexOfThread(url);
        if (index != -1) {
            thread = (UDPCommunicationThread) threads.get(index);
        } else {
            String protocol = getProtocol(url);
            if (protocol.equals("datagram")) {
                String host = getHost(url);
                int port = getPort(url);
                try {
                    if (isBroadcast(url)) {
                        if (port != -1) {
                            thread = new UDPMulticastCommunicationThread(
                                    scheduler, distributorThread, host, port);

                        } else {
                            throw new IllegalArgumentException(
                                    "Port must be given for broadcast URL");
                        }
                    } else {
                        thread = new UDPUnicastCommunicationThread(scheduler,
                                distributorThread);
                    }
                    setReceiveBufferSize(thread);
                    setSendBufferSize(thread);
                    scheduler.scheduleThread(thread);
                    setThread(url, thread);
                } catch (IOException e) {
                    raiseErrorEvent(new MALErrorEvent(e));
                }
            } else {
                throw new IllegalArgumentException("Protocol " + protocol
                        + " not supported.");
            }
        }
        return thread;
    }

    protected void removeReceivingThread(URL url) {
        int index = getIndexOfThread(url);
        if (index != -1) {
            CommunicationThread thread = (CommunicationThread) threads
                    .get(index);
            thread.putEvent(new StopEvent());
            threadURLs.remove(index);
            threads.remove(index);
        }
    }

    protected CommunicationThread getSendingThread(URL url) {
        String protocol = getProtocol(url);
        if (protocol.equals("datagram")) {
            // Use receiving
            return getReceivingThread(getDefaultUnicastURL());
        } else {
            throw new IllegalArgumentException("Protocol of " + url
                    + " not supported.");
        }
    }

    /** @return the index of the thread for url, or -1 if not found. */
    private int getIndexOfThread(URL url) {
        url = getProtocolSpecificUrl(url);
        int index = 0;
        while (index < threadURLs.size() && !threadURLs.get(index).equals(url)) {
            index++;
        }
        if (index == threadURLs.size()) {
            return -1;
        } else {
            return index;
        }
    }

    /** Sets the receiving thread for URL url. */
    private void setThread(URL url, CommunicationThread thread) {
        url = getProtocolSpecificUrl(url);
        int index = getIndexOfThread(url);
        if (index == -1) {
            threadURLs.add(url);
            threads.add(thread);
        } else {
            threads.set(index, thread);
        }
    }

    /**
     * This method is called isBroadcast, but in the case of UDP it actually
     * checks if the address is a multicast address.
     */
    protected boolean isBroadcast(URL url) {
        if (url == null) {
            throw new NullPointerException();
        }
        url = getProtocolSpecificUrl(url);
        String protocol = getProtocol(url);
        if ("datagram".equals(protocol)) {
            String host = getHost(url);
            return isIPMulticast(host) || isIPBroadcast(host);
        } else {
            return false;
        }
    }

    public URL[] getDefaultGroupAddresses() {
        return defaultGroupAddresses;
    }

    /**
     * @return the default receive buffer size for UDP sockets, or zero if not
     *         set.
     */
    public int getDefaultReceiveBufferSize() {
        return defaultReceiveBufferSize;
    }

    /**
     * @return the default send buffer size for UDP sockets.
     */
    public int getSendBufferSize() {
        return defaultSendBufferSize;
    }

    public void suggestReceiveBufferSize(int size) {
        suggestedReceiveBufferSize = size;
    }
    
    public void suggestSendBufferSize(int size) {
        suggestedSendBufferSize = size;
    }

    private void setReceiveBufferSize(UDPCommunicationThread thread) throws IOException {
        if (defaultReceiveBufferSize != 0) {
            thread.setReceiveBufferSize(suggestedReceiveBufferSize);
            defaultReceiveBufferSize = thread.getReceiveBufferSize();
            
        }
    }

    private void setSendBufferSize(UDPCommunicationThread thread) throws IOException {
        if (defaultSendBufferSize != 0) {
            thread.setSendBufferSize(suggestedSendBufferSize);
            defaultSendBufferSize = thread.getSendBufferSize();
        }
    }
    
    static boolean isIPMulticast(String address) {
        int index = address.indexOf(".");
        if (index == -1) {
            return false;
        } else {
            int firstPart = Integer.parseInt(address.substring(0, index));
            return 224 <= firstPart && firstPart <= 239;
        }
    }
    
    static boolean isIPBroadcast(String address) {
        int index = address.lastIndexOf(".");
        if (index == -1) {
            return false;
        } else {
            int fourthPart = Integer.parseInt(address.substring(index + 1));
            return fourthPart == 255;
        }
    }
}
