/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.palcomthreads;

import ist.palcom.base.Channel;
import ist.palcom.base.Coroutine;

/**
 * @author David Svensson, david@cs.lth.se
 */
public abstract class PalcomThread {

    /** In ready queue, executing or waiting its turn. */
    static final int STATE_READY = 0;

    /** Has been descheduled or not yet started, not in ready queue. */
    static final int STATE_STOPPED = 1;

    /** Terminated, cannot be started again. */
    static final int STATE_TERMINATED = 2;

    /** waitEvent has been called in a Mailbox, waiting for events. */
    static final int STATE_WAITING = 3;

    int state;

    Coroutine coroutine;

    private PalcomScheduler scheduler;

    private Mailbox mailbox;

    public PalcomThread() {
        state = STATE_STOPPED; // Initially suspended
        coroutine = new PalcomThreadCoroutine();
        mailbox = new Mailbox();
        setScheduler(null);
    }

    /** This method is overridden by subclasses. */
    public void run() {
    }

    /** @return Reference to the PalcomScheduler of this PalcomThread (if any). */
    public PalcomScheduler getScheduler() {
        return scheduler;
    }

    void setScheduler(PalcomScheduler scheduler) {
        this.scheduler = scheduler;
        mailbox.setScheduler(scheduler);
    }

    /** Make this PalcomThread ready to be scheduled. */
    void schedule() {
        if ((state == STATE_STOPPED || state == STATE_WAITING)
                && scheduler != null) {
            state = STATE_READY;
            scheduler.makeReadyForScheduling(this);
        }
    }

    /**
     * Remove this PalcomThread (==scheduler.getCurrentThread()) from scheduling
     * and give control back to the scheduler.
     */
    void deschedule() {
        if (state == STATE_READY) {
            state = STATE_STOPPED;
            coroutine.suspend();
        }
    }

    /** Permanently remove this PalcomThread from scheduling. */
    public void terminate() {
        if ((state == STATE_READY || state == STATE_WAITING)
                && scheduler != null) {
            state = STATE_TERMINATED;
            coroutine.terminate();
        }
    }

    /**
     * Give other PalcomThreads with the same or higher priority a chance to
     * execute (if any).
     */
    public void yield() {
        coroutine.suspend();
    }

    /**
     * @return The next event in the mailbox of this PalcomThread, or block
     *         until one arrives.
     */
    public Event waitEvent() {
        return mailbox.waitEvent();
    }

    /**
     * @return true if there are pending event(s) in this PalcomThread's
     *         mailbox, else return false.
     */
    public boolean askEvent() {
        return mailbox.askEvent();
    }

    /** Deposit an event for this PalcomThread. */
    public void putEvent(Event event) {
        mailbox.putEvent(event);
    }

    /**
     * Order a PeriodicEvent for delivery every period milliseconds. If
     * PeriodicEvents have already been ordered for this PalcomThread, that
     * order is cancelled.
     */
    public void enablePeriodicEvents(long period) {
        if (scheduler != null) {
            scheduler.enablePeriodicEvents(this, period);
        }
    }

    /** Stop delivering PeriodicEvents. */
    public void disablePeriodicEvents() {
        if (scheduler != null) {
            scheduler.disablePeriodicEvents(this);
        }
    }

    /**
     * Order an IOAvailableEvent to be delivered when input is available on the
     * Channel channel, or deliver an IOTimeoutEvent after timeout milliseconds
     * if no input came. Never deliver both. If IOEvents have already been
     * ordered for this PalcomThread, that order is cancelled.
     */
    public void enableIOEvent(Channel channel, int timeout) {
        if (scheduler != null) {
            scheduler.enableIOEvent(this, channel, timeout);
        }
    }

    /** Cancel previous call to enableIOEvent. */
    public void disableIOEvents() {
        if (scheduler != null) {
            scheduler.disableIOEvents(this);
        }
    }

    /** @return True if this thread should be sorted before t. */
    abstract boolean isBefore(PalcomThread t);

    private class PalcomThreadCoroutine extends Coroutine {
        public void run() {
            PalcomThread.this.run();
            PalcomThread.this.terminate();
        }
    }
}