/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.palcomthreads.test;

import java.io.IOException;

import ist.palcom.base.networking.DatagramPacket;
import ist.palcom.base.networking.DatagramSocket;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.IOAvailableEvent;
import ist.palcom.palcomthreads.IOTimeoutEvent;
import ist.palcom.palcomthreads.Monitor;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.palcomthreads.Semaphore;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class PalcomThreadsTest {

    public static void main(String[] args) {
        //ist.palcom.base.System.instance().enableConstantGC();
        PalcomThreadsTest test = new PalcomThreadsTest();
        test.testPlainRunMethod();
        System.out.println();
        test.testYield();
        System.out.println();
        test.testSemaphore();
        System.out.println();
        test.testMonitor();
        System.out.println();
        test.testMailbox();
        System.out.println();
        test.testPeriodicEvents();
        System.out.println();
        test.testTerminateOwnThread();
        System.out.println();
        test.testTerminateOtherThread();
        System.out.println();
        test.testIOEvent();
        System.out.println();
    }

    private void testPlainRunMethod() {
        System.out.println("--- testPlainRunMethod ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        PalcomThread t1 = new PalcomPriorityThread() {
            public void run() {
                System.out.println("Done.");
            }
        };
        scheduler.scheduleThread(t1);
        scheduler.run();
    }

    private void testYield() {
        System.out.println("--- testYield ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        PalcomThread t1 = new PalcomPriorityThread() {
            public void run() {
                System.out.println("t1 - 1");
                yield();
                System.out.println("t1 - 2");
                yield();
                System.out.println("t1 - 3");
                System.out.println("t1 done.");
            }
        };
        scheduler.scheduleThread(t1);
        PalcomThread t2 = new PalcomPriorityThread() {
            public void run() {
                System.out.println("t2 - 1");
                yield();
                System.out.println("t2 - 2");
                System.out.println("t2 done.");
            }
        };
        scheduler.scheduleThread(t2);
        scheduler.run();
    }

    private void testSemaphore() {
        System.out.println("--- testSemaphore ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final Semaphore semaphore = new Semaphore(scheduler);
        PalcomThread consumer = new PalcomPriorityThread() {
            public void run() {
                System.out.println("Consumer wants to consume.");
                semaphore.take();
                System.out.println("Consumer consumes now.");
                System.out.println("Consumer done.");
            }
        };
        scheduler.scheduleThread(consumer);
        PalcomThread producer = new PalcomPriorityThread() {
            public void run() {
                System.out.println("Producer produces now.");
                semaphore.give();
                System.out.println("Producer done.");
            }
        };
        scheduler.scheduleThread(producer);
        scheduler.run();
    }

    private void testMonitor() {
        System.out.println("--- testMonitor ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final ConsumerProducerMonitor monitor = new ConsumerProducerMonitor(
                scheduler);
        PalcomThread consumer = new PalcomPriorityThread() {
            public void run() {
                System.out.println("Consumer wants to consume.");
                monitor.consume();
                System.out.println("Consumer done.");
            }
        };
        scheduler.scheduleThread(consumer);
        PalcomThread producer = new PalcomPriorityThread() {
            public void run() {
                monitor.produce(1);
                System.out.println("Producer done.");
            }
        };
        scheduler.scheduleThread(producer);
        scheduler.run();
    }

    private class ConsumerProducerMonitor extends Monitor {
        private int value = 0;

        public ConsumerProducerMonitor(PalcomScheduler scheduler) {
            super(scheduler);
        }

        private Condition valueAvailable = new Condition();

        void produce(int value) {
            enterMonitor();
            try {
                System.out.println("Producer produces " + value + ".");
                this.value = value;
                valueAvailable.causeAll();
            } finally {
                exitMonitor();
            }
        }

        void consume() {
            enterMonitor();
            try {
                while (value == 0) {
                    valueAvailable.await();
                }
                System.out.println("Consumer consumes " + value + ".");
                value = 0;
            } finally {
                exitMonitor();
            }
        }
    }

    private void testMailbox() {
        System.out.println("--- testMailbox ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final PalcomThread consumer = new PalcomPriorityThread() {
            public void run() {
                System.out.println("Consumer will wait for an event.");
                /*Event event =*/ waitEvent();
               
                System.out.println("Consumer got an event.");
                System.out.println("Consumer done.");
            }
        };
        scheduler.scheduleThread(consumer);
        PalcomThread producer = new PalcomPriorityThread() {
            public void run() {
                Event event = new Event();
                System.out
                        .println("Producer will send an event to the consumer.");
                consumer.putEvent(event);
                System.out.println("Producer done.");
            }
        };
        scheduler.scheduleThread(producer);
        scheduler.run();
    }

    private void testPeriodicEvents() {
        System.out.println("--- testPeriodicEvents ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final PalcomThread thread = new PalcomPriorityThread() {
            public void run() {
                int seconds = 0;
                enablePeriodicEvents(1000);
                while (seconds < 3) {
                    /*Event event =*/ waitEvent();
                    seconds++;
                    System.out.println(seconds + " second(s).");
                }
                disablePeriodicEvents();
                System.out.println("Done.");
            }
        };
        scheduler.scheduleThread(thread);
        scheduler.run();
    }

    private void testTerminateOwnThread() {
        System.out.println("--- testTerminateOwnThread ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final PalcomThread thread = new PalcomPriorityThread() {
            public void run() {
                while (true) {
                    System.out.println("About to call terminate from "
                            + "within an eternal loop. If this message "
                            + "is seen more than once, terminate on "
                            + "one's own thread does not work.");
                    terminate();
                }
            }
        };
        scheduler.scheduleThread(thread);
        scheduler.run();
    }

    private void testTerminateOtherThread() {
        System.out.println("--- testTerminateOtherThread ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final PalcomThread t1 = new PalcomPriorityThread() {
            public void run() {
                while (true) {
                    System.out.println("If this message "
                            + "is seen more than once, terminate on "
                            + "another thread does not work.");
                    yield();
                }
            }
        };
        final PalcomThread t2 = new PalcomPriorityThread() {
            public void run() {
                t1.terminate();
            }
        };
        scheduler.scheduleThread(t1);
        scheduler.scheduleThread(t2);
        scheduler.run();
    }

    private void testIOEvent() {
        System.out.println("--- testIOEvent ---");
        PalcomScheduler scheduler = new PalcomScheduler();
        final int port = 15790;
        final PalcomThread consumer = new PalcomPriorityThread() {
            public void run() {
            	try {
	                DatagramSocket socket = new DatagramSocket(port);
	                System.out.println("Consumer will wait three seconds for an IO event.");
	                enableIOEvent(socket, 3000);
	                Event event = waitEvent();
	                if (event instanceof IOAvailableEvent) {
	                    DatagramPacket packet = socket.receive(1024);
                            System.out.print("Consumer received ");
                            byte[] data = packet.getData();
                            System.out
                                .print(new String(data, 0, packet.getLength()));
                            System.out.println(".");
	                } else if (event instanceof IOTimeoutEvent) {
	                    System.out.println("Error: I/O timed out.");
	                }
	                System.out.println("Consumer will wait half a second for another IO event.");
	                enableIOEvent(socket, 500);
	                event = waitEvent();
	                if (event instanceof IOTimeoutEvent) {
	                    System.out.println("I/O timed out, as expected.");
	                } else if (event instanceof IOAvailableEvent) {
	                    System.out.println("Error: I/O should have timed out.");
	                }
	                disableIOEvents();
	                socket.close();
	                System.out.println("Consumer done.");
	                
            	} catch (IOException e) {
            		System.out.println("Error: I/O: " + e);
            	}
            }
        };
        scheduler.scheduleThread(consumer);
        PalcomThread producer = new PalcomPriorityThread() {
            public void run() {
            	try {
	                System.out.println("Producer will wait for one second.");
	                enablePeriodicEvents(1000);
	                waitEvent();
	                System.out.println("Producer will send a packet to the consumer.");
	                DatagramSocket socket = new DatagramSocket();
	                byte[] data = { 'd', 'a', 't', 'a' };
	                DatagramPacket packet = new DatagramPacket(data, "localhost",
	                        port);
	                socket.send(packet);
	                System.out.println("Producer will wait for one more second.");
	                waitEvent();
	                System.out.println("Producer will send another packet to the consumer.");
	                socket.send(packet);
	                disablePeriodicEvents();
	                socket.close();
	                System.out.println("Producer done.");
	                
            	} catch (IOException e) {
            		System.out.println("Error: " + e);
            	}
            }
        };
        scheduler.scheduleThread(producer);
        scheduler.run();
    }
}
