/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.pubsub;

import java.util.Iterator;

import ist.palcom.communication.CommunicationManager;
import ist.palcom.hgraph.FuncFunc;
import ist.palcom.hgraph.HgraphNode;
import ist.palcom.hgraph.IHgraphNode;
import ist.palcom.hgraph.Path;
import ist.palcom.hgraph.SiblingFunc;
import ist.palcom.hgraph.ValueFunc;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.routing.RoutingManager;

/**
 * A class that manages topic-based publish/subscribe communication.
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class PubSubManager implements IHgraphNode{

	protected final PalcomScheduler scheduler;
	protected HgraphNode _node;

	/** The CommunicationManager used for network communication */
	protected final CommunicationManager communicationManager;


	/** True while started. */
	protected boolean started;

	private final RoutingManager routingManager;

	public PubSubManager(PalcomScheduler scheduler, CommunicationManager communicationManager) {
		this.scheduler = scheduler;
		this.communicationManager = communicationManager;
		this.routingManager = null;

		/******hgraph impl*******/
		_node = new HgraphNode("pubsub manager");
	}

	public PubSubManager(PalcomScheduler scheduler,
			RoutingManager routingManager) {
		this.scheduler = scheduler;
		this.routingManager = routingManager;
		this.communicationManager = null;

		/******hgraph impl*******/
		_node = new HgraphNode("pubsub manager");
	}

	/** Stops the PubSubManager. Has no effect if it is not started. */
	public void stop() {
		if (started) {
			started = false;

			if (communicationManager != null) {
				communicationManager.stopReceivingBroadcast();
			} else {
//				URL[] addresses = routingManager.getDefaultGroupAddresses();
//				for (int i = 0; i < addresses.length; i++) {
//					routingManager.stopReceiving(addresses[i], receiverThread); TODO routingManager needs to be changed
//				}
			}
		}
	}

	/**
	 * Subscribe to a topic. Registers a PalcomThread that will receive messages
	 * in MessageEvents.
	 */
	public void subscribe(String topic, PalcomThread subscriber) {
		if (!started) {
			started = true;

			if (communicationManager != null) {
                communicationManager.startReceivingBroadcast(topic, subscriber);
			} else {
//				URL[] addresses = routingManager.getDefaultGroupAddresses();
//				for (int i = 0; i < addresses.length; i++) {
//					routingManager.startReceiving(addresses[i], receiverThread); TODO routingManager needs to be changed
//				}
			}
		}
	}

	/**
	 * Stop subscribing to a topic. The subscriber will be removed from the list
	 * of registered subscribers for that topic.
	 */
	public void unsubscribe(String topic, PalcomThread subscriber) {
		communicationManager.stopReceivingBroadcast(topic, subscriber);
	}

    /**
     * Publish a message on a topic.
     */
    public void publish(String topic, Message message) {
    	message.insertHeaderNodeFirst(new MessageHeaderNode(
				MessageHeaderNode.NODE_FORMAT_PUBLISH, 
				new String[]{topic}));
    	 communicationManager.sendBroadcast(message);
    }
    
/****************hgraph impl below********************/
	
	public String getNodeName(){return _node.getNodeName();}
	public int setNodeName(String name, Path callerPath){return _node.setNodeName(name, callerPath);}
	public Path getPath(){return _node.getPath();}
	public void updatePath(){_node.updatePath();}
	public boolean hasChildNamed(String name){return _node.hasChildNamed(name);}
	public boolean isAncestorOf(IHgraphNode node) {return _node.isAncestorOf(node);}
	public int removeNodeChild(String name, Path callerPath){return _node.removeNodeChild(name, callerPath);}
	public String toXML(){return _node.toXML();}
	public String toXMLFull(){return _node.toXMLFull();}
	public String toXMLRecursively(){return _node.toXMLRecursively();}
	public String toXMLFullRecursively(){return _node.toXMLFullRecursively();}
	public int setNodeParent(IHgraphNode newParent, Path callerPath) {return _node.setNodeParent(newParent, callerPath);}
	public int addNodeChild(IHgraphNode node, Path callerPath) {return _node.addNodeChild(node, callerPath);}	
	public SiblingFunc getSiblingFunc(){return _node.getSiblingFunc();};
	public ValueFunc getValueFunc(){return _node.getValueFunc();};
	public FuncFunc getFuncFunc(){return _node.getFuncFunc();};	
	public IHgraphNode getNodeParent(){return _node.getNodeParent();}
	public Iterator getChildIterator(){return _node.getChildIterator();}

}