/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.rascal;

import ist.palcom.base.networking.DatagramPacket;
import ist.palcom.base.networking.DatagramSocket;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageArrayWithURLEvent;
import ist.palcom.mal.MessageEvent;
import ist.palcom.mal.StopEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.IOAvailableEvent;
import ist.palcom.palcomthreads.IOTimeoutEvent;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;

import java.io.IOException;

/**
 * RascalCommunicationThread is a CommunicationThread used to enable the
 * communication between the PalCom device and the RASCAL agent. 
 * It sends and receives PalCom messages using a socket opened on localhost.
 * 
 * @author rgh
 *
 */
public class RascalCommunicationThread  extends CommunicationThread {

	public static final String URL_PREFIX = "rascal://";
	public static final int LOCAL_SOCKET_PORT = 3001;
	public static final int REMOTE_SOCKET_PORT = 3000;
	        
	protected DatagramSocket socket;

	private PalcomThread distributorThread;
	
	public RascalCommunicationThread(PalcomScheduler scheduler,
			PalcomThread distributorThread) {
		
		this.distributorThread = distributorThread;
		
		try {
			socket = new DatagramSocket(LOCAL_SOCKET_PORT);
			
		} catch (IOException e) {
			distributorThread.putEvent(new MALErrorEvent(e));
		}
    }
	
	public void stop(){;}
	
    public void run() {
        final int timeout = 1000;
        enableIOEvent(socket, timeout);
        while (true) {
            Event event = waitEvent();
            if (event instanceof IOAvailableEvent) {
                try {
                    DatagramPacket packet = socket.receive(socket.getReceiveBufferSize());
	                if (packet != null) {
	                	RascalMessage rascalMsg = RascalMessage.parse(packet.getData(), 0, packet.getLength());
	    				
	            		byte[] copyArray = new byte[rascalMsg.getMessage().getTotalLength()];
	            		rascalMsg.getMessage().write(copyArray, 0);
	                	
                        if (distributorThread != null) 
                        	distributorThread.putEvent(new MessageEvent(rascalMsg.getMessage(), null));

	                	
	                }
	                enableIOEvent(socket, timeout);
	                
                } catch (IOException e) {
                	distributorThread.putEvent(new MALErrorEvent(e));
                }
            }
            else if (event instanceof IOTimeoutEvent) {
                // Try again
                enableIOEvent(socket, timeout);
            } else if (event instanceof MessageArrayWithURLEvent) {
                try {
                    MessageArrayWithURLEvent mawue = (MessageArrayWithURLEvent) event;

                    Message message = Message.parse(mawue.getMessageArray(), 0, mawue.getMessageArray().length);
                    
                    RascalMessage msg = new RascalMessage(new URN(""), message); 
                    byte[] messageData = new byte[msg.getTotalLength()];
                    msg.write(messageData, 0);

                    DatagramPacket datagram = new DatagramPacket(messageData,
                    		"localhost", REMOTE_SOCKET_PORT);
                    socket.send(datagram);
                    
                } catch (IOException e) {
					e.printStackTrace();
				}
            } else if (event instanceof RascalEvent) {
            	try {
	            	RascalMessage msg = new RascalMessage(
	            			((RascalEvent)event).getURN(), 
	            			((RascalEvent)event).getMsg()); 
	                
	                byte[] messageData = new byte[msg.getTotalLength()];
	                msg.write(messageData, 0);
	
	                DatagramPacket datagram = new DatagramPacket(messageData,
	                		"localhost", REMOTE_SOCKET_PORT);
	                socket.send(datagram);
                } catch (IOException e) {
					e.printStackTrace();
				}
                
            } else if (event instanceof StopEvent) {
                stop();
                terminate();
            }
        }
    }

	public URL getURL() {
		return new URL(RascalMediaManager.MY_UNICAST_ADDRESS);
	}
	
    public void setReceiveBufferSize(int size) throws IOException {
        socket.setReceiveBufferSize(size);
    }

    public void setSendBufferSize(int size) throws IOException {
        socket.setSendBufferSize(size);
    }
}