/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.rascal;

import ist.palcom.collections.LinkedList;
import ist.palcom.mal.CommunicationThread;
import ist.palcom.mal.DeviceCacheObject;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.mal.StopEvent;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.rascal.internalevent.RegisterHeartBeatEvent;
import ist.palcom.rascal.internalevent.StartReceiverRequest;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;

import java.io.IOException;
import java.util.ArrayList;

/**
 * This media manager sends and receives messages to the RASCAL agent
 * running on the same host through a socket. The RASCAL agent takes care
 * for the dispatching.
 * 
 * @author rgh
 *
 */
public class RascalMediaManager implements MediaManager {
	
	public static final String PALCOM_URL_PREFIX = "palcom:";
	
	/**
	 * This thread is used to distribute messages to the threads
	 * registered as listeners
	 */
	protected DistributorThread distributorThread;

	/**
	 * This thread is used to send/receive messages to the agent 
	 */
	private RascalCommunicationThread communicationThread;

	/**
	 * This thread belongs to the CommunicationManager
	 */
	private PalcomThread communicationListenerThread;
	
	/**
	 * List of threads listening for heartbeats
	 */
	private ArrayList heartBeatReceivers;
	
	protected final PalcomScheduler scheduler;
	
	public static final String MY_MULTICAST_ADDRESS = "rascal://localhost:0000";
	public static final String MY_UNICAST_ADDRESS = "rascal://localhost:1111";
	
    private int defaultReceiveBufferSize;
    private int defaultSendBufferSize;
    private int suggestedSendBufferSize;
    private int suggestedReceiveBufferSize;

    /**
     * @deprecated  Currently, this API is replaced by
     *              {@link ist.palcom.services.communication.rascal.RascalMediaManager#
     *              	RascalMediaManager(PalcomScheduler scheduler, PalcomThread errorHandler)}
     * This means that the RASCAL agent will run within the Palcom device instance. 
     */
    public RascalMediaManager(PalcomScheduler scheduler, PalcomThread errorHandler) {

		this.scheduler = scheduler;
		heartBeatReceivers = new ArrayList();
        
		defaultReceiveBufferSize = 8192;
        defaultSendBufferSize = 8192;
        suggestedSendBufferSize = 8192;
        suggestedReceiveBufferSize = 8192;

        distributorThread = new DistributorThread();
		communicationThread = new RascalCommunicationThread(scheduler, distributorThread);
		try {
			communicationThread.setReceiveBufferSize(defaultReceiveBufferSize);
			communicationThread.setSendBufferSize(defaultSendBufferSize);
		} catch (IOException e) {
			e.printStackTrace();
		}
		scheduler.scheduleThread(communicationThread);
	}

	public void registerAsListener(PalcomThread listener) {

		/*
    	 * It asks the RASCAL agent to start a "real" receiver for this thread. 
    	 */
		StartReceiverRequest request = new StartReceiverRequest(new URL(""));
		Message msg = new Message(request.getXMLData());
	
		RascalEvent event = new RascalEvent(msg, null);
		
		communicationThread.putEvent(event);

		communicationListenerThread = listener;
		if (distributorThread.getScheduler() == null) {
			scheduler.scheduleThread(distributorThread);
		}
		URL[] urls = getDefaultGroupAddresses();
		for(int i = 0; i < urls.length; ++i) {
			URL protocolSpecificURL = getProtocolSpecificUrl(urls[i]);
			//Creates and starts the threads
			getReceivingThread(protocolSpecificURL);
		}

	}

	public void registerForHeartBeats(PalcomThread listener) {

		/*
    	 * It asks the RASCAL agent to start a "real" heartbeat 
    	 * listener for this thread. 
    	 */
		RegisterHeartBeatEvent request = new RegisterHeartBeatEvent();
		Message msg = new Message(request.getXMLData());
		RascalEvent event = new RascalEvent(msg, null);
		communicationThread.putEvent(event);

		heartBeatReceivers.add(listener);
	}
	
	public DistributorThread getDistributorThread() {
		return distributorThread;
	}

	
    public void suggestSendBufferSize(int size) {
   		try {
   			/*
   			 * This event is sent to all the "real" media managers
   			 */
   			RegisterHeartBeatEvent request = new RegisterHeartBeatEvent();
   			Message msg = new Message(request.getXMLData());
   			RascalEvent event = new RascalEvent(msg, null);
   			communicationThread.putEvent(event);
   			
   			communicationThread.setSendBufferSize(defaultSendBufferSize);
   		} catch (IOException e) {
   			e.printStackTrace();
   		}
    }
	
	public void stop() {
		if ( communicationThread != null )
			communicationThread.putEvent(new StopEvent());
		
		distributorThread.terminate();
    }

    
	/**
	 * @return The protocol-specific part of the URL (the URL that can be used
	 *         for addressing the device).
	 */
	public static URL getProtocolSpecificUrl(URL url) {
		if (url == null) {
			throw new NullPointerException();
		}
		if (!url.getString().startsWith(PALCOM_URL_PREFIX)) {
			return url; // Considered a protocol-specific URL
		}
		int startIndex = PALCOM_URL_PREFIX.length();

		int colonSlashSlashIndex = url.getString().indexOf("://");
		if (colonSlashSlashIndex == -1) {
			throw new IllegalArgumentException("Wrong URL format");
		}
		int slashIndex = url.getString().indexOf("/", colonSlashSlashIndex + 3);

		if (slashIndex == -1) {
			return new URL(url.getString().substring(startIndex,
					url.getString().length()));
		} else {
			return new URL(url.getString().substring(startIndex, slashIndex));
		}
	}
	
	public URL getDefaultUnicastURL() {
		return new URL(MY_UNICAST_ADDRESS);
	}

	protected CommunicationThread getReceivingThread(URL url) {
		return communicationThread;
	}

	protected void removeReceivingThread(URL url) {
		communicationThread.putEvent(new StopEvent());
	}	

	protected CommunicationThread getSendingThread(URL url) {
		return communicationThread;
	}

	protected boolean isBroadcast(URL url) {
		if (url.toString().indexOf(MY_MULTICAST_ADDRESS) >= 0)
			return true;
		return false;
	}

	public URL[] getDefaultGroupAddresses() {
		
		URL address = new URL(PALCOM_URL_PREFIX + MY_MULTICAST_ADDRESS);
		
		return new URL[] {address};
	}
	
    /**
     * @return the default receive buffer size for UDP sockets, or zero if not
     *         set.
     */
    public int getDefaultReceiveBufferSize() {
        return defaultReceiveBufferSize;
    }

    /**
     * @return the default send buffer size for UDP sockets, or zero if not set.
     */
    public int getSendBufferSize() {
        return defaultSendBufferSize;
    }

    public void setDefaultReceiveBufferSize(int size) {
        defaultReceiveBufferSize = size;
        
        if (defaultReceiveBufferSize != 0) {
    		try {
            	communicationThread.setReceiveBufferSize(defaultReceiveBufferSize);
    		} catch (IOException e) {
    			e.printStackTrace();
    		}
        }
    }

	public boolean send(Message message, String receiverURN) {
		communicationThread.putEvent(new RascalEvent(message, new URN(receiverURN)));
		return true;
	}

	public void sendBroadcast(Message message) {
		communicationThread.putEvent(new RascalEvent(message, new URN("")));
	}

	public void addErrorHandler(PalcomThread handler) {;}


	public String getLocalDeviceCache() {
		// TODO Auto-generated method stub
		return null;
	}

	public String getLocalDeviceID() {
		// TODO Auto-generated method stub
		return null;
	}

	public boolean isDeviceRegistered(String deviceID) {
		// TODO Auto-generated method stub
		return false;
	}

	public void registerDevice(String deviceID, DeviceCacheObject device) {
		// TODO Auto-generated method stub
		
	}

	public void sendHeartAttack() {
		// TODO Auto-generated method stub
		
	}

	public void sendHeartBeat(boolean isHeartBeatAck) {
		// TODO Auto-generated method stub
		
	}

	public void setDiscoverySelector(Selector selector) {
		// TODO Auto-generated method stub
		
	}

	public void setLocalDeviceCache(String deviceCache) {
		// TODO Auto-generated method stub
		
	}

	public void setLocalDeviceID(String deviceID) {
		// TODO Auto-generated method stub
		
	}

	public void setLocalDiscoveryVersion(String discoveryVersion) {
		// TODO Auto-generated method stub
		
	}

	public void setLocalStatusFlag(byte statusFlag) {
		// TODO Auto-generated method stub
		
	}

	public boolean unRegisterDevice(String deviceID) {
		// TODO Auto-generated method stub
		return false;
	}
	
	public class DistributorThread extends PalcomPriorityThread {
		public void run() {
			while (true) {
				Event event = waitEvent();
			}
		}
	}

	public void sendHeartBeat(boolean isHeartBeatAck, LinkedList messageHeaderNodes) {
		// TODO Auto-generated method stub
	}
	
	public ArrayList getRegisteredDevices() {
		// TODO Auto-generated method stub
		return null;
	}
	
	public void setPeriod(long period) {
		//TODO Auto-generated method stub
	}
}
