/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

 * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

 * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ist.palcom.resource.manager;

import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.communication.CommunicationManager;
import ist.palcom.discovery.DiscoveryInfoXMLFactory;
import ist.palcom.discovery.DiscoveryManager;
import ist.palcom.discovery.InfoEvent;
import ist.palcom.discovery.SelectorEvent;
import ist.palcom.discovery.ServiceInstanceIDEvent;
import ist.palcom.discovery.ServiceStatusEvent;
import ist.palcom.mal.HeartAttackEvent;
import ist.palcom.mal.HeartBeatEvent;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.messaging.PointToPointManager;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.HeartBeat;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.LocalServiceIDReply;
import ist.palcom.resource.descriptor.LocalServiceIDRequest;
import ist.palcom.resource.descriptor.PRDConnection;
import ist.palcom.resource.descriptor.PRDConnectionList;
import ist.palcom.resource.descriptor.PRDConnectionListReply;
import ist.palcom.resource.descriptor.PRDConnectionListRequest;
import ist.palcom.resource.descriptor.PRDData;
import ist.palcom.resource.descriptor.PRDDevice;
import ist.palcom.resource.descriptor.PRDDeviceReply;
import ist.palcom.resource.descriptor.PRDDeviceRequest;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.PRDServiceFMDescription;
import ist.palcom.resource.descriptor.PRDServiceFMDescriptionReply;
import ist.palcom.resource.descriptor.PRDServiceFMDescriptionRequest;
import ist.palcom.resource.descriptor.PRDServiceList;
import ist.palcom.resource.descriptor.PRDServiceListItem;
import ist.palcom.resource.descriptor.PRDServiceListReply;
import ist.palcom.resource.descriptor.PRDServiceListRequest;
import ist.palcom.resource.descriptor.PRDSubList;
import ist.palcom.resource.descriptor.Request;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.SelectorReply;
import ist.palcom.resource.descriptor.SelectorRequest;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceInstanceIDReply;
import ist.palcom.resource.descriptor.ServiceInstanceIDRequest;
import ist.palcom.resource.descriptor.ServiceStatusReply;
import ist.palcom.resource.descriptor.ServiceStatusRequest;
import ist.palcom.xml.XMLException;
import ist.palcom.xml.XMLFactory;
import ist.palcom.xml.XMLRepresentable;

import java.util.ArrayList;

/**
 * @author David Svensson, david@cs.lth.se
 */
public class ResourceManager{

	/**
	 * The number of heartbeats a requestMessage can wait for a reply message 
	 * before it is sent out again.
	 */
	private int requestMessageTimeout = 5;

	private PalcomScheduler scheduler;

	private PointToPointManager pointToPointManager;

	private CommunicationManager communicationManager;

	private PRDDevice localPRDDevice;

	private Selector localSelector;

	/** List with PRDDataNotices where remote PRDDatas are kept. */
	private final LinkedList remotePRDDatas;

	private final LinkedList remoteDevices;

	private final LinkedList PRDDeviceReplys;

	private final LinkedList remoteSelectors;

	private final LinkedList remoteServiceInstanceIDs;

	private final LinkedList remoteServiceStatusList;

	private LinkedList requestMessages;

	private ArrayList PRDDeviceReceivers;

	private ReceiversHandler PRDServiceListReceiversHandler;

	private ReceiversHandler PRDServiceFMDescriptionReceiversHandler;

	private ReceiversHandler PRDConnectionReceiversHandler;

	private ReceiversHandler selectorReceiversHandler;

	private ReceiversHandler serviceInstanceIDReceiversHandler;

	private LocalSIDReceiversHandler localSIDReceiversHandler;

	private ServiceStatusReceiversHandler serviceStatusReceiversHandler;

	private ResourceManagerThread resourceManagerThread;

	/** XMLFactory used for creating PRDDataEvents from received data. */
	private XMLFactory factory;

	public ResourceManager(PalcomScheduler scheduler,
			PointToPointManager pointToPointManager,
			CommunicationManager communicationManager, PRDDevice localPRDDevice) {

		this.scheduler = scheduler;
		this.pointToPointManager = pointToPointManager;
		this.communicationManager = communicationManager;
		this.localPRDDevice = localPRDDevice;

		remotePRDDatas = new LinkedList();
		remoteDevices = new LinkedList();
		PRDDeviceReplys = new LinkedList();
		remoteSelectors = new LinkedList();
		remoteServiceInstanceIDs = new LinkedList();
		remoteServiceStatusList = new LinkedList();
		requestMessages = new LinkedList();

		PRDDeviceReceivers = new ArrayList();

		PRDServiceListReceiversHandler = new PRDServiceListReceiversHandler();
		PRDServiceFMDescriptionReceiversHandler = new PRDServiceFMDescriptionReceiversHandler();
		PRDConnectionReceiversHandler = new PRDConnectionReceiversHandler();
		selectorReceiversHandler = new SelectorReceiversHandler();
		serviceInstanceIDReceiversHandler = new ServiceInstanceIDReceiversHandler();
		localSIDReceiversHandler = new LocalSIDReceiversHandler();
		serviceStatusReceiversHandler = new ServiceStatusReceiversHandler();

		resourceManagerThread = new ResourceManagerThread();

		factory = new DiscoveryInfoXMLFactory();
	}

	/**
	 * Start the ResourceManager. Listening for remote PRDData requests will
	 * begin, and an initial announcement about this device will be sent out
	 * (the PRDDevice sent out in an PRDDataEvent with keepPRDData=true).
	 * 
	 * <p>
	 * The PubSubManager that was passed to the constructor must be started
	 * before this call.
	 */
	public void start() {
		scheduler.scheduleThread(resourceManagerThread);

		communicationManager.registerForHeartBeats(resourceManagerThread);
		localSelector = pointToPointManager
		.startReceiving2(resourceManagerThread);
	}

	/**
	 * Stop the ResourceManager.
	 * 
	 * <p>
	 * The PubSubManager that was passed to the constructor will not be stopped.
	 */
	public void stop() {
		resourceManagerThread.terminate();
	}

	/**
	 * Set the number of heartbeats a requestMessage can wait for a reply message 
	 * before it is sent out again.
	 */
	public void setRequestMessageTimeout(int requestMessageTimeout) {
		this.requestMessageTimeout = requestMessageTimeout;
	}

	/**
	 * (Re)initiate discovery. A heartbeat will be sent out, and all the stored
	 * PRDDatas will be sent to all the registered receivers. 
	 */
	public void initiateDiscovery() {
		communicationManager.sendHeartBeat(false);

		PRDDeviceReplyNotice deviceNotice = (PRDDeviceReplyNotice) PRDDeviceReplys.getFirst();
		while (deviceNotice != null) {
			PRDDevice PRDData = deviceNotice.PRDDeviceReply.getPRDDevice();
			firePRDData(PRDData, true, false);
			deviceNotice = (PRDDeviceReplyNotice) deviceNotice.getNext();
		}

		PRDDataNotice notice = (PRDDataNotice) remotePRDDatas.getFirst();
		while (notice != null) {
			PRDData PRDData = notice.PRDData;
			updateRemotePRDData(PRDData);
			notice = (PRDDataNotice) notice.getNext();
		}

	}

	/**
	 * Add a PRDDevice receiver thread. If the receiver is already registered,
	 * nothing happens. Otherwise, the following will happen:
	 * 
	 * <ul>
	 * <li> All known PRDDevices will immediately be sent to the receiver, in
	 * PRDDataEvents (using putEvent).
	 * 
	 * <li> Whenever a new PRDDevice arrives at the ResourceManager, an
	 * PRDDataEvent with the PRDData will be sent to the receiver.
	 * 
	 * <li> If a Device disappears, PRDDataEvents are sent to
	 * registered receivers with keepPRDData=false, so that the receivers can
	 * remove the PRDData from lists.
	 * </ul>
	 */
	public void updatePRDDeviceReceiver(PalcomThread receiver) {
		if (!PRDDeviceReceivers.contains(receiver)) {
			PRDDeviceReceivers.add(receiver);

			PRDDeviceReplyNotice notice = (PRDDeviceReplyNotice) PRDDeviceReplys.getFirst();
			while (notice != null) {
				PRDDevice PRDData = notice.PRDDeviceReply.getPRDDevice();
				receiver.putEvent(new InfoEvent(PRDData, true));
				notice = (PRDDeviceReplyNotice) notice.getNext();
			}
		}
	}

	/** Unregister a PRDDevice receiver thread. */
	public void removePRDDeviceReceiver(PalcomThread receiver) {
		PRDDeviceReceivers.remove(receiver);
	}

	/**
	 * Add a receiver thread for PRDServiceLists from the device and service identified by
	 * localSID. If the thread is already registered for that device, nothing
	 * happens. Otherwise, the following will happen:
	 * 
	 * <ul>
	 * <li> All known PRDServiceLists from the device will immediately be sent to
	 * the receiver thread, in PRDDataEvents (using putEvent).
	 * 
	 * <li> Whenever a PRDServiceList for that device arrives at the
	 * ResourceManager, an PRDDataEvent with the PRDData will be sent to the
	 * receiver.
	 * 
	 * <li> If no request for PRDServiceLists from that device is scheduled in the
	 * future, a first request will be sent directly.
	 * 
	 * <li> If a Device disappears, PRDDataEvents are sent to
	 * registered receivers with keepPRDData=false, so that the receivers can
	 * remove the PRDData from lists.
	 * </ul>
	 */
	public void updatePRDServiceListReceiver(LocalSID localSID, PalcomThread receiver) {
		if (PRDServiceListReceiversHandler.updateReceiver(localSID, receiver)) {
			PRDDataNotice notice = (PRDDataNotice) remotePRDDatas.getFirst();
			while (notice != null) {
				PRDData PRDData = notice.PRDData;
				if (PRDData instanceof PRDServiceList
						&& ((PRDServiceList)PRDData).getParentLocalSID().equals(localSID)) {
					receiver.putEvent(new InfoEvent(PRDData, true));
				}
				notice = (PRDDataNotice) notice.getNext();
			}
		}
	}

	/** Like updatePRDServiceListReceiver(LocalSID localSID, PalcomThread receiver) but for the 
	 *  top level of the tree.
	 */
	public void updatePRDServiceListReceiver(DeviceID deviceID, PalcomThread receiver) {
		updatePRDServiceListReceiver(new LocalSID(deviceID, null), receiver);
	}

	/**
	 * Unregister a receiver thread for PRDServiceLists from the device identified
	 * by localSID.
	 */
	public void removePRDServiceListReceiver(LocalSID localSID, PalcomThread receiver) {
		PRDServiceListReceiversHandler.removeReceiver(localSID, receiver);
	}

	/**
	 * Add a receiver thread for the PRDServiceFMDescription of the service
	 * identified by localSID. If the thread is already registered for that
	 * service, nothing happens. Otherwise, the behavior corresponds to the
	 * behavior for updatePRDServiceListReceiver.
	 */
	public void updatePRDServiceFMDescriptionReceiver(LocalSID localSID,
			PalcomThread receiver) {
		if (PRDServiceFMDescriptionReceiversHandler.updateReceiver(localSID,
				receiver)) {
			PRDDataNotice notice = (PRDDataNotice) remotePRDDatas.getFirst();
			while (notice != null) {
				PRDData PRDData = notice.PRDData;
//				System.err.println("ResourceManager.updatePRDSFMDRecv: looking at "+PRDData);
				if (PRDData instanceof PRDServiceFMDescription
						&& ((PRDServiceFMDescription)PRDData).getLocalSID().equals(localSID)) {
//					System.err.println("... putevent ");
					receiver.putEvent(new InfoEvent(PRDData, true));
				}
				notice = (PRDDataNotice) notice.getNext();
			}
		}
	}

	/**
	 * Unregister a receiver thread for the PRDServiceFMDescription of the service
	 * identified by serviceURN.
	 */
	public void removePRDServiceFMDescriptionReceiver(LocalSID localSID,
			PalcomThread receiver) {
		PRDServiceFMDescriptionReceiversHandler.removeReceiver(localSID, receiver);
	}

	/**
	 * Add a receiver thread for PRDConnections from the device identified by
	 * deviceID. If the thread is already registered for that device, nothing
	 * happens. Otherwise, the behavior corresponds to the behavior for
	 * updatePRDServiceListReceiver.
	 */
	public void updatePRDConnectionReceiver(DeviceID deviceID,
			PalcomThread receiver) {
		if (PRDConnectionReceiversHandler.updateReceiver(new LocalSID(deviceID, null), receiver)) {
			PRDDataNotice notice = (PRDDataNotice) remotePRDDatas.getFirst();
			while (notice != null) {
				PRDData PRDData = notice.PRDData;
				if (PRDData instanceof PRDConnectionList
						&& ((PRDConnectionList)PRDData).getDeviceID().equals(deviceID)) {
					PRDConnectionList cil = (PRDConnectionList)PRDData;
					for(int i = 0; i < cil.getNumPRDConnection(); ++i) {
						receiver.putEvent(new InfoEvent(cil.getPRDConnection(i), true));
					}
				}
				notice = (PRDDataNotice) notice.getNext();
			}
		}
	}

	/**
	 * Unregister a receiver thread for PRDConnections from the device
	 * identified by deviceID.
	 */
	public void removePRDConnectionReceiver(DeviceID deviceID,
			PalcomThread receiver) {
		PRDConnectionReceiversHandler.removeReceiver(new LocalSID(deviceID, null), receiver);
	}

	/**
	 * Add a receiver thread for the Selector of the service identified by
	 * localSID. The selector for the service is always received unless the service is unavailable.
	 */
	public void updateSelectorReceiver(LocalSID localSID,
			PalcomThread receiver) {
		selectorReceiversHandler.updateReceiver(localSID, receiver);
		SelectorNotice notice = (SelectorNotice) remoteSelectors.getFirst();
		while (notice != null) {
			if (localSID.equals(notice.localSID)) {
				receiver.putEvent(new SelectorEvent(notice.localSID, notice.selector));
			}
			notice = (SelectorNotice) notice.getNext();
		}
	}

	/**
	 * Unregister a receiver thread for Selectors from the service
	 * identified by localSID.
	 */
	public void removeSelectorReceiver(LocalSID localSID,
			PalcomThread receiver) {
		selectorReceiversHandler.removeReceiver(localSID, receiver);
	}

	/**
	 * Add a receiver thread for the ServiceInstanceID of the service identified by
	 * localSID. The serviceInstanceID for the service is always received unless the service is unavailable.
	 */
	public void updateServiceInstanceIDReceiver(LocalSID localSID,
			PalcomThread receiver) {
		serviceInstanceIDReceiversHandler.updateReceiver(localSID, receiver);
		ServiceInstanceIDNotice notice = (ServiceInstanceIDNotice) remoteServiceInstanceIDs.getFirst();
		while (notice != null) {
			if (localSID.equals(notice.localSID)) {
				receiver.putEvent(new ServiceInstanceIDEvent(notice.localSID, notice.instanceID));
			}
			notice = (ServiceInstanceIDNotice) notice.getNext();
		}
	}

	/**
	 * Unregister a receiver thread for ServiceInstanceIDs from the service
	 * identified by localSID.
	 */
	public void removeServiceInstanceIDReceiver(LocalSID localSID,
			PalcomThread receiver) {
		serviceInstanceIDReceiversHandler.removeReceiver(localSID, receiver);
	}

	/**
	 * Add a receiver thread for the LocalSID of the service identified by
	 * serviceInstanceID. The LocalSID for the service is always received unless the service is unavailable.
	 */
	public void updateLocalSIDReceiver(ServiceInstanceID serviceInstanceID,
			PalcomThread receiver) {
		localSIDReceiversHandler.updateReceiver(serviceInstanceID, receiver);
		ServiceInstanceIDNotice notice = (ServiceInstanceIDNotice) remoteServiceInstanceIDs.getFirst();
		while (notice != null) {
			if (serviceInstanceID.equals(notice.instanceID)) {
				receiver.putEvent(new ServiceInstanceIDEvent(notice.localSID, notice.instanceID));
			}
			notice = (ServiceInstanceIDNotice) notice.getNext();
		}
	}

	/**
	 * Unregister a receiver thread for LocalSIDs from the service
	 * identified by serviceInstanceID.
	 */
	public void removeLocalSIDReceiver(ServiceInstanceID serviceInstanceID,
			PalcomThread receiver) {
		localSIDReceiversHandler.removeReceiver(serviceInstanceID, receiver);
	}

	/**
	 * Add a receiver thread for the service status of the service identified by
	 * localSID. The service status for the service is always received unless the service is unavailable.
	 */
	public void updateServiceStatusReceiver(LocalSID localSID,
			PalcomThread receiver) {
		serviceStatusReceiversHandler.updateReceiver(localSID, receiver);
		ServiceStatusNotice notice = (ServiceStatusNotice) remoteServiceStatusList.getFirst();
		while (notice != null) {
			if (localSID.equals(notice.localSID)) {
				receiver.putEvent(new ServiceStatusEvent(notice.localSID, notice.status, notice.helpText));
			}
			notice = (ServiceStatusNotice) notice.getNext();
		}
	}

	/**
	 * Unregister a receiver thread for service status events from the service
	 * identified by localSID.
	 */
	public void removeServiceStatusReceiver(LocalSID localSID,
			PalcomThread receiver) {
		serviceStatusReceiversHandler.removeReceiver(localSID, receiver);
	}

	/**
	 * Unregister a receiver thread for all PRDDatas.
	 */
	public void removeReceiver(PalcomThread receiver) {
		PRDDeviceReceivers.remove(receiver);
		PRDConnectionReceiversHandler.removeReceiver(receiver);
		PRDServiceListReceiversHandler.removeReceiver(receiver);
		PRDServiceFMDescriptionReceiversHandler.removeReceiver(receiver);
		selectorReceiversHandler.removeReceiver(receiver);
		serviceInstanceIDReceiversHandler.removeReceiver(receiver);
		localSIDReceiversHandler.removeReceiver(receiver);
		serviceStatusReceiversHandler.removeReceiver(receiver);
	}

	/** @return The ReceiversHandler that matches PRDData's type, or null. */
	private ReceiversHandler getReceiversHandler(PRDData PRDData) {
		if (PRDData instanceof PRDServiceList) {
			return PRDServiceListReceiversHandler;
		} else if (PRDData instanceof PRDServiceFMDescription) {
			return PRDServiceFMDescriptionReceiversHandler;
		} else if (PRDData instanceof PRDConnection) {
			return PRDConnectionReceiversHandler;
		} else {
			return null;
		}
	}

	private void requestRemotePRDDevices(DeviceID deviceID, Selector discoverySelector) {
		Request request = new PRDDeviceRequest();
		Message requestMessage = new Message(request
				.getXMLData());
		requestMessage.setReceiverSelector(discoverySelector);
		requestMessage.setReceiverDeviceID(deviceID.toString());
		requestMessages.insertLast(new RequestMessageNotice(requestMessage, 
				new LocalSID(deviceID, null), request.getTagName()));
		boolean sent = pointToPointManager.send(requestMessage, localSelector, 
				deviceID, discoverySelector, false);
	}

	private void firePRDData(PRDData PRDData, boolean keepPRDData, boolean isRequest) {
		if (PRDData instanceof PRDDevice) {
			for (int i = 0; i < PRDDeviceReceivers.size(); i++) {
				PalcomThread receiver = (PalcomThread) PRDDeviceReceivers
				.get(i);
				receiver.putEvent(new InfoEvent(PRDData, keepPRDData, isRequest));
			}
		} else {
			ReceiversHandler handler = getReceiversHandler(PRDData);
			LocalSID localSID = null;
			if (PRDData instanceof PRDServiceList) {
				localSID = ((PRDServiceList)PRDData).getParentLocalSID();
			} else if (PRDData instanceof PRDServiceFMDescription) {
				localSID = ((PRDServiceFMDescription)PRDData).getLocalSID();
			} else if (PRDData instanceof PRDConnection) {
				localSID = new LocalSID(((PRDConnection)PRDData).getDeviceID(), null);
			}
			if (handler != null && localSID != null) {
				handler.firePRDData(localSID, PRDData, keepPRDData, isRequest);
			}
		}
	}

	/** Update the remote PRDData if it has changed and 
	 * send it to interested listeners. */
	private void updateRemotePRDData(PRDData PRDData) {
		PRDDataNotice notice = getRemotePRDDataNotice(PRDData);
		if(PRDData instanceof PRDConnectionList) {
			PRDConnectionList cil = (PRDConnectionList) PRDData;
			if (notice == null) {
				remotePRDDatas.insertLast(new PRDDataNotice(cil.getDeviceID(), cil));
				for(int i = 0; i < cil.getNumPRDConnection(); ++i) {
					firePRDData(cil.getPRDConnection(i), true, false);
				}
			} else {
				if (!notice.PRDData.equals(PRDData)) {
					PRDConnectionList oldList = (PRDConnectionList) notice.PRDData;
					int cilCounter = 0;
					for(int j = 0; j < oldList.getNumPRDConnection(); ++j) {
						PRDConnection oldPRDConnection = 
							oldList.getPRDConnection(j);
						if(cilCounter < cil.getNumPRDConnection()) {
							if(oldPRDConnection.equals(
									cil.getPRDConnection(cilCounter))) {
								firePRDData(oldPRDConnection, true, false);
								cilCounter++;
							} else {
								firePRDData(oldPRDConnection, false, false);
							}
						} else {
							firePRDData(oldPRDConnection, false, false);
						}
					}
					while(cilCounter < cil.getNumPRDConnection()) {
						firePRDData(cil.getPRDConnection(cilCounter), true, false);
						cilCounter++;
					}
					notice.PRDData = PRDData;
				} else {
					for(int i = 0; i < cil.getNumPRDConnection(); ++i) {
						firePRDData(cil.getPRDConnection(i), true, false);
					}
				}
			}
		} else {
			if (notice == null) {
				DeviceID address = null;
				if (PRDData instanceof PRDServiceList) {
					address = ((PRDServiceList)PRDData).getParentLocalSID().getDeviceID();
				} else if (PRDData instanceof PRDServiceFMDescription) {
					address = ((PRDServiceFMDescription)PRDData).getLocalSID().getDeviceID();
				}
				remotePRDDatas.insertLast(new PRDDataNotice(address, PRDData));
			} else {
				if (!notice.PRDData.equals(PRDData)) {
					// Only update if the PRDData has changed
					notice.PRDData = PRDData;
				}
			}
			firePRDData(PRDData, true, false);
		}
	}

	private void updateRemoteSelector(LocalSID localSID, Selector selector) {
		SelectorNotice notice = (SelectorNotice) remoteSelectors.getFirst();
		while(notice != null && !notice.localSID.equals(localSID)) {
			notice = (SelectorNotice) notice.getNext();
		}

		if(notice == null) {
			remoteSelectors.insertLast(new SelectorNotice(localSID, selector));
		} else {
			notice.selector = selector;
		}
		((SelectorReceiversHandler)selectorReceiversHandler).fireSelectorEvent(localSID, selector);
	}

	private void updateRemoteServiceInstanceID(LocalSID localSID, ServiceInstanceID instanceID) {
		ServiceInstanceIDNotice notice = (ServiceInstanceIDNotice) remoteServiceInstanceIDs.getFirst();
		while(notice != null && !notice.localSID.equals(localSID)) {
			notice = (ServiceInstanceIDNotice) notice.getNext();
		}

		if(notice == null) {
			remoteServiceInstanceIDs.insertLast(new ServiceInstanceIDNotice(localSID, instanceID));
		} else {
			notice.instanceID = instanceID;
		}
		((ServiceInstanceIDReceiversHandler)serviceInstanceIDReceiversHandler).fireServiceInstanceIDEvent(localSID, instanceID);
	}

	private void updateRemoteLocalSID(LocalSID localSID, ServiceInstanceID instanceID) {
		ServiceInstanceIDNotice notice = (ServiceInstanceIDNotice) remoteServiceInstanceIDs.getFirst();
		while(notice != null && !notice.instanceID.equals(instanceID)) {
			notice = (ServiceInstanceIDNotice) notice.getNext();
		}

		if(notice == null) {
			remoteServiceInstanceIDs.insertLast(new ServiceInstanceIDNotice(localSID, instanceID));
		} else {
			notice.localSID = localSID;
		}
		localSIDReceiversHandler.fireServiceInstanceIDEvent(localSID, instanceID);
	}

	private void updateRemoteServiceStatus(LocalSID localSID, byte status, String helpText) {
		ServiceStatusNotice notice = (ServiceStatusNotice) remoteServiceStatusList.getFirst();
		while(notice != null && !notice.localSID.equals(localSID)) {
			notice = (ServiceStatusNotice) notice.getNext();
		}

		if(notice == null) {
			remoteServiceStatusList.insertLast(new ServiceStatusNotice(localSID, status, helpText));
		} else {
			notice.status = status;
			notice.helpText = helpText;
		}
		serviceStatusReceiversHandler.fireServiceStatusEvent(localSID, status, helpText);
	}

	private void removeRemotePRDData(DeviceID deviceID) {
		PRDDataNotice notice = getRemotePRDDataNotice(deviceID);
		while (notice != null) {
			PRDData PRDData = notice.PRDData;
			remotePRDDatas.remove(notice);
			if(PRDData instanceof PRDConnectionList) {
				PRDConnectionList cil = (PRDConnectionList) PRDData;
				for(int i = 0; i < cil.getNumPRDConnection(); ++i) {
					firePRDData(cil.getPRDConnection(i), false, false);
				}
			} else {
				firePRDData(PRDData, false, false);
			}
			notice = getRemotePRDDataNotice(deviceID);
		}
	}

	private void removeRemotePRDDevice(DeviceID deviceID) {
		PRDDeviceReplyNotice notice = getPRDDeviceReplyNotice(deviceID);
		if (notice != null) {
			PRDDevice PRDDevice = notice.PRDDeviceReply.getPRDDevice();
			firePRDData(PRDDevice, false, false);
			PRDDeviceReplys.remove(notice);
		}
	}

	/** @return The HeartBeatNotice found for deviceID, or null. */
	private HeartBeatNotice getHeartBeatNotice(DeviceID deviceID) {
		HeartBeatNotice notice = (HeartBeatNotice) remoteDevices.getFirst();
		while (notice != null && !deviceID.equals(notice.deviceID)) {
			notice = (HeartBeatNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The PRDDeviceReplyNotice found for deviceID, or null. */
	private PRDDeviceReplyNotice getPRDDeviceReplyNotice(DeviceID deviceID) {
		PRDDeviceReplyNotice notice = (PRDDeviceReplyNotice) PRDDeviceReplys.getFirst();
		while (notice != null && !deviceID.equals(notice.deviceID)) {
			notice = (PRDDeviceReplyNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The first PRDDataNotice found for deviceID, or null. */
	private PRDDataNotice getRemotePRDDataNotice(DeviceID deviceID) {
		PRDDataNotice notice = (PRDDataNotice) remotePRDDatas.getFirst();
		while (notice != null && !deviceID.equals(notice.deviceID)) {
			notice = (PRDDataNotice) notice.getNext();
		}
		return notice;
	}

	/** @return The PRDDataNotice for the remote PRDData PRDData, or null. */
	private PRDDataNotice getRemotePRDDataNotice(PRDData PRDData) {
		PRDDataNotice notice = (PRDDataNotice) remotePRDDatas.getFirst();
		while (notice != null && !PRDData.isAboutTheSameEntityAs(notice.PRDData)) {
			notice = (PRDDataNotice) notice.getNext();
		}
		return notice;
	}

	private class ResourceManagerThread extends PalcomPriorityThread {

		public void run() {
			while (true) {
				Event event = waitEvent();

				if (event instanceof MessageEvent) {
					MessageEvent messageEvent = (MessageEvent) event;
					Message message = (Message) messageEvent
					.getMessage();

					try {
						XMLRepresentable representable = factory
						.createFromXMLData(message.getData(), 0,
								message.getDataLength());

						DeviceID deviceID = new DeviceID(message.getSenderDeviceID());
						Request request = null;

						if (representable instanceof PRDDeviceReply) {
							PRDDeviceReply PRDDeviceReply = (PRDDeviceReply) representable;
							PRDDevice PRDDevice = PRDDeviceReply.getPRDDevice();
							PRDDevice.setDeviceID(deviceID);

							RequestMessageNotice requestNotice = 
								(RequestMessageNotice) requestMessages.getFirst();
							while(requestNotice != null) {
								if(requestNotice.requestMessage.getReceiverDeviceID().equals(deviceID.toString()) 
										&& requestNotice.requestName.equals(
												new PRDDeviceRequest().getTagName())) {
									requestMessages.remove(requestNotice);
									break;
								}
								requestNotice = (RequestMessageNotice) requestNotice.getNext();
							}

							PRDDeviceReplyNotice notice = getPRDDeviceReplyNotice(deviceID);
							if (notice == null) {
								PRDDeviceReplys.insertLast(new PRDDeviceReplyNotice(
										deviceID, PRDDeviceReply));

								if(!PRDServiceListReceiversHandler.isEmpty()) {
									PRDServiceListReceiversHandler.requestAllNeeded(new LocalSID(deviceID, null));
								}
								if(!PRDConnectionReceiversHandler.isEmpty()) {
									PRDConnectionReceiversHandler.requestAllNeeded(new LocalSID(deviceID, null));
								}
								if(!selectorReceiversHandler.isEmpty()) {
									((SelectorReceiversHandler)selectorReceiversHandler).requestAllNeeded(
											new LocalSID(deviceID, null));
								}
								if(!serviceInstanceIDReceiversHandler.isEmpty()) {
									((ServiceInstanceIDReceiversHandler)serviceInstanceIDReceiversHandler).requestAllNeeded(
											new LocalSID(deviceID, null));
								}
								if(!serviceStatusReceiversHandler.isEmpty()) {
									((ServiceStatusReceiversHandler)serviceStatusReceiversHandler).requestAllNeeded(
											new LocalSID(deviceID, null));
								}
							} else {
								PRDDeviceReply oldReply = notice.PRDDeviceReply;
								if(!oldReply.getServiceCache().equals(PRDDeviceReply.getServiceCache())) {
									//If there are registered listeners request the changed PRDDatas otherwise remove the old ones
									if(!PRDServiceListReceiversHandler.isEmpty()) {
										PRDServiceListReceiversHandler.requestAllNeeded(new LocalSID(deviceID, null));
									}
									PRDDataNotice dataNotice = (PRDDataNotice) remotePRDDatas.getFirst();
									while (dataNotice != null) {
										PRDDataNotice nextNotice = (PRDDataNotice) dataNotice.getNext();
										PRDData inf = dataNotice.PRDData;
										if(inf instanceof PRDServiceList && 
												deviceID.equals(((PRDServiceList)inf).getParentLocalSID().getDeviceID())) {
											remotePRDDatas.remove(dataNotice);
										} else if(inf instanceof PRDServiceFMDescription && 
												deviceID.equals(((PRDServiceFMDescription)inf).getLocalSID().getDeviceID())) {
											remotePRDDatas.remove(dataNotice);
										}
										dataNotice = nextNotice;
									}

									//Clear any ServiceInstanceIDs for this device.
									ServiceInstanceIDNotice instanceIDNotice = 
										(ServiceInstanceIDNotice) remoteServiceInstanceIDs.getFirst();
									while(instanceIDNotice != null) {
										ServiceInstanceIDNotice nextNotice = (ServiceInstanceIDNotice) instanceIDNotice.getNext();
										if(instanceIDNotice.localSID.getDeviceID().equals(deviceID)) {
											remoteServiceInstanceIDs.remove(instanceIDNotice);
										}
										instanceIDNotice = nextNotice;
									}

								}
								if(!PRDConnectionReceiversHandler.isEmpty() && 
										!oldReply.getConnectionCache().equals(PRDDeviceReply.getConnectionCache())) {
									PRDConnectionReceiversHandler.requestAllNeeded(new LocalSID(deviceID, null));
								} else if(!oldReply.getConnectionCache().equals(PRDDeviceReply.getConnectionCache())) {
									PRDDataNotice dataNotice = (PRDDataNotice) remotePRDDatas.getFirst();
									while (dataNotice != null) {
										PRDDataNotice nextNotice = (PRDDataNotice) dataNotice.getNext();
										PRDData inf = dataNotice.PRDData;
										if(inf instanceof PRDConnectionList && 
												deviceID.equals(((PRDConnectionList)inf).getDeviceID())) {
											remotePRDDatas.remove(dataNotice);
										}
										dataNotice = nextNotice;
									}
								}
								if(!oldReply.getSelectorCache().equals(PRDDeviceReply.getSelectorCache())) {
									if(!selectorReceiversHandler.isEmpty()) {
										((SelectorReceiversHandler)selectorReceiversHandler).requestAllNeeded(
												new LocalSID(deviceID, null));
									} else {
										SelectorNotice selectorNotice = (SelectorNotice) remoteSelectors.getFirst();
										while(selectorNotice != null) {
											SelectorNotice nextNotice = (SelectorNotice) selectorNotice.getNext();
											if(selectorNotice.localSID.getDeviceID().equals(deviceID)) {
												remoteSelectors.remove(selectorNotice);
											}
											selectorNotice = nextNotice;
										}
									}
								}
								if(!oldReply.getServiceStatusCache().equals(PRDDeviceReply.getServiceStatusCache())) {
									if(!serviceStatusReceiversHandler.isEmpty()) {
										((ServiceStatusReceiversHandler)serviceStatusReceiversHandler).requestAllNeeded(
												new LocalSID(deviceID, null));
									} else {
										ServiceStatusNotice statusNotice = (ServiceStatusNotice) 
										remoteServiceStatusList.getFirst();
										while(statusNotice != null) {
											ServiceStatusNotice nextNotice = (ServiceStatusNotice) statusNotice.getNext();
											if(statusNotice.localSID.getDeviceID().equals(deviceID)) {
												remoteServiceStatusList.remove(statusNotice);
											}
											statusNotice = nextNotice;
										}
									}
								}
								notice.PRDDeviceReply = PRDDeviceReply;
							}

							HeartBeatNotice heartBeatNotice = getHeartBeatNotice(deviceID);
							if(heartBeatNotice != null) {
								PRDDevice.setDeviceStatus(heartBeatNotice.heartBeat.getDeviceStatus());
								PRDDevice.setDiscoverySelector(heartBeatNotice.heartBeat.getDiscoverySelector());
							}//TODO else what to do if there aren't any heartbeat for this device?

							firePRDData(PRDDevice, true,
									false);
						} else {

							LocalSID localSID = new LocalSID(deviceID, null);
							ServiceInstanceID serviceInstanceID = null;
							if (representable instanceof PRDConnectionListReply) {
								PRDConnectionList connectionList = 
									((PRDConnectionListReply) representable).getPRDConnectionList();
								connectionList.setDeviceID(deviceID);
								request = new PRDConnectionListRequest();
								for(int i = 0; i < connectionList.getNumPRDConnection(); ++i) {
									PRDConnection connection = connectionList.getPRDConnection(i);
									connection.getLocalSID2().setDeviceID(deviceID);
								}
								updateRemotePRDData(connectionList);
							} else if (representable instanceof PRDServiceListReply) {
								PRDServiceList serviceList = 
									((PRDServiceListReply) representable).getPRDServiceList();
								if(serviceList.getParentLocalSID() != null) {
									serviceList.getParentLocalSID().setDeviceID(deviceID);
								} else {
									serviceList.setParentLocalSID(new LocalSID(deviceID, null));
								}
								for(int i = 0; i < serviceList.getNumPRDServiceListItem(); ++i) {
									PRDServiceListItem listItem = serviceList.getPRDServiceListItem(i);
									if(listItem instanceof PRDService) {
										PRDService service = ((PRDService)listItem);
										service.getLocalSID().setDeviceID(deviceID);
										if(PRDServiceFMDescriptionReceiversHandler.
												receiversInterested(service.getLocalSID()) 
												&& service.getHasDesc()) {
											PRDServiceFMDescriptionReceiversHandler.requestAllNeeded(
													service.getLocalSID());
										}
										if(serviceInstanceIDReceiversHandler.receiversInterested(service.getLocalSID())) {
											serviceInstanceIDReceiversHandler.requestAllNeeded(service.getLocalSID());
										}
									} else if(listItem instanceof PRDSubList) {
										((PRDSubList)listItem).getListLocalSID().setDeviceID(deviceID);
										if(PRDServiceListReceiversHandler.
												receiversInterested(((PRDSubList)listItem).getListLocalSID())) {
											PRDServiceListReceiversHandler.
											requestAllNeeded(((PRDSubList)listItem).getListLocalSID());
										}
									}
								}
								localSID = serviceList.getParentLocalSID();
								request = new PRDServiceListRequest();
								updateRemotePRDData(serviceList);
							} else if (representable instanceof PRDServiceFMDescriptionReply) {
								PRDServiceFMDescription serviceFMDescription = 
									((PRDServiceFMDescriptionReply) representable).getPRDServiceFMDescription();
								if(serviceFMDescription.getLocalSID() != null) {
									serviceFMDescription.getLocalSID().setDeviceID(deviceID);
								} else {
									serviceFMDescription.setLocalSID(new LocalSID(deviceID, null));
								}
								localSID = serviceFMDescription.getLocalSID();
								request = new PRDServiceFMDescriptionRequest();
								updateRemotePRDData(serviceFMDescription);
							} else if (representable instanceof SelectorReply) {
								SelectorReply selectorReply = (SelectorReply) representable;
								selectorReply.getLocalSID().setDeviceID(deviceID);
								updateRemoteSelector(selectorReply.getLocalSID(), selectorReply.getSelector());
								localSID = selectorReply.getLocalSID();
								request = new SelectorRequest();
							} else if (representable instanceof ServiceInstanceIDReply) {
								ServiceInstanceIDReply instanceIDReply = (ServiceInstanceIDReply) representable;
								instanceIDReply.getLocalSID().setDeviceID(deviceID);
								instanceIDReply.getServiceInstanceID().setDeviceID(deviceID);
								updateRemoteServiceInstanceID(instanceIDReply.getLocalSID(), instanceIDReply.getServiceInstanceID());
								localSID = instanceIDReply.getLocalSID();
								request = new ServiceInstanceIDRequest();
							} else if (representable instanceof LocalServiceIDReply) {
								LocalServiceIDReply serviceIDReply = (LocalServiceIDReply) representable;
								serviceIDReply.getLocalSID().setDeviceID(deviceID);
								serviceIDReply.getServiceInstanceID().setDeviceID(deviceID);
								updateRemoteLocalSID(serviceIDReply.getLocalSID(), serviceIDReply.getServiceInstanceID());
								serviceInstanceID = serviceIDReply.getServiceInstanceID();
								request = new LocalServiceIDRequest();
							} else if (representable instanceof ServiceStatusReply) {
								ServiceStatusReply statusReply = (ServiceStatusReply) representable;
								statusReply.getLocalSID().setDeviceID(deviceID);
								updateRemoteServiceStatus(statusReply.getLocalSID(), 
										statusReply.getServiceStatus(), statusReply.getStatusHelpText());
								localSID = statusReply.getLocalSID();
								request = new ServiceStatusRequest();
							}

							RequestMessageNotice requestNotice = 
								(RequestMessageNotice) requestMessages.getFirst();
							while(requestNotice != null && request != null) {
								RequestMessageNotice nextNotice = (RequestMessageNotice) requestNotice.getNext();
								if(request instanceof LocalServiceIDRequest) {
									if(requestNotice.serviceInstanceID != null && 
											requestNotice.serviceInstanceID.equals(serviceInstanceID) && 
											requestNotice.requestName.equals(
													request.getTagName())) {
										requestMessages.remove(requestNotice);
										break;
									}
								} else { 
									if(requestNotice.localSID != null && requestNotice.localSID.equals(localSID) 
											&& requestNotice.requestName.equals(
													request.getTagName())) {
										requestMessages.remove(requestNotice);
										break;
									}
								}
								requestNotice = nextNotice;
							}
						}
					} catch (XMLException e) {
						e.printStackTrace();
						throw new IllegalArgumentException(
								"Error in discovery XML data: " + e);
					} 
				} else if (event instanceof HeartBeatEvent) {
					HeartBeat heartBeat = ((HeartBeatEvent) event).getHeartBeat();

					if(heartBeat != null) {
						HeartBeatNotice notice = (HeartBeatNotice) remoteDevices.getFirst();
						while(notice != null && !notice.deviceID.equals(heartBeat.getDeviceID())) {
							notice = (HeartBeatNotice) notice.getNext();
						}
						DeviceID deviceID = heartBeat.getDeviceID();
						Selector discoverySelector = heartBeat.getDiscoverySelector();
						if(notice == null) {
							remoteDevices.insertLast(new HeartBeatNotice(deviceID,
									heartBeat));
							if(heartBeat.getDiscoveryVersion().equals(DiscoveryManager.DISCOVERY_VERSION)) {
								requestRemotePRDDevices(deviceID, discoverySelector);
							} else {
								//TODO An event should be sent to upper layers announcing this information
								System.out.println("ResourceManager: Device " + heartBeat.getDeviceID().toString() + 
								" is not using a compatible discovery version.");
							}
						} else {
							String deviceCache = heartBeat.getDeviceCache();
							if(!deviceCache.equals(notice.heartBeat.getDeviceCache()) && 
									heartBeat.getDiscoveryVersion().equals(DiscoveryManager.DISCOVERY_VERSION)) {
								requestRemotePRDDevices(deviceID, discoverySelector);
							}
							notice.heartBeat = heartBeat;
						}
						if(!heartBeat.getIsHeartBeatAck()) {
							RequestMessageNotice requestNotice = 
								(RequestMessageNotice) requestMessages.getFirst();
							while(requestNotice != null) {
								requestNotice.heartBeatCounter++;

								if(requestNotice.heartBeatCounter >= requestMessageTimeout) {
									requestNotice.heartBeatCounter = 0;
									Message message = requestNotice.requestMessage;
									Message requestMessage = message.getCopy();
									requestMessage.setHeaderNodes(new LinkedList());
									if(!pointToPointManager.send(requestMessage, localSelector,
											new DeviceID(message.getReceiverDeviceID()), message.getReceiverSelector(), false)) {
										RequestMessageNotice prevNotice = 
											(RequestMessageNotice) requestNotice.getPrev();
										requestMessages.remove(requestNotice);
										requestNotice = prevNotice;
										if(requestNotice != null) {
											requestNotice = (RequestMessageNotice) requestNotice.getNext();
										} else {
											requestNotice = (RequestMessageNotice) 
											requestMessages.getFirst();
										}
									} else {
										requestNotice = (RequestMessageNotice) requestNotice.getNext();
									}
								} else {
									requestNotice = (RequestMessageNotice) requestNotice.getNext();
								}
							}
						}
					}
				} else if (event instanceof HeartAttackEvent) {
					DeviceID deviceID = new DeviceID(((HeartAttackEvent) event).getDeviceID());
					HeartBeatNotice heartBeatNotice = (HeartBeatNotice) remoteDevices.getFirst();
					while (heartBeatNotice != null && !heartBeatNotice.deviceID.equals(deviceID)) {
						heartBeatNotice = (HeartBeatNotice) heartBeatNotice.getNext();
					}
					if(heartBeatNotice != null) {
						remoteDevices.remove(heartBeatNotice);
					}

					//Remove any requestMessages waiting for response
					RequestMessageNotice requestNotice = 
						(RequestMessageNotice) requestMessages.getFirst();
					while(requestNotice != null) {

						if(requestNotice.requestMessage.getReceiverDeviceID().equals(deviceID.toString())) {
							RequestMessageNotice prevNotice = 
								(RequestMessageNotice) requestNotice.getPrev();
							requestMessages.remove(requestNotice);
							requestNotice = prevNotice;

						}
						if(requestNotice != null) {
							requestNotice = (RequestMessageNotice) requestNotice.getNext();
						} else {
							requestNotice = (RequestMessageNotice) 
							requestMessages.getFirst();
						}

					}
					communicationManager.unregisterDevice(deviceID);

					removeRemotePRDDevice(deviceID);
					removeRemotePRDData(deviceID);
				}
			}
		}
	}

	/** Handles receivers for different deviceIDs and LocalSIDs for one kind of PRDData. */
	private abstract class ReceiversHandler {
		/** The list of ReceiverNotices. */
		private LinkedList receiverLists;

		ReceiversHandler() {
			receiverLists = new LinkedList();
		}

		abstract Request createRequest(LocalSID localSID);

		ReceiverByLocalSIDNotice getReceivers(LocalSID localSID) {
			ReceiverByLocalSIDNotice notice = (ReceiverByLocalSIDNotice) receiverLists.getFirst();
			while (notice != null) {
				if(notice.localSID.equals(localSID)) {
					return notice;
				}
				notice = (ReceiverByLocalSIDNotice) notice.getNext();
			}
			return notice;
		}

		boolean isEmpty() {
			return receiverLists.isEmpty();
		}

		/**
		 * @return true if there are interested receivers for the given LocalSID. If
		 *         the LocalSIDs servicePartID is null, it returns true if there are
		 *         interested receivers for any services at the device level.
		 */
		boolean receiversInterested(LocalSID localSID) {
			ReceiverByLocalSIDNotice notice = (ReceiverByLocalSIDNotice) receiverLists.getFirst();
			while (notice != null) {
				if (localSID.equals(notice.localSID)) {
					return true;
				}
				notice = (ReceiverByLocalSIDNotice) notice.getNext();
			}
			return false;
		}

		boolean updateReceiver(LocalSID localSID, PalcomThread thread) {
			boolean threadAdded = false;
			ReceiverByLocalSIDNotice notice = getReceivers(localSID);
			if (notice == null) {
				notice = new ReceiverByLocalSIDNotice(localSID);
				receiverLists.insertLast(notice);
				requestPRDDatas(localSID);
			}
			if (!notice.receiverThreads.contains(thread)) {
				notice.receiverThreads.add(thread);
				threadAdded = true;
			}

			return true;//threadAdded;
		}

		void removeReceiver(PalcomThread thread) {
			ReceiverByLocalSIDNotice notice = (ReceiverByLocalSIDNotice) receiverLists.getFirst();
			while (notice != null) {
				notice.receiverThreads.remove(thread);
				if (notice.receiverThreads.isEmpty()) {
					receiverLists.remove(notice);
				}
			}
		}

		void removeReceiver(LocalSID localSID, PalcomThread thread) {
			ReceiverByLocalSIDNotice notice = getReceivers(localSID);
			if (notice != null) {
				notice.receiverThreads.remove(thread);
				if (notice.receiverThreads.isEmpty()) {
					receiverLists.remove(notice);
				}
			}
		}

		void requestPRDDatas(LocalSID localSID) {
			HeartBeatNotice heartBeatNotice = getHeartBeatNotice(localSID.getDeviceID());
			if (heartBeatNotice != null) {
				Selector remoteSelector = heartBeatNotice.heartBeat
				.getDiscoverySelector();
				if (remoteSelector != null) {
					Request request = createRequest(localSID);
					Message requestMessage = new Message(request
							.getXMLData());

					requestMessage.setReceiverDeviceID(localSID.getDeviceID().toString());
					requestMessage.setReceiverSelector(remoteSelector);
					boolean sent = pointToPointManager.send(requestMessage, localSelector, 
							localSID.getDeviceID(), remoteSelector, false);
					if(sent) {
						requestMessages.insertLast(new RequestMessageNotice(requestMessage, localSID, 
								request.getTagName()));
					}
				}
			}
		}

		void requestAllNeeded(LocalSID localSID) {
//			if(localSID.getServicePartID() == null) {
//			ReceiverByLocalSIDNotice notice = (ReceiverByLocalSIDNotice) receiverLists.getFirst();
//			while (notice != null) {
//			if (localSID.getDeviceID().equals(notice.localSID.getDeviceID())) {
//			requestPRDDatas(notice.localSID);
//			}
//			notice = (ReceiverByLocalSIDNotice) notice.getNext();
//			}
//			} else {
			ReceiverByLocalSIDNotice notice = (ReceiverByLocalSIDNotice) receiverLists.getFirst();
			while (notice != null ) {
				if (localSID.equals(notice.localSID)) {
					requestPRDDatas(notice.localSID);
				}
				notice = (ReceiverByLocalSIDNotice) notice.getNext();
			}
//			}
		}

		void firePRDData(LocalSID localSID, PRDData PRDData, boolean keepPRDData, boolean isRequest) {
			ReceiverByLocalSIDNotice notice = getReceivers(localSID);
			if (notice != null) {
				for (int i = 0; i < notice.receiverThreads.size(); i++) {
					PalcomThread receiver = (PalcomThread) notice.receiverThreads
					.get(i);
					receiver.putEvent(new InfoEvent(PRDData, keepPRDData, isRequest));
				}
			}
		}
	}

	private class PRDServiceListReceiversHandler extends ReceiversHandler {
		Request createRequest(LocalSID lSID) {
			return new PRDServiceListRequest(lSID);
		}
	}

	private class PRDServiceFMDescriptionReceiversHandler extends ReceiversHandler {
		Request createRequest(LocalSID lSID) {
			return new PRDServiceFMDescriptionRequest(lSID);
		}
	}

	private class PRDConnectionReceiversHandler extends ReceiversHandler {
		Request createRequest(LocalSID lSID) {
			return new PRDConnectionListRequest();
		}
	}

	private class SelectorReceiversHandler extends ReceiversHandler {
		Request createRequest(LocalSID lSID) {
			return new SelectorRequest(lSID);
		}

		void fireSelectorEvent(LocalSID localSID, Selector selector) {
			ReceiverByLocalSIDNotice notice = getReceivers(localSID);
			if (notice != null) {
				for (int i = 0; i < notice.receiverThreads.size(); i++) {
					PalcomThread receiver = (PalcomThread) notice.receiverThreads
					.get(i);
					receiver.putEvent(new SelectorEvent(localSID, selector));
				}
			}
		}

		void requestAllNeeded(LocalSID localSID) {
			if(localSID.getServicePartID() == null) {
				SelectorNotice notice = (SelectorNotice) remoteSelectors.getFirst();
				while(notice != null) { 
					if(notice.localSID.getDeviceID().equals(localSID.getDeviceID())) {
						requestPRDDatas(notice.localSID);
					}
					notice = (SelectorNotice) notice.getNext();
				}
			} else {
				requestPRDDatas(localSID);
			}
		}
	}

	private class ServiceStatusReceiversHandler extends ReceiversHandler {
		Request createRequest(LocalSID lSID) {
			return new ServiceStatusRequest(lSID);
		}

		void fireServiceStatusEvent(LocalSID localSID, byte status, String helpText) {
			ReceiverByLocalSIDNotice notice = getReceivers(localSID);
			if (notice != null) {
				for (int i = 0; i < notice.receiverThreads.size(); i++) {
					PalcomThread receiver = (PalcomThread) notice.receiverThreads
					.get(i);
					receiver.putEvent(new ServiceStatusEvent(localSID, status, helpText));
				}
			}
		}

		void requestAllNeeded(LocalSID localSID) {
			if(localSID.getServicePartID() == null) {
				ServiceStatusNotice notice = (ServiceStatusNotice) remoteServiceStatusList.getFirst();
				while(notice != null) { 
					if(notice.localSID.getDeviceID().equals(localSID.getDeviceID())) {
						requestPRDDatas(notice.localSID);
					}
					notice = (ServiceStatusNotice) notice.getNext();
				}
			} else {
				requestPRDDatas(localSID);
			}
		}
	}

	private class ServiceInstanceIDReceiversHandler extends ReceiversHandler {
		Request createRequest(LocalSID lSID) {
			return new ServiceInstanceIDRequest(lSID);
		}

		void fireServiceInstanceIDEvent(LocalSID localSID, ServiceInstanceID instanceID) {
			ReceiverByLocalSIDNotice notice = getReceivers(localSID);
			if (notice != null) {
				for (int i = 0; i < notice.receiverThreads.size(); i++) {
					PalcomThread receiver = (PalcomThread) notice.receiverThreads
					.get(i);
					receiver.putEvent(new ServiceInstanceIDEvent(localSID, instanceID));
				}
			}
		}

		void requestAllNeeded(LocalSID localSID) {
			if(localSID.getServicePartID() == null) {
				ServiceInstanceIDNotice notice = 
					(ServiceInstanceIDNotice) remoteServiceInstanceIDs.getFirst();
				while(notice != null) { 
					if(notice.localSID.getDeviceID().equals(localSID.getDeviceID())) {
						requestPRDDatas(notice.localSID);
					}
					notice = (ServiceInstanceIDNotice) notice.getNext();
				}
			} else {
				requestPRDDatas(localSID);
			}
		}
	}

	/** Handles receivers for different deviceIDs and ServiceInstanceIDs for LocalSIDs. */
	private class LocalSIDReceiversHandler {

		/** The list of ReceiverNotices. */
		private LinkedList receiverLists;

		LocalSIDReceiversHandler() {
			receiverLists = new LinkedList();
		}

		Request createRequest(ServiceInstanceID serviceInstanceID) {
			return new LocalServiceIDRequest(serviceInstanceID);
		}

		ReceiverByServiceInstanceIDNotice getReceivers(ServiceInstanceID serviceInstanceID) {
			ReceiverByServiceInstanceIDNotice notice = 
				(ReceiverByServiceInstanceIDNotice) receiverLists.getFirst();
			while (notice != null) {
				if(notice.serviceInstanceID.equals(serviceInstanceID)) {
					return notice;
				}
				notice = (ReceiverByServiceInstanceIDNotice) notice.getNext();
			}
			return notice;
		}

		boolean isEmpty() {
			return receiverLists.isEmpty();
		}

		/**
		 * @return true if there are interested receivers for the given ServiceInstanceID.
		 */
		boolean receiversInterested(ServiceInstanceID serviceInstanceID) {
			ReceiverByServiceInstanceIDNotice notice = 
				(ReceiverByServiceInstanceIDNotice) receiverLists.getFirst();
			while (notice != null) {
				if (serviceInstanceID.equals(notice.serviceInstanceID)) {
					return true;
				}
				notice = (ReceiverByServiceInstanceIDNotice) notice.getNext();
			}
			return false;
		}

		boolean updateReceiver(ServiceInstanceID serviceInstanceID, PalcomThread thread) {
			boolean threadAdded = false;
			ReceiverByServiceInstanceIDNotice notice = getReceivers(serviceInstanceID);
			if (notice == null) {
				notice = new ReceiverByServiceInstanceIDNotice(serviceInstanceID);
				receiverLists.insertLast(notice);
				requestPRDDatas(serviceInstanceID);
			}
			if (!notice.receiverThreads.contains(thread)) {
				notice.receiverThreads.add(thread);
				threadAdded = true;
			}

			return threadAdded;
		}

		void removeReceiver(PalcomThread thread) {
			ReceiverByLocalSIDNotice notice = (ReceiverByLocalSIDNotice) receiverLists.getFirst();
			while (notice != null) {
				notice.receiverThreads.remove(thread);
				if (notice.receiverThreads.isEmpty()) {
					receiverLists.remove(notice);
				}
			}
		}

		void removeReceiver(ServiceInstanceID serviceInstanceID, PalcomThread thread) {
			ReceiverByServiceInstanceIDNotice notice = getReceivers(serviceInstanceID);
			if (notice != null) {
				notice.receiverThreads.remove(thread);
				if (notice.receiverThreads.isEmpty()) {
					receiverLists.remove(notice);
				}
			}
		}

		void requestPRDDatas(ServiceInstanceID serviceInstanceID) {
			HeartBeatNotice heartBeatNotice = getHeartBeatNotice(serviceInstanceID.getDeviceID());
			if (heartBeatNotice != null) {
				Selector remoteSelector = heartBeatNotice.heartBeat
				.getDiscoverySelector();
				if (remoteSelector != null) {
					Request request = createRequest(serviceInstanceID);
					Message requestMessage = new Message(request
							.getXMLData());

					requestMessage.setReceiverDeviceID(serviceInstanceID.getDeviceID().toString());
					requestMessage.setReceiverSelector(remoteSelector);
					boolean sent = pointToPointManager.send(requestMessage, localSelector, 
							serviceInstanceID.getDeviceID(), remoteSelector, false);
					if(sent) {
						requestMessages.insertLast(new RequestMessageNotice(requestMessage, 
								serviceInstanceID, request.getTagName()));
					}
				}
			}
		}

		void requestAllNeeded(ServiceInstanceID serviceInstanceID) {
			ReceiverByServiceInstanceIDNotice notice = 
				(ReceiverByServiceInstanceIDNotice) receiverLists.getFirst();
			while (notice != null) {
				if (serviceInstanceID.equals(notice.serviceInstanceID)) {
					requestPRDDatas(notice.serviceInstanceID);
				}
				notice = (ReceiverByServiceInstanceIDNotice) notice.getNext();
			}
		}

		void fireServiceInstanceIDEvent(LocalSID localSID, ServiceInstanceID instanceID) {
			ReceiverByServiceInstanceIDNotice notice = getReceivers(instanceID);
			if (notice != null) {
				for (int i = 0; i < notice.receiverThreads.size(); i++) {
					PalcomThread receiver = (PalcomThread) notice.receiverThreads
					.get(i);
					receiver.putEvent(new ServiceInstanceIDEvent(localSID, instanceID));
				}
			}
		}
	}

	private class RequestMessageNotice extends Link{
		int heartBeatCounter;

		Message requestMessage;

		LocalSID localSID;

		String requestName;

		ServiceInstanceID serviceInstanceID;

		RequestMessageNotice(Message requestMessage, LocalSID localSID, String tagName) {
			this.requestMessage = requestMessage;
			heartBeatCounter = 0;
			requestName = tagName;
			this.localSID = localSID;
		}

		RequestMessageNotice(Message requestMessage, ServiceInstanceID serviceInstanceID, String tagName) {
			this.requestMessage = requestMessage;
			heartBeatCounter = 0;
			requestName = tagName;
			this.serviceInstanceID = serviceInstanceID;
		}
	}

	private class PRDDataNotice extends Link {
		DeviceID deviceID;

		PRDData PRDData;

		PRDDataNotice(DeviceID deviceID, PRDData PRDData) {
			this.deviceID = deviceID;
			this.PRDData = PRDData;
		}
	}

	private class SelectorNotice extends Link {
		LocalSID localSID;

		Selector selector;

		SelectorNotice(LocalSID localSID, Selector selector) {
			this.localSID = localSID;
			this.selector = selector;
		}
	}

	private class PRDDeviceReplyNotice extends Link {
		DeviceID deviceID;

		PRDDeviceReply PRDDeviceReply;

		PRDDeviceReplyNotice(DeviceID deviceID, PRDDeviceReply PRDDeviceReply) {
			this.deviceID = deviceID;
			this.PRDDeviceReply = PRDDeviceReply;
		}
	}

	private class ServiceInstanceIDNotice extends Link {
		LocalSID localSID;

		ServiceInstanceID instanceID;

		ServiceInstanceIDNotice(LocalSID localSID, ServiceInstanceID instanceID) {
			this.localSID = localSID;
			this.instanceID = instanceID;
		}
	}

	private class HeartBeatNotice extends Link {
		DeviceID deviceID;

		HeartBeat heartBeat;

		HeartBeatNotice(DeviceID deviceID, HeartBeat heartBeat) {
			this.deviceID = deviceID;
			this.heartBeat = heartBeat;
		}
	}

	/**
	 * Keeps a list of receiver threads for information for a localSID containing a deviceID.
	 */
	private class ReceiverByLocalSIDNotice extends Link {

		LocalSID localSID;

		ArrayList receiverThreads;

		ReceiverByLocalSIDNotice(LocalSID localSID) {
			this.localSID = localSID;
			receiverThreads = new ArrayList();
		}
	}

	/**
	 * Keeps a list of receiver threads for information for a serviceInstanceID containing a deviceID.
	 */
	private class ReceiverByServiceInstanceIDNotice extends Link {

		ServiceInstanceID serviceInstanceID;

		ArrayList receiverThreads;

		ReceiverByServiceInstanceIDNotice(ServiceInstanceID serviceInstanceID) {
			this.serviceInstanceID = serviceInstanceID;
			receiverThreads = new ArrayList();
		}
	}	

	private class ServiceStatusNotice extends Link {
		LocalSID localSID;

		byte status;

		String helpText;

		ServiceStatusNotice(LocalSID localSID, byte serviceStatus, String statusHelpText) {
			this.localSID = localSID;
			status = serviceStatus;
			helpText = statusHelpText;
		}
	}
}