/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing.util;


import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URN;

/**
 * DSDV-Table
 * 
 * @author Jacob Mahler-Andersen, jmahle@daimi.au.dk
 */
public class DSDVTable {
	
	private static final String seperator = "$$";
	
	private LinkedList list;

	
	/**
	 * 
	 * DSDV-Table representation
	 *
	 */
	public DSDVTable() {
		reset();
		
	}
	
	public void addPeer(DSDVPeer peer) {
		list.insertLast(peer);
	}
	
	/**
	 * Delete a peer based on the URN
	 * @param routingURN
	 */
	public void deletePeer(URN routingURN, Selector routingSelector) {
		Link current = list.getFirst(); 
		while (true) {
			if (current == null)
				break;
			
			DSDVPeer p = (DSDVPeer) current;
			
			if (p.getRoutingURN().equals(routingURN) && p.getRoutingSelector().equals(routingSelector)) {
				list.remove(current);
				break;
			}
			
			current = current.getNext();
		}
	}
	
	/**
	 * The metric must = 0 and next_hop must match
	 * @param pong true if update after a pong message and false after ping message
	 * @param routingURN the URN
	 * @return true if found peer
	 */
	public boolean updateStatus(boolean pong, URN routingURN) {
		DSDVPeer peer = (DSDVPeer) list.getFirst();
		
		boolean more = peer != null;
		while (more) {
			if (peer.getMetric() == 0 && peer.getRoutingURN().equals(routingURN)) {
				
				// Pong
				if (pong) {
					peer.setStatus(0);
					
				// Ping
				} else {
					peer.setStatus(peer.getStatus()+1);
				}
			}
			
			peer = (DSDVPeer) peer.getNext();
			more = peer != null;
		}
		
		return false;
	}
	
	/**
	 * Get a peer based on the URN
	 * @return DSDVPeer 
	 */
	public DSDVPeer getPeer(URN routingURN) {
		Link current = list.getFirst(); 
		while (true) {
			if (current == null)
				break;
			
			DSDVPeer p = (DSDVPeer) current;
			
			if (p.getRoutingURN().equals(routingURN)) {
				return (DSDVPeer) current;
			}
			
			current = current.getNext();
		}
		
		return null;
	}
	
	public LinkedList getPeers() {
		return list;
	}
	
	public void reset() {
		//list.clear();
		list = new LinkedList();
	}

	
	/**
	 * Marshalling the DSDV-table.
	 * @return String
	 */
	public String marshall() {
		String output = "";
		
		DSDVPeer peer = (DSDVPeer) list.getFirst();
		
		boolean more = peer != null;
		while (more) {
			output += peer.marshall();
			output += seperator;
			
			peer = (DSDVPeer) peer.getNext();
			more = peer != null;
		}
		
		return output;
	}
	
	/**
	 * Unmarshalling
	 * @param input The marshalled String
	 * @return DSDVTable
	 */
	public static DSDVTable unmarshall(String input) {
		DSDVTable output = new DSDVTable();
		
		int i = 0;
		int j = input.indexOf(seperator); 
		String current = "";
		
		while (j > -1) {
			current = input.substring(i, j);
			output.addPeer(DSDVPeer.unmarshall(current));
			
			i = j + seperator.length();
			j = input.indexOf(seperator, i);
		}
		
		return output;
	}

}
