/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing.util;

import ist.palcom.collections.Link;
import ist.palcom.collections.LinkedList;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;

/**
 * A class that manages topic-based publish/subscribe communication.
 * 
 * @author David Svensson, david@cs.lth.se
 */
public class RoutingPubSubManager {

	protected final PalcomScheduler scheduler;

	/** The MediaManager used for network communication */
	protected final MediaManager mediaManager;

	private PalcomThread receiverThread;

	/**
	 * A list of SubscriberListNotices with a list of SubscriberNotices for each
	 * topic.
	 */
	private LinkedList subscriberLists;

	/** True while started. */
	protected boolean started;

	private final AbstractRoutingManager routingManager;

	public RoutingPubSubManager(PalcomScheduler scheduler, MediaManager mediaManager) {
		this.scheduler = scheduler;
		this.mediaManager = mediaManager;
		this.routingManager = null;
		receiverThread = new ReceiverThread();

		subscriberLists = new LinkedList();
	}

	public RoutingPubSubManager(PalcomScheduler scheduler,
			AbstractRoutingManager routingManager) {
		this.scheduler = scheduler;
		this.routingManager = routingManager;
		this.mediaManager = null;
		receiverThread = new ReceiverThread();

		subscriberLists = new LinkedList();
	}

	/** Stops the PubSubManager. Has no effect if it is not started. */
	public void stop() {
		if (started) {
			started = false;

//			if (mediaManager != null) {
//				URL[] addresses = mediaManager.getDefaultGroupAddresses();
//				for (int i = 0; i < addresses.length; i++) {
//					mediaManager.stopReceiving(addresses[i], receiverThread);
//				}
//			} else {
//				URL[] addresses = routingManager.getDefaultGroupAddresses();
//				for (int i = 0; i < addresses.length; i++) {
//					routingManager.stopReceiving(addresses[i], receiverThread);
//				}
//			}
			receiverThread.terminate();
		}
	}

	/**
	 * Subscribe to a topic. Registers a PalcomThread that will receive messages
	 * in MessageEvents.
	 */
	public void subscribe(String topic, PalcomThread subscriber) {
		if (!started) {
			started = true;

			scheduler.scheduleThread(receiverThread);
//			if (mediaManager != null) {
//                mediaManager.startReceivingBroadcast(receiverThread);
//			} else {
//				URL[] addresses = routingManager.getDefaultGroupAddresses();
//				for (int i = 0; i < addresses.length; i++) {
//					routingManager.startReceiving(addresses[i], receiverThread);
//				}
//			}
		}

		LinkedList subscribers = getSubscriberList(topic);
		if (subscribers == null) {
			SubscriberListNotice listNotice = new SubscriberListNotice(topic);
			subscriberLists.insertLast(listNotice);
			subscribers = listNotice.getList();
		}
		subscribers.insertLast(new SubscriberNotice(subscriber));
	}

	/**
	 * Stop subscribing to a topic. The subscriber will be removed from the list
	 * of registered subscribers for that topic.
	 */
	public void unsubscribe(String topic, PalcomThread subscriber) {
		LinkedList subscribers = getSubscriberList(topic);
		if (subscribers != null) {
			SubscriberNotice notice = (SubscriberNotice) subscribers.getFirst();
			while (notice != null) {
				SubscriberNotice nextNotice = (SubscriberNotice) notice
						.getNext();
				if (notice.getSubscriber() == subscriber) {
					subscribers.remove(notice);
				}
				notice = nextNotice;
			}
		}
	}

    /**
     * Publish a message on a topic.
     */
    public void publish(String topic, Message message, Selector publisherSelector) {
        message.setSenderSelector(publisherSelector);
        broadcastMessage(topic, message, publisherSelector);
    }

	/**
	 * Returns a LinkedList with SubscriberNotices for a topic. If the topic is
	 * subscribed to, null will be returned returned.
	 */
	private LinkedList getSubscriberList(String topic) {
		SubscriberListNotice notice = (SubscriberListNotice) subscriberLists
				.getFirst();
		while (notice != null && !notice.getTopic().equals(topic)) {
			notice = (SubscriberListNotice) notice.getNext();
		}
		if (notice == null) {
			return null;
		} else {
			return notice.getList();
		}
	}

    /** Broadcasts a message with a certain topic out to the network. */
    protected void broadcastMessage(String topic, Message message,
            Selector senderSelector) {
        Message headerMessage = new Message(topic.getBytes());
        Message multiPartMessage = new Message(Message.TYPE_MULTIPART);
        multiPartMessage.addPart(headerMessage);
        multiPartMessage.addPart(message);

        mediaManager.sendBroadcast(multiPartMessage);
    }

	private class ReceiverThread extends PalcomPriorityThread {
		public void run() {
			while (true) {
				MessageEvent event = (MessageEvent) waitEvent();
				Message message = event.getMessage();
				String topic = getHeader(message);
				if (topic != null) {
					fireMessage(topic, message.getPart(1));
				}
			}
		}
	}

	/**
	 * Notifies all subscribers about a message.
	 */
	private void fireMessage(String topic, Message message) {
		LinkedList subscriberList = getSubscriberList(topic);
		if (subscriberList != null) {
			SubscriberNotice notice = (SubscriberNotice) subscriberList
					.getFirst();
			while (notice != null) {
				notice.getSubscriber()
						.putEvent(new MessageEvent(message, this));
				notice = (SubscriberNotice) notice.getNext();
			}
		}
	}

	/**
	 * @return The header string of message, or null if the message is in some
	 *         way misformatted (it should be a two-part MultiPartMessage with a
	 *         data message as its first part, containing the header).
	 */
	private static String getHeader(Message message) {
		if (message.getType() == Message.TYPE_MULTIPART) {
			if (message.getNumberOfParts() == 2
					&& message.getPart(0).getType() == 
							Message.TYPE_SINGLE) {
				// It is a multi-part message with a data message as its first
				// part of two
				Message headerMessage = message.getPart(0);
				return new String(headerMessage.getData());
			}
		}
		return null;
	}

	/** A list element in a list of subscribers. */
	private class SubscriberNotice extends Link {
		private PalcomThread subscriber;

		SubscriberNotice(PalcomThread subscriber) {
			this.subscriber = subscriber;
		}

		PalcomThread getSubscriber() {
			return subscriber;
		}
	}

	/** A list element in a list with a list of subscribers for each topic. */
	private class SubscriberListNotice extends Link {
		private String topic;

		private LinkedList list;

		SubscriberListNotice(String topic) {
			this.topic = topic;
			list = new LinkedList();
		}

		String getTopic() {
			return topic;
		}

		LinkedList getList() {
			return list;
		}
	}
}