/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.routing.util;

import ist.palcom.collections.LinearMap;
import ist.palcom.collections.LinearSet;
import ist.palcom.collections.LinkedList;
import ist.palcom.mal.DeviceCacheObject;
import ist.palcom.mal.HeartAttackEvent;
import ist.palcom.mal.HeartBeatEvent;
import ist.palcom.mal.MALErrorEvent;
import ist.palcom.mal.MediaManager;
import ist.palcom.mal.Message;
import ist.palcom.mal.MessageEvent;
import ist.palcom.mal.MessageHeaderNode;
import ist.palcom.palcomthreads.Event;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.palcomthreads.PalcomScheduler;
import ist.palcom.palcomthreads.PalcomThread;
import ist.palcom.resource.descriptor.HeartBeat;
import ist.palcom.resource.descriptor.Selector;
import ist.palcom.resource.descriptor.URL;
import ist.palcom.resource.descriptor.URN;
import ist.palcom.routing.AbstractRoutingManager2;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class handles simple routing with multiple MediaManagers. The policy is
 * to choose the first found manager that has the device registered.
 * 
 * @author Thomas Forsstrm, thomasf@cs.lth.se
 */
public class SimpleRoutingManager extends AbstractRoutingManager2 {

	private static final String ROUTING_MARK = "<R>";

	private ArrayList managers;

	private MessageThread internalThread;

	private PalcomThread communicationHandlerThread, heartBeatReceiver;

	private LinearMap deviceIDs;

	private PalcomScheduler scheduler;

	private int shortIDCounter;

	private String markID;

	public SimpleRoutingManager(ArrayList mediaManagers,
			PalcomScheduler scheduler, String mark) {
		this.managers = mediaManagers;

		internalThread = new MessageThread();

		communicationHandlerThread = null;

		deviceIDs = new LinearMap();

		this.scheduler = scheduler;

		shortIDCounter = 1;

		markID = mark;

	}

	/** Stop all MediaManagers and the internal thread. */
	public void stop() {
		super.stop();
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).stop();
		}
		internalThread.terminate();
	}

	// TODO how should this method work?
	/**
	 * Subscribe to error events from all media managers. Unhandled errors
	 * within the MediaManagers will be send to the handler in the form of
	 * {@link MALErrorEvent} events. Notice: Currently, only one error handler
	 * at a time is supported
	 */
	public void addErrorHandler(PalcomThread handler) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).addErrorHandler(handler);
		}
	}
	
	/** Add another mediamanager and start to listen for events from it. */
	public void addMediaManager(MediaManager manager) {
		manager.registerForHeartBeats(internalThread);
		manager.registerAsListener(internalThread);
		managers.add(manager);
	}

	/** Set the device cache information in all media managers for this device. */
	public void setLocalDeviceCache(String localDeviceCache) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i))
			.setLocalDeviceCache(localDeviceCache);
		}
	}

	// TODO How should this work?
	/**
	 * Returns the device cache for the first found media manager. This works
	 * since they all have the same information in their device cache.
	 */
	public String getLocalDeviceCache() {
		if (!managers.isEmpty()) {
			return ((MediaManager) managers.get(0)).getLocalDeviceCache();
		}
		return null;
	}

	/** Set the device ID in all media managers for this device. */
	public void setLocalDeviceID(String localDeviceID) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).setLocalDeviceID(localDeviceID);
		}
	}

	/** Set the status flag in all media managers for this device. */
	public void setLocalStatusFlag(byte statusFlag) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).setLocalStatusFlag(statusFlag);
		}
	}

	// TODO Remove the following method.
	// There should only be the one with a byte param
	// (This one is currently needed for Pal-J)
	public void setLocalStatusFlag(int statusFlag) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i))
			.setLocalStatusFlag((byte) statusFlag);
		}
	}

	/** Set the discovery version in all media managers for this device. */
	public void setLocalDiscoveryVersion(String discoveryVersion) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i))
			.setLocalDiscoveryVersion(discoveryVersion);
		}
	}

	/** Set the discovery selector in all media managers for this device. */
	public void setDiscoverySelector(Selector selector) {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).setDiscoverySelector(selector);
		}
	}

	/** Send a message to a specified deviceID. */
	public boolean send(Message message, String receiverDeviceID) {
		message = markMessage(message);
		Iterator itr = managers.iterator();
		while (itr.hasNext()) {
			MediaManager manager = (MediaManager) itr.next();
			if (manager.isDeviceRegistered(receiverDeviceID)) {
				Message[] messages = new Message[] {message};
				if(getSendBufferSize(receiverDeviceID) >= 0) {
					messages = getChoppedMessageHandler().chopMessage(
							getSendBufferSize(receiverDeviceID), message);
				}
				boolean success = true;
				for (int i = 0; i < messages.length; ++i) {
					if (!manager.send(messages[i], receiverDeviceID)) {
						success = false;
					}
				}
				return success;
			}
		}
		return false;
	}

	/** Send a message to all devices. */
	public void sendBroadcast(Message message) {
		message = markMessage(message);
		Message[] messages = new Message[] {message};
		if(getSendBufferSize(null) >= 0) {
			messages = getChoppedMessageHandler().chopMessage(
					getSendBufferSize(null), message);
		}
		for (int j = 0; j < messages.length; ++j) {
			for (int i = 0; i < managers.size(); ++i) {
				MediaManager manager = (MediaManager) managers.get(i);
				manager.sendBroadcast(messages[j]);
			}
		}
	}

	/**
	 * Private method used for sending routed broadcast messages to all
	 * MediaManagers except to the source where they came from.
	 */
	private void sendBroadcast(Message message, MediaManager source) {
		message = markMessage(message);
		Message[] messages = new Message[] {message};
		if(getSendBufferSize(null) >= 0) {
			messages = getChoppedMessageHandler().chopMessage(getSendBufferSize(null), message);
		}
		for (int j = 0; j < messages.length; ++j) {
			for (int i = 0; i < managers.size(); ++i) {
				MediaManager manager = (MediaManager) managers.get(i);
				if (!manager.equals(source)) {
					manager.sendBroadcast(messages[j]);
				}
			}
		}
	}

	/**
	 * Used for sending broadcast messages to a certain MediaManager.
	 */
	private void sendBroadcastToSpecifiedMAL(Message message, MediaManager destination) {
		int index = 0;
		if(!managers.isEmpty()) {
			MediaManager manager = (MediaManager) managers.get(index);
			while(index < managers.size() && !manager.equals(destination)) {
				index++;
				manager = (MediaManager) managers.get(index);
			}
			if(index < managers.size()) {
				manager.sendBroadcast(message);
			}
		}
	}

	/** Send a message saying that the device will close down */
	public void sendHeartAttack() {
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).sendHeartAttack();
		}
	}

	/**
	 * Register as a listener for heartBeatEvents. This method is used by test
	 * programs and the ResourceManager. Currently only one listener at a time
	 * is supported.
	 */
	public void registerForHeartBeats(PalcomThread resourceManager) {
		heartBeatReceiver = resourceManager;
		for (int i = 0; i < managers.size(); ++i) {
			MediaManager manager = ((MediaManager) managers.get(i));
			manager.registerForHeartBeats(internalThread);

		}
	}

	/**
	 * Send a heartbeat message to all devices saying that this device is still
	 * active.
	 */
	public void sendHeartBeat(boolean isHeartBeatAck) {
		for (int i = 0; i < managers.size(); ++i) {
			MediaManager manager = ((MediaManager) managers.get(i));
			MessageHeaderNode hNode = new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_MARK, new String[]{ROUTING_MARK, markID});
			LinkedList headerNodes = new LinkedList();
			headerNodes.insertFirst(hNode);
			manager.sendHeartBeat(isHeartBeatAck, headerNodes);
		}
	}

	/** Remove the device with ID deviceID from the list of registered devices. */
	public void unregisterDevice(String deviceID) {
		super.unregisterDevice(deviceID);
		for (int i = 0; i < managers.size(); ++i) {
			((MediaManager) managers.get(i)).unRegisterDevice(deviceID);
		}
		unRegisterShortID(getShortID(deviceID));
	}

	/**
	 * Registers as a listener to messages received by this manager from the
	 * network. Currently only one listener at a time is supported.
	 * 
	 * @param communicationManager
	 *            The DistributorThread used in CommunicationManager
	 */
	public void registerAsListener(PalcomThread communicationManager) {
		communicationHandlerThread = communicationManager;
		for (int i = 0; i < managers.size(); ++i) {
			MediaManager manager = ((MediaManager) managers.get(i));
			manager.registerAsListener(internalThread);
		}

		scheduler.scheduleThread(internalThread);
	}

	/**
	 * Get the send buffer size of the media manager that has registered the
	 * device with deviceID.
	 */
	private int getSendBufferSize(String deviceID) {
		if (deviceID != null) {
			Iterator itr = managers.iterator();
			while (itr.hasNext()) {
				MediaManager manager = (MediaManager) itr.next();
				if (manager.isDeviceRegistered(deviceID)) {
					return manager.getSendBufferSize();
				}
			}
		} // If deviceID is null or it was not found then take the smallest
		// answer.
		int smallest = ((MediaManager) managers.get(0)).getSendBufferSize();
		for (int i = 0; i < managers.size(); ++i) {
			int size = ((MediaManager) managers.get(i)).getSendBufferSize();
			if (size < smallest) {
				smallest = size;
			}
		}
		return smallest;
	}

	/**
	 * Get the shortID registered to this deviceID. Creates a new one and
	 * registers it if it is not registered.
	 */
	private String getShortID(String deviceID) {
		ArrayList devices = deviceIDs.values();
		int index = devices.indexOf(deviceID);
		String shortID = null;
		LinearSet shortIDs = deviceIDs.keySet();
		Object[] shortIDArray = shortIDs.toArray();
		if (index >= 0) {
			shortID = (String) shortIDArray[index];
		} else {
			shortID = "shID" + shortIDCounter;
			shortIDCounter++;
			deviceIDs.put(shortID, deviceID);
		}
		return shortID;
	}

	private String getRegisteredDeviceID(String shortID) {
		return (String) deviceIDs.get(shortID);
	}

	private void unRegisterShortID(String shortID) {
		deviceIDs.remove(shortID);
	}

	private Message markMessage(Message message) {
		MessageHeaderNode searchNode = (MessageHeaderNode) message.getHeaderNodes().getFirst();
		while(searchNode != null && !(searchNode.getFormatIdentifier() == MessageHeaderNode.NODE_FORMAT_MARK && 
				searchNode.getData()[0].equals(ROUTING_MARK))) {
			searchNode = (MessageHeaderNode) searchNode.getNext();
		}
		if(searchNode != null) {
			String[] data = searchNode.getData();
			message.getHeaderNodes().remove(searchNode);
			String[] newData = new String[data.length + 1];
			for(int i = 0; i < data.length; ++i) {
				newData[i] = data[i];
			}
			newData[newData.length-1] = markID;
			message.insertHeaderNodeLast(new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_MARK, newData));
		} else {
			message.insertHeaderNodeLast(new MessageHeaderNode(
					MessageHeaderNode.NODE_FORMAT_MARK, new String[]{ROUTING_MARK, markID}));
		}
		return message;
	}

	// The following methods are left from the old wire protocol,
	// they should be removed when the RoutingManager interface is changed.
	/**
	 * @deprecated
	 */
	public URL[] getDefaultGroupAddresses() {
		return null;
	}

	/**
	 * @deprecated
	 */
	public void stopReceiving(URL url, PalcomThread receiver) {
	}

	/**
	 * @deprecated
	 */
	public URL startReceiving(URL url, PalcomThread receiver) {
		return null;
	}

	/**
	 * @deprecated
	 */
	public void stopReceiving(PalcomThread receiver) {
	}

	/**
	 * @deprecated
	 */
	public URL startReceiving(PalcomThread receiver) {
		return null;
	}

	/**
	 * @deprecated
	 */
	public Selector startReceiving2(PalcomThread receiver) {
		return null;
	}

	/**
	 * @deprecated
	 */
	public void stopReceiving(Selector selector) {
	}

	/**
	 * @deprecated
	 */
	public void stopReceiving(URL url) {
	}

	/**
	 * @deprecated
	 */
	public boolean send(Message msg, URN urn1, Selector selector1, URN urn2,
			Selector selector2) {
		Iterator itr = managers.iterator();
		while (itr.hasNext()) {
			MediaManager manager = (MediaManager) itr.next();
			if (manager.isDeviceRegistered(urn2.getDeviceURN().toString())) {
				return manager.send(msg, urn2.getDeviceURN().toString());
			}
		}
		return false;
	}

	/**
	 * This private class is the thread for receiving messages from the
	 * mediamanager. The thread will distribute messages based on their type
	 * routing/ordinary.
	 */
	private class MessageThread extends PalcomPriorityThread {

		private Event event;

		public MessageThread() {
			event = null;
		}

		public void run() {
			while (true) {
				event = waitEvent();
				if (event instanceof MessageEvent) {
					Message message = ((MessageEvent) event).getMessage();

					// Remove any routingS node that might be left from a
					// message to be routed.
					LinkedList headerNodes = message.getHeaderNodes();
					MessageHeaderNode searchNode = (MessageHeaderNode) headerNodes
					.getFirst();
					while (searchNode != null
							&& searchNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_ROUTINGS) {
						searchNode = (MessageHeaderNode) searchNode.getNext();
					}

					// Found routingS node which means that the message was
					// routed.
					if (searchNode != null) {

						// Remove the old routingS node.
						headerNodes.remove(searchNode);

						// Required to make a new message where old rawData is
						// removed
						Message routedMessage = adjustMessageRawData(message);

						routedMessage.setSenderDeviceID(message.getSenderDeviceID());
						routedMessage.setHeaderNodes(headerNodes);
						message = routedMessage;
					}

					// Look for routingR node
					headerNodes = message.getHeaderNodes();
					searchNode = (MessageHeaderNode) headerNodes.getFirst();
					while (searchNode != null
							&& searchNode.getFormatIdentifier() != MessageHeaderNode.NODE_FORMAT_ROUTINGR) {
						searchNode = (MessageHeaderNode) searchNode.getNext();
					}

					String senderShortID = getShortID(message.getSenderDeviceID());
					MessageHeaderNode routingSNode = new MessageHeaderNode(
							MessageHeaderNode.NODE_FORMAT_ROUTINGS,
							new String[] { senderShortID });

					// Found routingR node which means that the message should
					// be routed.
					if (searchNode != null) {
						if(!isMarked(message)) {
							String receiverShortID = searchNode.getData()[0];
							String deviceID = getRegisteredDeviceID(receiverShortID);

							// Remove the routingR node and insert a routingS node
							// instead.
							headerNodes.remove(searchNode);
							headerNodes.insertLast(routingSNode);

							// Required to make a new message where old rawData is
							// removed
							Message routedMessage = adjustMessageRawData(message);

							routedMessage.setHeaderNodes(headerNodes);
							routedMessage = getChoppedMessageHandler().rebuildChopped(routedMessage);

							if (routedMessage != null) {
								send(routedMessage, deviceID);
							}
						}

						// This message is either broadcast and/or to the router
						// device itself
					} else {

						message = getChoppedMessageHandler().rebuildChopped(message);
						if (message != null) {
							MessageEvent changedEvent = new MessageEvent(
									message, ((MessageEvent) event).getSource());
							communicationHandlerThread.putEvent(changedEvent);

							if(!isMarked(message)) {
								MessageHeaderNode hNode = (MessageHeaderNode) headerNodes
								.getFirst();
								if (hNode != null) {
									byte formatIdentifier = (byte)hNode
									.getFormatIdentifier();
									message = message.getCopy();
									headerNodes = message.getHeaderNodes();
									headerNodes.insertLast(routingSNode);
									// Connection message
									if (formatIdentifier == MessageHeaderNode.NODE_FORMAT_CONNECTION) {
										// Broadcasted connection message
										if (hNode.getFunctionIdentifier() == MessageHeaderNode.FUNCTION_BROADCASTED) {
											sendBroadcast(
													message,
													(MediaManager) ((MessageEvent) event)
													.getSource());
										}
										// Publish message
									} else if (formatIdentifier == MessageHeaderNode.NODE_FORMAT_PUBLISH) {
										sendBroadcast(
												message,
												(MediaManager) ((MessageEvent) event)
												.getSource());
									}
								}
							}
						}
					}

				} else if (event instanceof HeartBeatEvent) {
					if (heartBeatReceiver != null) {
						heartBeatReceiver.putEvent(event);
					}
					HeartBeatEvent hEvent = (HeartBeatEvent) event;

//					Have this message been here before?
					ArrayList marks = hEvent.getMarks();
					boolean marked = false;
					for(int j = 0; j < marks.size(); ++j) {
						String[] mark = (String[])marks.get(j);
						if(mark[0].equals(ROUTING_MARK)) {
							for(int k = 1; k < mark.length; ++k) {
								if(markID.equals(mark[k])) {
									marked = true;
								}
							}
						}
					}
					if(!marked) {
						//This routing manager does not route heartbeats.
						//It answers with the registered devices from the other media manager(s).
						HeartBeat heartBeat = hEvent.getHeartBeat();
						if (!heartBeat.getIsHeartBeatAck()) {
							for(int i = 0; i < managers.size(); ++i) {
								MediaManager manager = (MediaManager) managers.get(i);
								if(!((MediaManager)hEvent.getSource()).equals(manager)) {
									ArrayList registeredDevices = manager.getRegisteredDevices();

									for(int k = 0; k < registeredDevices.size(); ++k) {
										DeviceCacheObject device = (DeviceCacheObject) registeredDevices.get(k);
										if(!((MediaManager)hEvent.getSource()).isDeviceRegistered(device.getDeviceID())) {
											String shortID = getShortID(device.getDeviceID());
											MessageHeaderNode heartBeatNode = new MessageHeaderNode(
													MessageHeaderNode.NODE_FORMAT_HEARTBEATACK,
													new String[] { device.getDeviceCache(),
															device.getDeviceStatus()});

											MessageHeaderNode routingSNode = new MessageHeaderNode(
													MessageHeaderNode.NODE_FORMAT_ROUTINGS,
													new String[] { shortID });
											Message message = new Message();
											LinkedList headerNodes = new LinkedList();
											headerNodes.insertFirst(heartBeatNode);
											headerNodes.insertLast(routingSNode);
											message.setHeaderNodes(headerNodes);

											sendBroadcastToSpecifiedMAL(message, (MediaManager)hEvent.getSource());
										}
									}
								}
							}
						}
					}
				} else if (event instanceof HeartAttackEvent) {
					if (heartBeatReceiver != null) {
						heartBeatReceiver.putEvent(event);
					}
					HeartAttackEvent hEvent = (HeartAttackEvent) event;

					ArrayList marks = hEvent.getMarks();
					boolean marked = false;
					for(int j = 0; j < marks.size(); ++j) {
						String[] mark = (String[])marks.get(j);
						if(mark[0].equals(ROUTING_MARK)) {
							for(int k = 0; k < mark.length; ++k) {
								if(markID.equals(mark[k])) {
									marked = true;
								}
							}
						}
					}
					if(!marked) {

						String shortID = getShortID(hEvent.getDeviceID());

						unRegisterShortID(shortID);

						MessageHeaderNode heartAttackNode = new MessageHeaderNode(
								MessageHeaderNode.NODE_FORMAT_HEARTATTACK);

						MessageHeaderNode routingSNode = new MessageHeaderNode(
								MessageHeaderNode.NODE_FORMAT_ROUTINGS,
								new String[] { shortID });

						Message routedMessage = new Message();
						LinkedList headerNodes = new LinkedList();
						headerNodes.insertFirst(heartAttackNode);
						headerNodes.insertLast(routingSNode);
						for(int l = 0; l < marks.size(); ++l) {
							headerNodes.insertLast(new MessageHeaderNode(
									MessageHeaderNode.NODE_FORMAT_MARK, 
									(String[])marks.get(l)));
						}
						routedMessage.setHeaderNodes(headerNodes);

						sendBroadcast(routedMessage,
								(MediaManager) ((HeartAttackEvent) event)
								.getSource());
					}
				}
			}
		}

		private boolean isMarked(Message message) {
			MessageHeaderNode searchNode = (MessageHeaderNode) message.getHeaderNodes().getFirst();
			while(searchNode != null && !(searchNode.getFormatIdentifier() == MessageHeaderNode.NODE_FORMAT_MARK && 
					searchNode.getData()[0].equals(ROUTING_MARK))) {
				searchNode = (MessageHeaderNode) searchNode.getNext();
			}
			if(searchNode != null) {
				String[] marks = searchNode.getData();
				for(int i = 0; i < marks.length; ++i) {
					if(marks[i].equals(markID)) {
						return true;
					}
				}
			}
			return false;
		}
	}
}