/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.services.test.echo;

import ist.palcom.device.DeviceContext;
import ist.palcom.discovery.CommandEvent;
import ist.palcom.palcomthreads.PalcomPriorityThread;
import ist.palcom.resource.descriptor.Command;
import ist.palcom.resource.descriptor.DeviceID;
import ist.palcom.resource.descriptor.LocalSID;
import ist.palcom.resource.descriptor.PRDService;
import ist.palcom.resource.descriptor.Param;
import ist.palcom.resource.descriptor.ServiceID;
import ist.palcom.resource.descriptor.ServiceInstanceID;
import ist.palcom.resource.descriptor.ServiceProxy;
import ist.palcom.services.AbstractService;

import java.io.IOException;

/**
 * EchoService is a service that sends out values received in in-commands
 * directly in out-commands. It has a service description that is used by
 * customers for interacting with the service.
 */
public class EchoService extends AbstractService {

    /** Thread that takes care of commands. */
    private CommandThread commandThread;

    public EchoService(DeviceContext context)
            throws IOException {
        super(context, "EchoService", true, false, PRDService.UNICAST, "P1", 
        		false, "1", "Sends the in-command as out-command.", createServiceProxy(),
        		new LocalSID(context.getDeviceInfo().getDeviceID(), context.getNewLocalSIDString()));
        
        ServiceID serviceID = new ServiceID(
        		new DeviceID("X:1"), "TF18", new DeviceID("X:1"), "TF18", null, null, null, null);
		setServiceInstanceID(new ServiceInstanceID(context.getDeviceInfo().getDeviceID(), serviceID, 
				context.getNewInstanceNumber(serviceID)));

        commandThread = new CommandThread();
        getServiceProxy().addInCommandReceiver(commandThread);
    }

    /**
     * Utility method for creating the service proxy for the service. The
     * service proxy holds the structure of commands and parameters for the
     * service (it is an instance of the ServiceDescription of the service).
     */
    static ServiceProxy createServiceProxy() {
        ServiceProxy sp = new ServiceProxy("EchoService");
        Command inCommand = new Command("In", Command.DIRECTION_IN);
        Param inValueParam = new Param("value", "text/plain");
        inCommand.addParamAndInfo(inValueParam);
        sp.addControlItemAndInfo(inCommand);
        Command outCommand = new Command("Out", Command.DIRECTION_OUT);
        Param outValueParam = new Param("value", "text/plain");
        outCommand.addParamAndInfo(outValueParam);
        sp.addControlItemAndInfo(outCommand);
        return sp;
    }

    /** The start method is overridden for scheduling the command thread. */
    public void start() throws IOException {
        context.getScheduler().scheduleThread(commandThread);
        super.start();
    }

    /** The stop method is overridden for terminating the command thread. */
    public void stop() throws IOException {
        commandThread.terminate();
        super.stop();
    }

    /** CommandThread takes care of in-commands and sends out out-commands. */
    private class CommandThread extends PalcomPriorityThread {
        public void run() {
            System.out.println("EchoService is now running.");
            while (true) {
                CommandEvent event = (CommandEvent) waitEvent();
                if ("In".equals(event.getCommand().getID())) {
                    // When an in-command comes in, send out the in param value
                    // in the out-command
                    Command inCommand = event.getCommand();
                    Param inParam = inCommand.getParam("value");
                    byte[] value = inParam.getData();
                    Command outCommand = (Command) getServiceProxy().findItem(
                            "Out");
                    Param outParam = outCommand.getParam("value");
                    outParam.setData(value);
                    outCommand.invoke();
                }
            }
        }
    }
}
