/*
   Copyright (c) 2004-2007, The PalCom Project consortium.
   See http://www.ist-palcom.org/consortium.

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

       * Redistributions in binary form must reproduce the above
         copyright notice, this list of conditions and the following
         disclaimer in the documentation and/or other materials provided
         with the distribution.

       * Neither the name of the PalCom Project nor the names of its
         contributors may be used to endorse or promote products derived
         from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ist.palcom.storage.persistent;

/**
 * A simple example of Persistent class.
 * This example show how to create a class that can
 * use the Persistent interface and use this to store
 * objects persistently into the Storage
 * 
 * @author Konrad Tollmar, konrad.tollmar@design.lth.se
 */

import java.io.IOException;

import ist.palcom.storage.pfs.Storage;
import ist.palcom.storage.pio.StorageAccessStream;


public class PersistentStore {

	static final int MAXSIZE = 100;
	static final int INTSIZE = 4;
	
	StorageAccessStream ras;
	int free;
	int fsize;
	int rsize;
	int ind[] = new int[MAXSIZE];
	
	/**
     * Create a PersistentStore
     * @param store the Storage to use
     * @param fname the filename for Storage to use
     * @param clear to create a new file or append to an existing file
     */
    
	public PersistentStore(Storage store, String fname, boolean clear) {
		try {
			ras = new StorageAccessStream(store,fname);
			if (clear) {
				clear();
			}
			else {
				open();
			}
		} catch (IOException e) {
			e.printStackTrace();
		}	
	}
	
	/**
     * Open a new PersistentStore
     */
	
	public void open() {
		try {
			free = INTSIZE;
			ras.seek(0);
			fsize = ras.readInt(); //read(); 
			rsize = 0;
			for (int i=0; i<fsize; i++) {
				ras.seek(free);
				int len = ras.readInt(); //read();
				if (len > 0)
					ind[rsize++] = free;
				else
					len = Math.abs(len);
				free = free + len + INTSIZE;
			}			
		} catch (IOException e) {
			e.printStackTrace();
		}	
	}
	
	/**
     * Close the PersistentStore
     */

	public void close() {
		try { 
			// compact();
			ras.close(); 
		} catch (IOException ioe) {}
	}

	/**
     * Clear all ojects in the PersistentStore
     */

	public void clear() {
		try {
			// reset all variables 
			ras.seek(0);
			ras.writeInt(0);
			free = INTSIZE;
			fsize = 0;
			rsize = 0;
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
     * Compact the PersistentStore
     * This is currently NOT implemented but will be needed
     */

	public void compact() {
		// Todo: Compact file
		// if (ind[i] == -1) ... ; // deleted
	}
	
	/**
     * Add an object to the PersistentStore
     * @param obj the Object that should be added
     */
	
	public int add(Persistent obj) {
		try {	
			byte[] persisted = obj.persist();
			
	        ind[rsize++] = free; // set pointer to free 

	        ras.seek(0);
	        ras.writeInt(++fsize); // inc stored #obj
			
			int len = persisted.length;
			ras.seek(free);
			ras.writeInt(len);
	        ras.write(persisted, 0, len);

	        free = free + len + INTSIZE;
	        
		    return rsize;
		    
		} catch (IOException ioe) {}
	    return -1;
	}
	
	/**
     * Change an object in the PersistentStore to a new object
     * @param index the index of the old object
     * @param obj the object that should be added
     */
	public void set(int index, Persistent obj) {
		try {	
			// invalidate reference index
			ras.seek(ind[index]);
			int nsize = ras.readInt(); //read();
			ras.seek(ind[index]);
			ras.writeInt(-nsize); // (MAXBUFF + nsize); 
			
			// add new obj
	        ind[index] = free;
	        ras.seek(0);
	        ras.writeInt(++fsize); 
			byte[] persisted = obj.persist();
			int len = persisted.length;
			ras.seek(ind[index]);
			ras.writeInt(len);
	        ras.write(persisted, 0, len);

	        // update free pointer
	        free = free + len + INTSIZE;
		} catch (IOException ioe) {}
	}
	
	/**
     * Get an object in the PersistentStore
     * @param index the index of the object
     * @param obj a ref to the object
     */
	
	public void get(int index, Persistent obj) {
		try {	
			int pos = ind[index]; 
			ras.seek(pos);
			int len = ras.readInt(); //read();
			byte[] persisted = new byte[len];
			// ras.seek(ind[index] + 1);
			ras.read(persisted,0,len);
	    	obj.resurrect( persisted );
		} catch (IOException ioe) {}
	}	
	
	/**
     * Delete an object in the PersistentStore
     * @param index the index of the object
     */
	
	public void delete(int index) {
		try {

			// invalidate index
			ras.seek(ind[index]);
			int nsize = ras.readInt(); // MAXBUFF + ras.read();
			ras.seek(ind[index]);
			ras.writeInt(-nsize);

			// reindex and resize
			for (int i = index; i< rsize; i++) {
				ind[i] = ind[i+1];
			}
			--rsize;
			    
		} catch (IOException ioe) {}
	}

	/**
     * Get the size of the PersistentStore
     */
	
	public int size() {
		return rsize;
	}
	
	/**
     * Pretty print the PersistentStore, mostly for debugging
     * @param obj need a empty object as an template
     */
	
	public void print(Persistent obj) {

		System.out.println("--print ps--");
		System.out.print("size: ");
		System.out.print(rsize);
		System.out.print("/");
		System.out.print(fsize);
		System.out.print(" ");
    	
    	for (int i = 0; i<size(); i++) {
    	    this.get(i, obj);
    	    System.out.print(ind[i]);
    	    System.out.print(":");
    	    if (obj != null) {
        		System.out.print(obj.toString());
        		System.out.print(" ");    	
    	    }
    	}
    	System.out.println();
		System.out.println("--end ps--");
	}

}
